/* 
 * Mach Operating System
 * Copyright (c) 1991,1990,1989,1988 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS 
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */

#ifndef	_MACH_EXEC_H_
#define	_MACH_EXEC_H_

#include <mach/machine/vm_types.h>
#include <mach/vm_prot.h>

typedef vm_offset_t exec_entry_t;

typedef int exec_read_func_t(void *handle, vm_offset_t file_ofs, void *buf, vm_size_t size,
			     vm_size_t *out_actual);

typedef int exec_read_exec_func_t(void *handle, vm_offset_t file_ofs, vm_size_t file_size,
			          vm_offset_t mem_addr, vm_size_t mem_size, vm_prot_t mem_prot);

/*
 * Routines exported from libmach_exec.a
 */

/* Generic function to interpret an executable "file" and "load" it into "memory".
   All file I/O and destination memory accesses go through the provided functions.

   The read() function is used to read metadata from the file into the local address space.

   The read_exec() function is used to the actual executable code and data sections
   into whatever address space it's being loaded into.
   'file_size' may be less than 'mem_size'; the remaining data must be zero-filled.
   'mem_size' is always greater than zero, but 'file_size' may be zero.
   'mem_prot' is the Mach VM protection permissions for that segment.
   No two read_exec() calls for one executable
   will load data into the same virtual memory page,
   although they may load from arbitrary (possibly overlapping) file positions.

   The 'handle' is an opaque pointer passed on to the read() and read_exec() functions.

   On return, *out_entry contains the virtual address at which to start executing.
*/
int exec_load(exec_read_func_t *read, exec_read_exec_func_t *read_exec,
	      void *handle, exec_entry_t *out_entry);

/*
 * Error codes
 */

#define	EX_NOT_EXECUTABLE	6000	/* not a recognized executable format */
#define	EX_WRONG_ARCH		6001	/* valid executable, but wrong architecture */
#define EX_CORRUPT		6002	/* looks like an executable, but is mangled */
#define EX_BAD_LAYOUT		6003	/* something wrong with the memory or file image layout */


#endif	/* _MACH_EXEC_H_ */
