/*
 *  (c) Copyright 1991 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * Copyright (c) 1990,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * THE UNIVERSITY OF UTAH AND CSL PROVIDE THIS SOFTWARE IN ITS "AS IS"
 * CONDITION, AND DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM ITS USE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: som.h 1.1 94/12/03$
 */

/*
 * HP object formats
 */

#ifndef _PARISC_SOM_H_
#define _PARISC_SOM_H_

struct aux_id { 
	unsigned int mandatory : 1;
	unsigned int copy : 1;
	unsigned int append : 1;
	unsigned int ignore : 1;
	unsigned int reserved : 12;
	unsigned int type : 16;
	unsigned int length;
};

struct som_exec_auxhdr {
	struct aux_id som_auxhdr;	/* som auxiliary header header  */
	long exec_tsize;		/* text size in bytes           */
	long exec_tmem;			/* offset of text in memory     */
	long exec_tfile;		/* location of text in file     */
	long exec_dsize;		/* initialized data             */
	long exec_dmem;			/* offset of data in memory     */
	long exec_dfile;		/* location of data in file     */
	long exec_bsize;		/* uninitialized data (bss)     */
	long exec_entry;		/* offset of entrypoint         */
	long exec_flags;		/* loader flags                 */
	long exec_bfill;		/* bss initialization value     */
};

struct sys_clock { 
	unsigned int secs; 
	unsigned int nanosecs; 
};


/*
 * system_id values
 */

#define	MID_HP800	800	/* hp800 BSD binary */
#define	MID_HPUX	0x20b	/* HP-UX or OSF SOM binary */

/*
 * a_magic values
 */

#define EXEC_MAGIC	0x0107		/* normal executable */
#define SHARE_MAGIC	0x0108		/* shared executable */
#define DEMAND_MAGIC	0x010B		/* demand-load executable */


struct header {
        short int system_id;                 /* magic number - system        */
        short int a_magic;                   /* magic number - file type     */
        unsigned int version_id;             /* version id; format= YYMMDDHH */
        struct sys_clock file_time;          /* system clock- zero if unused */
        unsigned int entry_space;            /* index of space containing 
                                                entry point                  */
        unsigned int entry_subspace;         /* index of subspace for 
                                                entry point                  */
        unsigned int entry_offset;           /* offset of entry point        */
        unsigned int aux_header_location;    /* auxiliary header location    */
        unsigned int aux_header_size;        /* auxiliary header size        */
        unsigned int som_length;             /* length in bytes of entire som*/
        unsigned int presumed_dp;            /* DP value assumed during 
						compilation 		     */
        unsigned int space_location;         /* location in file of space 
                                                dictionary                   */
        unsigned int space_total;            /* number of space entries      */
        unsigned int subspace_location;      /* location of subspace entries */
        unsigned int subspace_total;         /* number of subspace entries   */
        unsigned int loader_fixup_location;  /* space reference array        */
        unsigned int loader_fixup_total;     /* total number of space 
                                                reference records            */
        unsigned int space_strings_location; /* file location of string area
                                                for space and subspace names */
        unsigned int space_strings_size;     /* size of string area for space 
                                                 and subspace names          */
        unsigned int init_array_location;    /* location in file of 
                                                initialization pointers      */
        unsigned int init_array_total;       /* number of init. pointers     */
        unsigned int compiler_location;      /* location in file of module 
                                                dictionary                   */
        unsigned int compiler_total;         /* number of modules            */
        unsigned int symbol_location;        /* location in file of symbol 
                                                dictionary                   */
        unsigned int symbol_total;           /* number of symbol records     */
        unsigned int fixup_request_location; /* location in file of fix_up 
                                                requests                     */
        unsigned int fixup_request_total;    /* number of fixup requests     */
        unsigned int symbol_strings_location;/* file location of string area 
                                                for module and symbol names  */
        unsigned int symbol_strings_size;    /* size of string area for 
                                                module and symbol names      */
        unsigned int unloadable_sp_location; /* byte offset of first byte of 
                                                data for unloadable spaces   */
        unsigned int unloadable_sp_size;     /* byte length of data for 
                                                unloadable spaces            */
        unsigned int checksum;
};

typedef struct {
  struct header fhdr;
  struct som_exec_auxhdr ehdr;
} som_exec;

#endif	/* _PARISC_SOM_H_ */
