/* 
 * Copyright (c) 1990,1991,1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: genassym.c 1.14 95/01/20$
 *	Author: Bob Wheeler, University of Utah CSL
 */

/*
 *	Generate assembly symbols
 */

#include <mach/message.h>
#include <mach/rpc.h>
#include <kern/thread.h>
#include <sys/types.h>
#include <kern/task.h>
#include <kern/act.h>
#include <ipc/ipc_space.h>
#include <ipc/ipc_port.h>
#include <ipc/ipc_pset.h>
#include <vm/vm_map.h>
#include <iotypes.h>
#include <machine/rpb.h>
#include <machine/pdc.h>
#include <machine/iomod.h>
#include <machine/pmap.h>
#include <mach/kern_return.h>

#if PORT_GENERATIONS
#error msg.s currently cannot handle port generations (could be fixed)
#endif


main()
{
	struct parisc_kernel_state *ksp = (struct parisc_kernel_state *)0;
	struct parisc_saved_state *ssp = (struct parisc_saved_state *)0;
	struct parisc_fp_state *fsp = (struct parisc_fp_state *)0;
	struct PCB *pcb = (struct PCB *)0;
	struct mapping *map = (struct mapping *)0;
	struct phys_entry *ptov = (struct phys_entry *)0;
	struct vtop_entry *vtop = (struct vtop_entry *)0;
#ifdef	HPT
	struct hpt_entry *hpt = (struct hpt_entry *)0;
#endif
	struct thread *thread = (struct thread *)0;
	struct vm_map *vm_map = (struct vm_map *)0;
	struct pmap *pmap = (struct pmap *)0;
        struct eml_dispatch *disp = (struct eml_dispatch *)0;
	struct task *task = (struct task *)0;
	struct Act *act = (struct Act *)0;
	struct rpb *rp = (struct rpb *)0;
	struct pagezero *pzp = (struct pagezero *)0;
	struct iomod *hpa = (struct iomod *)0;
	mach_msg_header_t *msgh = 0;
	struct ipc_space *space = 0;
	struct ipc_entry *entry = 0;
	struct ipc_port *port = 0;
	struct ipc_pset *pset = 0;
	struct ipc_target *ipt = 0;
	int bit;

	printf("\n/*\n * Kernel State Structure\n */\n");
	printf("#define\tKS_R2\t0x%x\n", &ksp->r2);
	printf("#define\tKS_R3\t0x%x\n", &ksp->r3);
	printf("#define\tKS_R4\t0x%x\n", &ksp->r4);
	printf("#define\tKS_R5\t0x%x\n", &ksp->r5);
	printf("#define\tKS_R6\t0x%x\n", &ksp->r6);
	printf("#define\tKS_R7\t0x%x\n", &ksp->r7);
	printf("#define\tKS_R8\t0x%x\n", &ksp->r8);
	printf("#define\tKS_R9\t0x%x\n", &ksp->r9);
	printf("#define\tKS_R10\t0x%x\n", &ksp->r10);
	printf("#define\tKS_R11\t0x%x\n", &ksp->r11);
	printf("#define\tKS_R12\t0x%x\n", &ksp->r12);
	printf("#define\tKS_R13\t0x%x\n", &ksp->r13);
	printf("#define\tKS_R14\t0x%x\n", &ksp->r14);
	printf("#define\tKS_R15\t0x%x\n", &ksp->r15);
	printf("#define\tKS_R16\t0x%x\n", &ksp->r16);
	printf("#define\tKS_R17\t0x%x\n", &ksp->r17);
	printf("#define\tKS_R18\t0x%x\n", &ksp->r18);
	printf("#define\tKS_SIZE\t0x%x\n", sizeof(struct parisc_kernel_state));

	printf("\n/*\n * Save State Structure\n */\n");
	printf("#define\tSS_FLAGS\t0x%x\n", &ssp->flags);
	printf("#define\tSS_R1\t0x%x\n", &ssp->r1);
	printf("#define\tSS_R2\t0x%x\n", &ssp->r2);
	printf("#define\tSS_R3\t0x%x\n", &ssp->r3);
	printf("#define\tSS_R4\t0x%x\n", &ssp->r4);
	printf("#define\tSS_R5\t0x%x\n", &ssp->r5);
	printf("#define\tSS_R6\t0x%x\n", &ssp->r6);
	printf("#define\tSS_R7\t0x%x\n", &ssp->r7);
	printf("#define\tSS_R8\t0x%x\n", &ssp->r8);
	printf("#define\tSS_R9\t0x%x\n", &ssp->r9);
	printf("#define\tSS_R10\t0x%x\n", &ssp->r10);
	printf("#define\tSS_R11\t0x%x\n", &ssp->r11);
	printf("#define\tSS_R12\t0x%x\n", &ssp->r12);
	printf("#define\tSS_R13\t0x%x\n", &ssp->r13);
	printf("#define\tSS_R14\t0x%x\n", &ssp->r14);
	printf("#define\tSS_R15\t0x%x\n", &ssp->r15);
	printf("#define\tSS_R16\t0x%x\n", &ssp->r16);
	printf("#define\tSS_R17\t0x%x\n", &ssp->r17);
	printf("#define\tSS_R18\t0x%x\n", &ssp->r18);
	printf("#define\tSS_R19\t0x%x\n", &ssp->r19);
	printf("#define\tSS_R20\t0x%x\n", &ssp->r20);
	printf("#define\tSS_R21\t0x%x\n", &ssp->r21);
	printf("#define\tSS_R22\t0x%x\n", &ssp->r22);
	printf("#define\tSS_R23\t0x%x\n", &ssp->r23);
	printf("#define\tSS_R24\t0x%x\n", &ssp->r24);
	printf("#define\tSS_R25\t0x%x\n", &ssp->r25);
	printf("#define\tSS_R26\t0x%x\n", &ssp->r26);
	printf("#define\tSS_R27\t0x%x\n", &ssp->r27);
	printf("#define\tSS_R28\t0x%x\n", &ssp->r28);
	printf("#define\tSS_R29\t0x%x\n", &ssp->r29);
	printf("#define\tSS_R30\t0x%x\n", &ssp->r30);
	printf("#define\tSS_R31\t0x%x\n", &ssp->r31);
	printf("#define\tSS_SR0\t0x%x\n", &ssp->sr0);
	printf("#define\tSS_SR1\t0x%x\n", &ssp->sr1);
	printf("#define\tSS_SR2\t0x%x\n", &ssp->sr2);
	printf("#define\tSS_SR3\t0x%x\n", &ssp->sr3);
	printf("#define\tSS_SR4\t0x%x\n", &ssp->sr4);
	printf("#define\tSS_SR5\t0x%x\n", &ssp->sr5);
	printf("#define\tSS_SR6\t0x%x\n", &ssp->sr6);
	printf("#define\tSS_SR7\t0x%x\n", &ssp->sr7);
	printf("#define\tSS_CR0\t0x%x\n", &ssp->cr0);
	printf("#define\tSS_CR8\t0x%x\n", &ssp->cr8);
	printf("#define\tSS_CR9\t0x%x\n", &ssp->cr9);
	printf("#define\tSS_CR11\t0x%x\n", &ssp->cr11);
	printf("#define\tSS_CR12\t0x%x\n", &ssp->cr12);
	printf("#define\tSS_CR13\t0x%x\n", &ssp->cr13);
	printf("#define\tSS_CR15\t0x%x\n", &ssp->cr15);
	printf("#define\tSS_IISQH\t0x%x\n", &ssp->iisq_head);
	printf("#define\tSS_IISQT\t0x%x\n", &ssp->iisq_tail);
	printf("#define\tSS_IIOQH\t0x%x\n", &ssp->iioq_head);
	printf("#define\tSS_IIOQT\t0x%x\n", &ssp->iioq_tail);
	printf("#define\tSS_CR19\t0x%x\n", &ssp->cr19);
	printf("#define\tSS_CR20\t0x%x\n", &ssp->cr20);
	printf("#define\tSS_CR21\t0x%x\n", &ssp->cr21);
	printf("#define\tSS_CR22\t0x%x\n", &ssp->cr22);
	printf("#define\tSS_GOTO\t0x%x\n", &ssp->bsd_goto);
	printf("#define\tSS_SIZE\t0x%x\n", sizeof(struct parisc_saved_state));

	printf("\n/*\n * Floating-point state structure\n */\n");
	printf("#define\tFS_FR0\t0x%x\n", &fsp->fr0);
	printf("#define\tFS_FR1\t0x%x\n", &fsp->fr1);
	printf("#define\tFS_FR2\t0x%x\n", &fsp->fr2);
	printf("#define\tFS_FR3\t0x%x\n", &fsp->fr3);
	printf("#define\tFS_FR4\t0x%x\n", &fsp->fr4);
	printf("#define\tFS_FR5\t0x%x\n", &fsp->fr5);
	printf("#define\tFS_FR6\t0x%x\n", &fsp->fr6);
	printf("#define\tFS_FR7\t0x%x\n", &fsp->fr7);
	printf("#define\tFS_FR8\t0x%x\n", &fsp->fr8);
	printf("#define\tFS_FR9\t0x%x\n", &fsp->fr9);
	printf("#define\tFS_FR10\t0x%x\n", &fsp->fr10);
	printf("#define\tFS_FR11\t0x%x\n", &fsp->fr11);
	printf("#define\tFS_FR12\t0x%x\n", &fsp->fr12);
	printf("#define\tFS_FR13\t0x%x\n", &fsp->fr13);
	printf("#define\tFS_FR14\t0x%x\n", &fsp->fr14);
	printf("#define\tFS_FR15\t0x%x\n", &fsp->fr15);
	printf("#define\tFS_FR16\t0x%x\n", &fsp->fr16);
	printf("#define\tFS_FR17\t0x%x\n", &fsp->fr17);
	printf("#define\tFS_FR18\t0x%x\n", &fsp->fr18);
	printf("#define\tFS_FR19\t0x%x\n", &fsp->fr19);
	printf("#define\tFS_FR20\t0x%x\n", &fsp->fr20);
	printf("#define\tFS_FR21\t0x%x\n", &fsp->fr21);
	printf("#define\tFS_FR22\t0x%x\n", &fsp->fr22);
	printf("#define\tFS_FR23\t0x%x\n", &fsp->fr23);
	printf("#define\tFS_FR24\t0x%x\n", &fsp->fr24);
	printf("#define\tFS_FR25\t0x%x\n", &fsp->fr25);
	printf("#define\tFS_FR26\t0x%x\n", &fsp->fr26);
	printf("#define\tFS_FR27\t0x%x\n", &fsp->fr27);
	printf("#define\tFS_FR28\t0x%x\n", &fsp->fr28);
	printf("#define\tFS_FR29\t0x%x\n", &fsp->fr29);
	printf("#define\tFS_FR30\t0x%x\n", &fsp->fr30);
	printf("#define\tFS_FR31\t0x%x\n", &fsp->fr31);

	printf("#define\tPCB_SS\t0x%x\n", &pcb->ss);
	printf("#define\tPCB_FS\t0x%x\n", &pcb->fs);
	printf("#define\tPCB_ACT\t0x%x\n", &pcb->ss.act);

	printf("\n/*\n * Mapping structure parameters\n */\n");
	printf("#define\tPARISC_MAP_HLINKN\t0x%x\n", &map->hash_link.next);
	printf("#define\tPARISC_MAP_HLINKP\t0x%x\n", &map->hash_link.prev);
	printf("#define\tPARISC_MAP_PLINKN\t0x%x\n", &map->phys_link.next);
	printf("#define\tPARISC_MAP_PLINKP\t0x%x\n", &map->phys_link.prev);
	printf("#define\tPARISC_MAP_SPACE\t0x%x\n", &map->space);
	printf("#define\tPARISC_MAP_OFFSET\t0x%x\n", &map->offset);
	printf("#define\tPARISC_MAP_PAGE\t0x%x\n", &map->tlbpage);
	printf("#define\tPARISC_MAP_PROT\t0x%x\n", &map->tlbprot);

	printf("\n/*\n * physical to virtual mapping parameters\n */\n");
	printf("#define\tPTOV_LINKN\t0x%x\n", &ptov->phys_link.next);
	printf("#define\tPTOV_LINKP\t0x%x\n", &ptov->phys_link.prev);
	printf("#define\tPTOV_WRITER\t0x%x\n", &ptov->writer);
	printf("#define\tPTOV_TLBPROT\t0x%x\n", &ptov->tlbprot);

	printf("\n/*\n * virtual to physical mapping parameters\n */\n");
	printf("#define\tVTOP_LINKN\t0x%x\n", &vtop->hash_link.next);
	printf("#define\tVTOP_LINKP\t0x%x\n", &vtop->hash_link.prev);

#ifdef HPT
	printf("\n/*\n * Hashed Page Table parameters\n */\n");
	printf("#define\tHPT_TAG\t0x%x\n", 0);
	printf("#define\tHPT_VTOP\t0x%x\n", &hpt->vtop_entry);
	printf("#define\tHPT_PROT\t0x%x\n", &hpt->tlbprot);
	printf("#define\tHPT_PAGE\t0x%x\n", &hpt->tlbpage);
#endif

	printf("\n/*\n * values from kern/thread.h\n */\n");
	printf("#define\tTHREAD_TOP_ACT\t0x%x\n", &thread->top_act);
        printf("#define\tTHREAD_RECOVER\t0x%x\n", &thread->mthread.recover);

	printf("#define\tACT_REF_COUNT\t0x%x\n", &act->ref_count.ref_count);
	printf("#define\tACT_TASK\t0x%x\n", &act->task);
	printf("#define\tACT_THREAD\t0x%x\n", &act->thread);
	printf("#define\tACT_LOWER\t0x%x\n", &act->lower);
	printf("#define\tACT_HIGHER\t0x%x\n", &act->higher);
	printf("#define\tACT_HANDLERS\t0x%x\n", &act->handlers);
	printf("#define\tACT_ACTIVE\t0x%x\n", &act->active);
        printf("#define\tACT_IPT\t0x%x\n", &act->ipt);
        printf("#define\tACT_IPT_NEXT\t0x%x\n", &act->ipt_next);
        printf("#define\tACT_USER_STACK\t0x%x\n", &act->mact.user_stack);
        printf("#define\tACT_USER_RBUF\t0x%x\n", &act->mact.user_rbuf);
        printf("#define\tACT_USER_RBUF_SIZE\t0x%x\n", &act->mact.user_rbuf_size);
	printf("#define\tACT_USER_ACT\t0x%x\n", &act->mact.user_act);
        printf("#define\tACT_PCB\t0x%x\n", &act->mact.pcb);
	printf("#define\tACT_RETURN_ROUTINE\t0x%x\n", &act->mact.return_routine);
	printf("#define\tACT_REENTRY_ROUTINE\t0x%x\n", &act->mact.reentry_routine);
	printf("#define\tACT_SYSCALL_PORT\t0x%x\n", &act->mact.reentry_routine);
	printf("#define\tACT_SPACE\t0x%x\n", &act->mact.space);
	printf("#define\tACT_PID\t0x%x\n", &act->mact.pid);
	printf("#define\tACT_SPARE\t0x%x\n", act->mact.spare);

	printf("\n/*\n * values from kern/task.h\n */\n");
        printf("#define\tTASK_EML_DISPATCH\t0x%x\n", &task->eml_dispatch);
	printf("#define\tTASK_SPACE\t0x%x\n", &task->itk_space);
	printf("#define\tTASK_SYSCALL_IPC_ENTRY\t0x%x\n", &task->syscall_ipc_entry);

	printf("\n/*\n * values from kern/syscall_emulation.h\n */\n");
	printf("#define\tDISP_MIN 0x%x\n", &disp->disp_min);
        printf("#define\tDISP_COUNT 0x%x\n", &disp->disp_count);
        printf("#define\tDISP_VECTOR 0x%x\n", &disp->disp_vector[0]);

	/* Restart Parameter Block, Processor Internal Memory */
	printf("#define\tRP_SP %d\n", &rp->rp_sp);
	printf("#define\tRP_FLAG %d\n", &rp->rp_flag);
	printf("#define\tRP_PIM %d\n", &rp->rp_pim);
	printf("#define\tRP_PIM_SP %d\n", &rp->rp_pim.pim_gr[30]);
	printf("#define\tRP_PIM_R4 %d\n", &rp->rp_pim.pim_gr[4]);
	printf("#define\tRPBSTRUCTSIZE %d\n", sizeof(struct rpb));
	printf("#define\tPIMSTRUCTSIZE %d\n", sizeof(struct pim));
	printf("#define\tMEM_PDC %d\n", &pzp->mem_pdc);
	printf("#define\tMEM_TOC %d\n", &pzp->ivec_toc);
	printf("#define\tMEM_TOC_LEN %d\n", &pzp->ivec_toclen);
	printf("#define\tHPA_IOCMD %d\n", &hpa->io_command);

	/*** Stuff for msg.s ***/

	/* The code depends on ipc_entry being 16, 20 or 24 bytes long.  */
	if (sizeof(struct ipc_entry) != 16 &&
	    sizeof(struct ipc_entry) != 20 &&
	    sizeof(struct ipc_entry) != 24)
		return 10;

	printf("#define\tIPC_ENTRY_SIZE\t%d\n", sizeof(struct ipc_entry));

        printf("#define\tMSGH_BITS\t0x%x\n", &msgh->msgh_bits);
        printf("#define\tMSGH_SIZE\t0x%x\n", &msgh->msgh_size);
        printf("#define\tMSGH_REMOTE_PORT\t0x%x\n", &msgh->msgh_remote_port);
        printf("#define\tMSGH_LOCAL_PORT\t0x%x\n", &msgh->msgh_local_port);
        printf("#define\tMSGH_SEQNO\t0x%x\n", &msgh->msgh_seqno);
        printf("#define\tMSGH_ID\t0x%x\n", &msgh->msgh_id);

        printf("#define\tSPACE_TABLE\t0x%x\n", &space->is_table);
        printf("#define\tSPACE_TABLE_SIZE\t0x%x\n", &space->is_table_size);

	printf("#define\tENTRY_BITS\t0x%x\n", &entry->ie_bits);
	printf("#define\tENTRY_OBJECT\t0x%x\n", &entry->ie_object);

	/* Fields in struct ipc_port */
	printf("#define\tIP_CUR_TARGET\t0x%x\n", &port->ip_cur_target);
	printf("#define\tIP_SEQNO\t0x%x\n", &port->ip_seqno);
	printf("#define\tIP_PSET\t0x%x\n", &port->ip_pset);
	printf("#define\tIP_RECEIVER_NAME\t0x%x\n", &port->ip_receiver_name);
	printf("#define\tIP_BITS\t0x%x\n", &port->ip_bits);

	/* Fields in struct ipc_target */
	printf("#define\tIPT_TYPE\t0x%x\n", &ipt->ipt_type);
	for (bit = 0; bit < 32; bit++)
		if ((1 << bit) == IPT_TYPE_MIGRATE_RPC)
			printf("#define\tIPT_TYPE_MIGRATE_RPC_BIT\t%d\n", 31-bit);
	printf("#define\tIPT_USER_ENTRYPOINT\t0x%x\n", &ipt->ipt_rpcinfo.entry_pc);
	printf("#define\tIPT_USER_DATA\t0x%x\n", &ipt->ipt_rpcinfo.entry_dp);
	printf("#define\tIPT_ACTS\t0x%x\n", &ipt->ipt_acts);
	printf("#define\tIPT_WAITING\t0x%x\n", &ipt->ipt_waiting);

	for (bit = 0; bit < 32; bit++) {
		if ((1 << bit) == MACH_PORT_TYPE_SEND)
			printf("#define\tMACH_PORT_TYPE_SEND_BIT\t%d\n", 31-bit);
		if ((1 << bit) == MACH_MSGH_BITS_COMPLEX)
			printf("#define\tMACH_MSGH_COMPLEX_BIT\t%d\n", 31-bit);
	}

        printf("#define\tMACH_MSG_OPTION_MIGR\t0x%x\n",
		MACH_SEND_MSG | MACH_RCV_MSG | MACH_MSG_OPTION_NONE);
	printf("#define\tMACH_MSGH_BITS_MIGR\t0x%x\n",
	       MACH_MSGH_BITS(MACH_MSG_TYPE_COPY_SEND, MACH_MSG_TYPE_MAKE_SEND_ONCE));

	printf("#define\tKERN_SUCCESS\t%d\n", KERN_SUCCESS);
	printf("#define\tKERN_ABORTED\t%d\n", KERN_ABORTED);
	return 0;
}
