/* 
 * Copyright (c) 1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 *	Utah $Hdr: medusa.c 1.6 94/12/14$
 */

/*
 * NOTES: The driver does not pad packets to ETHERMIN. Should it?
 */

#include "medusa.h"
#if NMEDUSA > 0

#include <sys/types.h>
#include <machine/asp.h>
#include <machine/psl.h>
#ifdef	MACH_KERNEL
#include <machine/machspl.h>
#include <mach/machine/vm_types.h>
#include <kern/time_out.h>
#include <device/device_types.h>
#include <device/errno.h>
#include <device/io_req.h>
#include <device/if_hdr.h>
#include <device/if_ether.h>
#include <device/net_status.h>
#include <device/net_io.h>
#include <mach_glue.h>
#include <sys/time.h>
#include <device.h>
#include <medreg.h>
#include <grf_stireg.h>
#else
#include <machine/spl.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/time.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/buf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/syslog.h>
#include <sys/proc.h>
#include <sys/user.h>
#include <tty.h>

#include <net/if.h>
#include <net/netisr.h>
#include <net/route.h>

#ifdef	INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#endif

#include <hpdev/device.h>
#include <hpsgc/medreg.h>
#include <hpsgc/grf_stireg.h>
#endif

/*
 * This is the header that goes onto the front of each FDDI packet.
 * Much of the info is constant, so we put a template in the softc
 * structure to save time. The transmit/receive routines convert
 * between the ether_header structure and this template.
 */
struct fddihdr {
	unsigned long 	magic;		/* FDDI preamble */
	char  		dst_mac[6];	/* Destination MAC */
	char  		src_mac[6];	/* Source MAC */
	unsigned char 	dsap;		/* Dst. Service Access Protocol */
	unsigned char 	ssap;		/* Src. Service Access Protocol */
	unsigned char 	control;
	unsigned char 	protocol[3];	/* Protocol ID/Origin code */
	unsigned short 	ethertype;	/* Ethertype (IP,ARP,etc.) */
};

#ifdef	MACH_KERNEL
#define FDDIMTU	(NET_RCV_MAX - sizeof(struct fddihdr) - CRC_SIZE_ALGN)
#else
#define FDDIMTU	(MED_BUFSIZE - sizeof(struct fddihdr) - CRC_SIZE_ALGN)
#endif

struct med_softc {
	struct fddihdr	template;	/* First, to ensure alignment */
	
#ifdef	MACH_KERNEL
	struct	ifnet	ds_if;		/* generic interface header */
	u_char	ds_addr[6];		/* Ethernet hardware address */
#else
	struct		arpcom	med_ac;
#define	ds_if	med_ac.ac_if
#define	ds_addr	med_ac.ac_enaddr
#endif	
	char 		*base;		/* I/O bus address */
	int		unit;
	int		flags;
#define MED_RUNNING  0x1
#define MED_SLEEPING 0x4
#define MED_OPEN     0x8
};

struct med_softc	med_softc[NMEDUSA];
int			medintr(), medinit(), medioctl(), medreset(),
			medstart();
void			med_vcopy();
int			meddebug = 0;

/*
 * grfconfig() is called way too early during bootup, so calling the 
 * medusa initialization routines will not work properly. Therefore,
 * duplicate some of the grfconfig() code here.
 */
medconfig()
{
	struct hp_device hd;
	struct modtab *mptr, *getmod();
	int noinit = 0;
	int unit = 0;

	while ((mptr = getmod(IODC_TP_FIO, SVMOD_FIO_SGC, noinit)) != 0) {
		hd.hp_unit = unit;
		hd.hp_addr = (char *)mptr->m_hpa;
		if (medprobe(&hd)) {
			medattach(&hd);
			unit++;
		}
		noinit = 1;
	}
}

/*
 * Get STI ROM device ID and indicate (via return code) if it is a Medusa.
 */
medprobe(dev)
	struct hp_device *dev;
{
	caddr_t		addr = dev->hp_addr;
	int		unit = dev->hp_unit;
	int		_om_, id;

	if (unit < 0 || unit > NMEDUSA) {
		printf("fddi%d: board out of range [0..%d]\n", unit, NMEDUSA);
		return(0);
	}

	_om_ = rsm(PSW_P);
	id = STI_ID_HI(STI_TYPE_BWGRF, addr);
	if (_om_ & PSW_P) ssmv(_om_);
	if (id == MED_STI_ID) {
		printf("medprobe: Medusa FDDI card at 0x%x\n", addr);
		return(1);
	}
	return(0);
}

/*
 * Attach the card to the system.
 */
void
medattach(dev)
	struct hp_device	*dev;
{
	int			i;
	struct med_softc	*mp = &med_softc[dev->hp_unit];
	unsigned long		*rom;
	struct ifnet		*ifp;

	mp->base = dev->hp_addr;
	mp->unit = dev->hp_unit;
	rom = (unsigned long *) mp->base;

	for (i = 0; i < MED_MAC_LEN; i++) {
		mp->ds_addr[i] = (rom[(MED_MAC_OFFSET >> 2) + i]) & 0xFF;
	}
	printf("medattach: Medusa FDDI id [%s]\n", ether_sprintf(mp->ds_addr));

	/*
	 * Initialize header template.
	 */
	mp->template.magic   = MED_PKT_HDR;
	mp->template.dsap    = 170;
	mp->template.ssap    = 170;
	mp->template.control = 3;
	bcopy(mp->ds_addr, mp->template.src_mac, sizeof(mp->ds_addr));

	ifp = &(mp->ds_if);
	ifp->if_unit = dev->hp_unit;
	ifp->if_flags = IFF_BROADCAST;
#ifdef IFF_SIMPLEX
	ifp->if_flags |= IFF_SIMPLEX;
#endif
#ifdef	MACH_KERNEL
	ifp->if_header_size = sizeof(struct ether_header);
	ifp->if_header_format = HDR_ETHERNET;
	ifp->if_address_size = 6;
	ifp->if_address = (char *)&mp->ds_addr[0];
	if_init_queues(ifp);
#else
	ifp->if_name = "fddi";
	ifp->if_init = medinit;
	ifp->if_output = ether_output;
	ifp->if_start = medstart;
	ifp->if_ioctl = medioctl;
	ifp->if_reset = medreset;
	ifp->if_next = NULL;
	if_attach(ifp); 
#endif
	ifp->if_mtu = FDDIMTU;	/* ifattach sets this to ETHERMTU */
}

/*
 * This routine is in part an entry point for the "if" code. 
 */
medreset(unit)
	int	unit;
{
	med_softc[unit].ds_if.if_flags &= ~IFF_RUNNING;
	med_softc[unit].flags &= ~MED_RUNNING;
	medinit(unit);
	return(1);
}

/*
 * Another routine that interfaces the "if" layer to this driver.  
 * Simply resets the structures that are used by "upper layers".  
 */	
medinit(unit)
	int	unit;
{
	struct	ifnet	*ifp;
	int		s;

	ifp = &(med_softc[unit].ds_if);
#ifndef	MACH_KERNEL
	if (ifp->if_addrlist == (struct ifaddr *)0)
		return(0);
#endif
	s = splimp();
	medhwinit(unit);
	med_softc[unit].ds_if.if_flags |= IFF_RUNNING;
#ifdef MACH_KERNEL
	medstart(unit);
#else
	medstart(ifp);
#endif
	splx(s);
}

#ifndef	MACH_KERNEL
medioctl(ifp, cmd, data)
	struct ifnet	*ifp;
	int		cmd;
	caddr_t		data;
{
	struct ifaddr		*ifa = (struct ifaddr *)data;
	int			unit = ifp->if_unit;
	struct med_softc 	*mp = &med_softc[unit];
	int			s, error = 0;

	s = splimp();
	switch (cmd) {

	case SIOCSIFADDR:
		ifp->if_flags |= IFF_UP;
		medinit(unit);
		switch (ifa->ifa_addr->sa_family) {

		case AF_INET:
			((struct arpcom *)ifp)->ac_ipaddr =
				IA_SIN(ifa)->sin_addr;
			arpwhohas((struct arpcom *)ifp, &IA_SIN(ifa)->sin_addr);
			break;
		}
		break;

	case SIOCSIFFLAGS:
		if ((ifp->if_flags & IFF_UP) == 0 && mp->flags & MED_RUNNING) {
			printf("fddi%d ioctl(): board is not running\n", unit);
			mp->flags &= MED_RUNNING;
		} else if (ifp->if_flags & IFF_UP &&
			   (mp->flags & MED_RUNNING) == 0) {
			medinit(unit);
		}
		break;

	default:
		error = EINVAL;
	}
	splx(s);
	return (error);
}
#endif	/* MACH_KERNEL */

/*
 * Try to output each packet on the send queue.
 */
#ifdef	MACH_KERNEL
medstart(unit)
	int			unit;
#else
medstart(ifp)
	struct ifnet 		*ifp;
#endif
{
#ifdef	MACH_KERNEL
	io_req_t		m;
	struct med_softc	*mp = &med_softc[unit];
	struct	ifnet		*ifp = &(mp->ds_if);
#else
	struct mbuf		*m;
	int			unit = ifp->if_unit;
	struct med_softc	*mp = &med_softc[unit];
#endif

	if ((mp->ds_if.if_flags & IFF_RUNNING) == 0)
		return(0);

	do {
		IF_DEQUEUE(&mp->ds_if.if_snd, m);
		if (m == 0)
			return(0);
		medxmit(unit, m);
	} while (1);
	return(0);
}

medxmit(unit, m)
	int	unit;
#ifdef	MACH_KERNEL
	io_req_t	m;
#else
	struct	mbuf	*m;
#endif
{
	struct med_softc	*mp = &med_softc[unit];
	unsigned long 		med_card_base = (unsigned long) mp->base;
#ifdef	MACH_KERNEL
	struct ether_header	*eh_p = (struct ether_header *)m->io_data;
#else
	struct	mbuf		*tm_p = m;
	struct ether_header	*eh_p = mtod(m, struct ether_header *);
	u_char			*mb_p = mtod(m, u_char *) +
						sizeof(struct ether_header);
	u_short			count = m->m_len - sizeof(struct ether_header);
#endif
	u_short			size = 0;
	u_char 			*start, *sp;
	struct mbuf		*copy_m;

	/*
	 * The Medusa does not see packets that are addressed to itself. This
	 * includes broadcast packets. The first case should never happen.
	 * The second case should be handled by higher level code when
	 * IFF_SIMPLEX is defined.
	 */
#if 0
	if (bcmp((caddr_t)mp->template.src_mac,
	    (caddr_t)eh_p->ether_dhost, sizeof(mp->template.src_mac)) == 0) {
		panic("medxmit(): Why am I here?");
	}
#ifndef	IFF_SIMPLEX
	if (bcmp((caddr_t)etherbroadcastaddr,
	    (caddr_t)eh_p->ether_dhost, sizeof(etherbroadcastaddr)) == 0) {
		bcopy((caddr_t)mp->template.src_mac, (caddr_t)eh_p->ether_shost,
		      sizeof(eh_p->ether_shost));
		copy_m = m_copy(m, 0, M_COPYALL);
		m_adj(copy_m, sizeof(struct ether_header));
		ether_input(&mp->ds_if, eh_p, copy_m);
	}
#endif
#endif
	/*
	 * Grab a transmit buffer.
	 */
	sp = start = (u_char *) in_word(TX_FREE_FIFO);
	if (((unsigned long) start & 0x80000000L) == 0) {
		printf("send: buffer is empty\n");
		return(0);
	}
	
	/*
	 * Convert the Ether header into an FDDI header and copy it out to
	 * the transmit buffer.
	 */
	bcopy(eh_p->ether_dhost, mp->template.dst_mac,
	      sizeof(eh_p->ether_dhost));
	mp->template.ethertype = eh_p->ether_type;
	size = sizeof(mp->template);
	med_copy(&mp->template, sp, size);
	sp += size;

#ifdef	MACH_KERNEL
	/*
	 * Copy the entire buffer out to the transmit buffer.
	 */
	med_copy(eh_p + 1, sp, m->io_count - sizeof(struct ether_header));
	size += m->io_count - sizeof(struct ether_header);
#else
	/*
	 * Now loop through the mbufs, copying data into the transmit buffer.
	 */
	do {
		if (count) {
			med_copy(mb_p, sp, count);
			size += count;
			sp   += count;

			if ((tm_p = tm_p->m_next) == (struct mbuf *)0)
				break;
		}
		else if ((tm_p = tm_p->m_next) == (struct mbuf *)0)
			break;
		count = tm_p->m_len;
		mb_p = mtod(tm_p, u_char *);
	} while (1);
	
	mp->ds_if.if_opackets++;
#endif	/* MACH_KERNEL */
	
	if (meddebug) {
		printf("medxmit: 0x%x %d\n", start, size);
		printfddi(start, size);
	}

	/*
	 * Send the packet on its way.
	 */
	out_word (TX_READY_FIFO, start + size);
	
#ifdef	MACH_KERNEL
	iodone(m);
#else
	m_freem(m);
#endif
}

/*
 * Hardware interrupt routine.
 */
unsigned long	fdditimer = 0;

int
medintr(unit)
	int	unit;
{
	struct med_softc	*mp = &med_softc[unit];
	unsigned long 		med_card_base = (unsigned long) mp->base;
	unsigned long 		event = in_word (INTERRUPT_EVENT);
	char 			*buf;
	int			len;

	if (meddebug)
		printf("medintr: event = 0x%x\n", event);

#if 0
	if (fdditimer)
		printf("medintr: time = %d\n", mfctl(16) - fdditimer);
#endif

#ifndef	MACH_KERNEL
	/*
	 * Someone has the card open, and wants the packets for themselves.
	 */
	if (mp->flags & MED_OPEN) {
		out_word (INTERRUPT_EVENT, event);
		if (mp->flags & MED_SLEEPING) {
			mp->flags &= ~MED_SLEEPING;
			wakeup((caddr_t)mp);
		}
		return;
	}
#endif

	do {
		while ((unsigned long)
		       (buf = (char *)in_word (RX_READY_FIFO)) & 0x80000000) {

			len = (unsigned long)buf & ~(-MED_BUFSIZE);
			buf = (char *)((unsigned long)buf &  (-MED_BUFSIZE));

			if (meddebug) {
				printf("medintr: buf 0x%x, len %d\n", buf, len); 
				printfddi(buf, len);
			}

			if (len > sizeof(struct fddihdr)) {
				len -= CRC_SIZE;	/* Trailing CRC */
				medrecv(unit, buf, len);
			}
			out_word (RX_FREE_FIFO, buf);
		}
		event = in_word (INTERRUPT_EVENT);
		out_word (INTERRUPT_EVENT, event);

	} while ((in_word (STATUS_BUFF_FIFO) & 0x4));
}

medrecv(unit, buf, len)
	int unit;
	char *buf;
	int len;
{
	struct med_softc 	*mp = &med_softc[unit];
	unsigned long 		med_card_base = (unsigned long) mp->base;	
	struct fddihdr		fh, *fhp = &fh;
#ifndef	MACH_KERNEL
	struct ether_header	eh, *ehp = &eh;
    	struct mbuf 		*m, *m_devget();
#else
	ipc_kmsg_t 		kmsg;
	struct packet_header	*pkt;
	struct ether_header	*ehp;

	if ((kmsg = net_kmsg_get()) == IKM_NULL) {
		mp->ds_if.if_rcvdrops++;
		return;
	}
	ehp = (struct ether_header *) (&net_kmsg(kmsg)->header[0]);
	pkt = (struct packet_header *)(&net_kmsg(kmsg)->packet[0]);
#endif
	
	/*
	 * Pull the the FDDI header off the card and construct an
	 * ethernet style header to pass up to higher level code.
	 */
	med_copy(buf, fhp, sizeof(struct fddihdr));
	bcopy(fhp->src_mac, ehp->ether_shost, sizeof(ehp->ether_shost));
	bcopy(fhp->dst_mac, ehp->ether_dhost, sizeof(ehp->ether_dhost));
	ehp->ether_type = fhp->ethertype;
	len -= sizeof(struct fddihdr);
	buf += sizeof(struct fddihdr);

#ifdef	MACH_KERNEL
	pkt->type   = ehp->ether_type;
	pkt->length = len + sizeof(struct packet_header);
	out_word(VENOM_SUP_CONTROL, VENOM_CONTROL);
	med_vcopy(buf, (pkt + 1), len);
	net_packet(&(mp->ds_if), kmsg, pkt->length, ethernet_priority(kmsg));
#else
	/*
	 * Pull the rest of the packet off the card using mbuf copy code.
	 */
	out_word(VENOM_SUP_CONTROL, VENOM_CONTROL);
	m = m_devget((char *)buf, len, 0, &mp->ds_if, med_vcopy);
	if (m == 0)
		return;

	mp->ds_if.if_ipackets++;
	ether_input(&mp->ds_if, ehp, m);
#endif
}

/*
 * Perform necessary device initializations for the card.
 */
medhwinit(unit)
	int			unit;
{
	struct med_softc	*mp = &med_softc[unit];
	unsigned long 		med_card_base = (unsigned long) mp->base;
	unsigned long 		buf;
	int			i;

	/*
	 * Send all zeros to do the reset
	 */
	out_word (RESET_FIFO, 0x00000000);
	out_word (RESET_MISC, 0x00000000);
	out_word (RESET_FDDI, 0x00000000);
	
	/*
	 * Do initializations of other card registers.
	 */
	out_word (CHECKSUM_UNIT, INIT_CHECKSUM_UNIT);
	out_word (TEST_FCG, INIT_TEST_FCG);
	out_word (TEST_ELM, INIT_TEST_ELM);
	out_word (TEST_MAC, INIT_TEST_MAC);
	out_word (INTERRUPT_MASK, INIT_INTERRUPT_MASK);
	out_word (INTERRUPT_EVENT, (~INIT_INTERRUPT_EVENT & 0xff));
	out_word (MISC_OUTPUT, INIT_MISC_OUTPUT);

	/*
	 * Send all ones to release the reset
	 */
	out_word (RESET_FIFO, 0xFFFFFFFF);
	out_word (RESET_MISC, 0xFFFFFFFF);
	out_word (RESET_FDDI, 0xFFFFFFFF);

	/*
	 * ?
	 */
	out_word (RESET_FIFO, 0x00000000);
	out_word (RESET_FIFO, 0xffffffff);

	/*
	 * Initialize the transmit and receive onboard buffers
	 */
	buf = (CARD_MEMORY_BASE) + ((MED_NUM_BUFS) * (MED_BUFSIZE));
	for (i = 0; i < MED_NUM_RX_BUFS; i++) {
		buf -= MED_BUFSIZE;
		out_word(RX_FREE_FIFO, buf);
	}
	for (i = 0; i < MED_NUM_TX_BUFS; i++) {
		buf -= MED_BUFSIZE;
		out_word(TX_READY_FIFO, buf);
	}

	out_word(ELM_CNTRL_B, ELM_PCM_CNTRL_STOP);
	while ((in_word(ELM_STATUS_B) & ELM_PCM_STATE) != ELM_PCM_OFF);

	/*
	 * Load configuration values into the ELM registers
 	 */
	out_word(ELM_INTR_MASK, ELM_INTR_MASK_CONFIG);
	out_word(ELM_LE_THRESHOLD, ELM_LE_THRESHOLD_CONFIG);
	out_word(ELM_A_MAX, ELM_A_MAX_CONFIG);
	out_word(ELM_LS_MAX, ELM_LS_MAX_CONFIG);
	out_word(ELM_TB_MIN, ELM_TB_MIN_CONFIG);
	out_word(ELM_T_OUT, ELM_T_OUT_CONFIG);
	out_word(ELM_LC_SHORT, ELM_LC_SHORT_CONFIG);
	out_word(ELM_T_SCRUB, ELM_T_SCRUB_CONFIG);
	out_word(ELM_NS_MAX, ELM_NS_MAX_CONFIG);
	out_word(ELM_CNTRL_A, ELM_CNTRL_A_CONFIG);
	out_word(ELM_CNTRL_B, ELM_PC_MAINT|0x600);

	out_word (MAC_CNTRL_A, ~MAC_ON);
	out_word(MAC_INTR_MASK_A, 0x00000000);
	out_word(MAC_INTR_MASK_B, 0x00000000);
	out_word(MAC_MSA, 0);
	out_word(MAC_MLA_A, mp->ds_addr[4] << 8 | mp->ds_addr[5]);
	out_word(MAC_MLA_B, mp->ds_addr[2] << 8 | mp->ds_addr[3]);
	out_word(MAC_MLA_C, mp->ds_addr[0] << 8 | mp->ds_addr[1]);
	out_word(MAC_T_REQ, MAC_T_REQ_CONFIG);
	out_word(MAC_T_MAX, MAC_T_MAX_CONFIG);
	out_word(MAC_CNTRL_A, MAC_CNTRL_A_CONFIG);
	out_word(MAC_CNTRL_B, MAC_CNTRL_B_CONFIG);
	
	while (!(in_word(MAC_TX_STATUS) & MAC_RING_OPERATIONAL));

	out_word (INTERRUPT_EVENT, 0XFFFFFFFF);
	out_word (INTERRUPT_MASK,  0x00000020);  /* Allow receive interrupts */

	if (mp->base == 0XF8000000) {
		aspitab(INT_GRF2, SPLFDDI_2, medintr, mp->unit, 0);
		aspitab(INT_GRF1, SPLFDDI_1, medintr, mp->unit, 0);
	}
	mp->flags |= MED_RUNNING;
}

medopen(dev, flags)
	dev_t dev;
{
	int			unit = minor(dev);
	struct med_softc 	*mp = &med_softc[unit];
	unsigned long 		med_card_base = (unsigned long) mp->base;

	if (meddebug)
		printf("medopen: unit = %d\n", unit);

	if (unit < 0 || unit >= NMEDUSA || !mp->base)
		return (ENXIO);

#ifdef	MACH_KERNEL
	mp->ds_if.if_flags |= IFF_UP;
	medinit(unit);
#else
	if ((mp->flags & MED_RUNNING) == 0) {
		medhwinit(unit);
	}
	mp->flags |= MED_OPEN;
#endif
	return(0);
}

medclose(dev, flags)
	dev_t dev;
{
	int			unit = minor(dev);
	struct med_softc 	*mp = &med_softc[unit];

	if (meddebug)
		printf("medclose: unit = %d\n", unit);

	mp->flags &= ~MED_OPEN;
	return(0);
}

#ifdef	MACH_KERNEL
medoutput(dev, ior)
	dev_t	dev;
	io_req_t ior;
{
	int			unit = minor(dev);
	struct med_softc 	*mp = &med_softc[unit];

	if (meddebug)
		printf("medoutput: unit = %d, ior = 0x%x\n", unit, ior);

	if (unit < 0 || unit >= NMEDUSA || !mp->base)
		return (ENXIO);

	return (net_write(&(mp->ds_if), medstart, ior));
}

medsetinput(dev, receive_port, priority, filter, filter_count)
	dev_t		dev;
	mach_port_t	receive_port;
	int		priority;
	filter_t	filter[];
	unsigned int	filter_count;
{
	int			unit = minor(dev);
	struct med_softc 	*mp = &med_softc[unit];

	if (unit < 0 || unit >= NMEDUSA || !mp->base)
		return (ENXIO);

	return (net_set_filter(&(mp->ds_if), receive_port, priority,
			       filter, filter_count));
}

medgetstat(dev, flavor, status, count)
	dev_t	dev;
	int	flavor;
	dev_status_t	status;		/* pointer to OUT array */
	unsigned int	*count;		/* out */
{
	int			unit = minor(dev);
	struct med_softc 	*mp = &med_softc[unit];

	if (unit < 0 || unit >= NMEDUSA || !mp->base)
		return (ENXIO);

	return (net_getstat(&(mp->ds_if), flavor, status, count));
}

medsetstat(dev, flavor, status, count)
	dev_t	dev;
	int	flavor;
	dev_status_t	status;
	unsigned int	count;
{
	int			unit = minor(dev);
	struct med_softc 	*mp = &med_softc[unit];

	if (unit < 0 || unit >= NMEDUSA || !mp->base)
		return (ENXIO);

	printf("medsetstat: 0x%x 0x%x\n", flavor,
	       ((struct net_status *)status)->flags);

	return (D_SUCCESS);
}
#else
/*
 * Simple write interface to the FDDI card. NOTE: The user buffer *must*
 * be word aligned.
 */
medwrite(dev, uio)
	dev_t		dev;
	struct uio	*uio;
{
	int			size, unit = minor(dev);
	struct med_softc 	*mp = &med_softc[unit];
	unsigned long 		med_card_base = (unsigned long) mp->base;
	u_char 			*start;

	if (unit < 0 || unit >= NMEDUSA || !mp->base)
		return (ENXIO);

	if ((unsigned long) uio->uio_iov->iov_base & 0x3) {
		printf("medwrite: unaligned 0x%x\n", uio->uio_iov->iov_base);
		return(EIO);
	}

	/*
	 * Grab a transmit buffer.
	 */
#if 1
	while (((unsigned long)
	       (start = (u_char *) in_word (TX_FREE_FIFO)) & 0x80000000) == 0)
		;
#else
	start = (u_char *) in_word(TX_FREE_FIFO);
	if (((unsigned long) start & 0x80000000L) == 0) {
		printf("send: buffer is empty\n");
		return(EIO);
	}
#endif
	
	size = uio->uio_resid;
	meduiomove(start, size, uio);

	if (meddebug)
		printfddi(start, size);
	
	/*
	 * Send the packet on its way.
	 */
	out_word (TX_READY_FIFO, start + size);
	
	return(0);
}

/*
 * Simple read interface to the FDDI card. NOTE: The user buffer *must*
 * be word aligned.
 */
medread(dev, uio)
	dev_t		dev;
	struct uio	*uio;
{
	int			unit = minor(dev);
	struct med_softc 	*mp = &med_softc[unit];
	unsigned long 		med_card_base = (unsigned long) mp->base;
	char 			*buf;
	int			len;
	int			done = 0;

	if (unit < 0 || unit >= NMEDUSA || !mp->base)
		return (ENXIO);

	if ((unsigned long) uio->uio_iov->iov_base & 0x3) {
		printf("medwrite: unaligned 0x%x\n", uio->uio_iov->iov_base);
		return(EIO);
	}

	do {
		while ((in_word (STATUS_BUFF_FIFO) & 0x4) == 0) {
			mp->flags |= MED_SLEEPING;
			sleep((caddr_t)mp, TTIPRI);
		}

		if ((unsigned long)
		    (buf = (char *)in_word (RX_READY_FIFO)) & 0x80000000) {

			len = (unsigned long)buf & ~(-MED_BUFSIZE);
			buf = (char *)((unsigned long)buf &  (-MED_BUFSIZE));

			if (meddebug) {
				printf("medintr: buf 0x%x, len %d\n", buf, len); 
				printfddi(buf, len);
			}

			if (len <= sizeof(struct fddihdr)) {
				out_word (RX_FREE_FIFO, buf);
				continue;
			}
			
			len -= CRC_SIZE;	/* Trailing CRC */
			meduiomove(buf, len, uio);
			out_word (RX_FREE_FIFO, buf);
			done = 1;
		}
	} while (!done);

	return(0);
}

meduiomove(cp, n, uio)
	register caddr_t cp;
	register int n;
	register struct uio *uio;
{
	register struct iovec *iov;
	u_int cnt;

	while (n > 0 && uio->uio_resid) {
		iov = uio->uio_iov;
		cnt = iov->iov_len;
		if (cnt == 0) {
			uio->uio_iov++;
			uio->uio_iovcnt--;
			continue;
		}
		if (cnt > n)
			cnt = n;
		switch (uio->uio_segflg) {

		case UIO_USERSPACE:
		case UIO_USERISPACE:
			if (uio->uio_rw == UIO_READ) {
				copyout(cp, iov->iov_base, cnt);
			}
			else
				copyin(iov->iov_base, cp, cnt);
			break;

		default:
			panic("meduiomove\n");
		}
		iov->iov_base += cnt;
		iov->iov_len -= cnt;
		uio->uio_resid -= cnt;
		uio->uio_offset += cnt;
		cp += cnt;
		n -= cnt;
	}
	return (0);
}

printmbufs(m)
	struct	mbuf	*m;
{
	struct	mbuf		*tm_p = m;
	u_char			*mb_p = mtod(m, u_char *);
	u_short			count = m->m_len;
	int			i;

	printf("pmbuf: 0x%x %d\n", m, count);

	do {
		if (count) {
			for (i = 0; i < count; i++)
				printf("%x ", (u_char) *mb_p++);
			printf("\n");

			if ((tm_p = tm_p->m_next) == (struct mbuf *)0)
				break;
		}
		else if ((tm_p = tm_p->m_next) == (struct mbuf *)0)
			break;
		count = tm_p->m_len;
		mb_p = mtod(tm_p, u_char *);
	} while (1);
}
#endif /* MACH_KERNEL */
printfddi(buf, len)
	u_long	*buf;
	int	len;
{
	int i;

	if (len > 128)
		len = 128;

	for (i = 0; i < len/4; i++)
		printf("%x ", *buf++);

	printf("\n");
}
#endif /* NMEDUSA */
