/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

#include "scsiha.h"
#define NSCSI	NSCSIHA
#if NSCSI > 0

/*
 * Controller independent interfaces to SCSI.
 */

#ifdef	MACH_KERNEL
#include <device/errno.h>
#include <kern/thread.h>
#include <mach_glue.h>
#include <machine/machspl.h>
#include <scsi.h>
#include <scsi_common.h>
#else
#include <sys/param.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#include <hpsgc/bsd_glue.h>
#include <machine/spl.h>
#include <hpsgc/scsi.h>
#include <hpsgc/scsi_common.h>
#endif

struct scsi_cmd nullcmd;

#define SCSIBUS(dev)	((minor(dev)>>6)&0x03)
#define SCSITARGET(dev)	((minor(dev)>>3)&0x07)
#define SCSILUN(dev)	((minor(dev))&0x07)

#define MKSCSIDEV(sd, d) {\
        sd.bus = SCSIBUS(d),\
        sd.target = SCSITARGET(d),\
        sd.lun = SCSILUN(d);\
}

struct {
	struct scsi_request_sense	*sense;
	int				open;
} scsi_data[NSCSI];

unsigned long scsi_count = 0;

#ifdef DBG
int scsidebug = 0;
unsigned long scsi_trace_index = 0;
unsigned long scsi_trace_mask = -1;		/* default to tracing all targets */
#define TRACE_SIZE	128
struct strace {
	unsigned long	st_count;
	vm_map_t	st_map;
	vm_offset_t	st_addr;
	vm_size_t	st_len;
	struct scsi_device st_sd;
	union scsi_cdb	st_cdb;
} scsi_trace[TRACE_SIZE];

struct {
	union scsi_cdb	sl_cdb;
	unsigned long	sl_id;
	vm_offset_t	sl_di;
} scsi_last[NSCSI][SCSI_NTARGETS];

/* Must be used at splbio() */
#define TRACE_IT(cmd) {\
	if (scsi_trace_mask & (1<<(cmd)->device.target)) {\
                struct strace *st; \
                st = &scsi_trace[scsi_trace_index++]; \
                if (scsi_trace_index >= TRACE_SIZE) \
                        scsi_trace_index = 0; \
                st->st_count = (cmd)->id; \
	        st->st_map = (cmd)->data_map; \
	        st->st_addr = (cmd)->data_addr; \
	        st->st_len = (cmd)->data_len; \
	        st->st_cdb = (cmd)->cdb; \
	        st->st_sd = (cmd)->device; \
	} \
	scsi_last[(cmd)->device.bus][(cmd)->device.target].sl_cdb = (cmd)->cdb; \
	scsi_last[(cmd)->device.bus][(cmd)->device.target].sl_id = (cmd)->id; \
	scsi_last[(cmd)->device.bus][(cmd)->device.target].sl_di = (cmd)->driver_info; \
}
#else
#define TRACE_IT(cmd)
#endif

char    *scsi_sense_key_text[] = {
	/* 00   */      "no sense",
	/* 01   */      "recovered error",
	/* 02   */      "unit not ready",
	/* 03   */      "medium error",
	/* 04   */      "hardware error",
	/* 05   */      "illegal request",
	/* 06   */      "unit attention",
	/* 07   */      "data protect",
	/* 08   */      "blank check",
	/* 09   */      "vendor unique command",
	/* 0A   */      "copy aborted",
	/* 0B   */      "aborted command",
	/* 0C   */      "equal",
	/* 0D   */      "volume overflow",
	/* 0E   */      "miscompare",
	/* 0F   */      "misc sense error",

/* software invented sense error codes for more specific error indications */

	/* 10   */      "no sense data",          /* need this for non-zero return */
	/* 11   */      "no tape present",        /* tell user he forgot to insert a tape */
	/* 12   */      "not yet ready",          /* drive is coming online (loading or testing) */
	/* 13   */      "record length mismatch", /* requested record length did not match recorded record length */
};

int scsi_sense_key_count = (sizeof(scsi_sense_key_text)/sizeof(scsi_sense_key_text[0]));

struct scsi_controller *scsi_controller[NSCSI];

#ifdef	MACH_KERNEL
zone_t scsi_command_zone = ZONE_NULL;
#else
u_long scsi_command_size = 0;
#endif

scsi_init()
{
#ifdef	MACH_KERNEL
	vm_offset_t cram;
#endif
	vm_size_t size;
	static scsi_inited = 0;
	extern unsigned long dcache_line_size;

	if (scsi_inited) return;
	scsi_inited = 1;

	size = MAX(sizeof(struct scsi_request_sense), sizeof(struct scsi_cmd));
	size = (size + (dcache_line_size-1)) & ~(dcache_line_size-1);

#ifdef	MACH_KERNEL
	scsi_command_zone = zinit(size, 64 * NBPG, NBPG, FALSE,
				  "scsi commands/sense");
	if (kmem_alloc(kernel_map, &cram, 16 * NBPG) != KERN_SUCCESS)
		panic("scsi_init: no memory for command zone");
	zcram(scsi_command_zone, cram, 16 * NBPG);
#else
	scsi_command_size = size;
#endif
}

int
scsi_add(bus, scp)
int bus;
struct scsi_controller *scp;
{
	if (bus > NSCSI) {
		return(EINVAL);
	}
	else if (scsi_controller[bus]) {
		panic("scsi_add: attempt to replace bus!");
	}

	scsi_init();
	scsi_controller[bus] = scp;
	return(0);
}

int
scsi_del(bus)
int bus;
{
	if (bus > NSCSI) {
		return(EINVAL);
	}
	else if (!scsi_controller[bus]) {
		panic("scsi_del: attempt to delete non-existent bus!");
	}

	scsi_controller[bus] = (struct scsi_controller *)0;
	return(0);
}

int
scsi_command(action, cmd)
int action;
struct scsi_cmd *cmd;
{
	struct scsi_controller	*sc;
	int x;

	if (!cmd)
		panic("scsi_command: no cmd");

	if ((cmd->device.bus >= NSCSI) || !(sc = scsi_controller[cmd->device.bus]))
		panic("scsi_command: no controller");

#ifdef DBG
	if (scsidebug) {
		if ((cmd->flags & (SCSI_DATA_IN|SCSI_DATA_OUT)) == (SCSI_DATA_IN|SCSI_DATA_OUT) ||
		   (cmd->flags & (SCSI_WAIT|SCSI_CALL)) == (SCSI_WAIT|SCSI_CALL))
			panic("scsi_command: conficting command flags\n");
     	}
#endif
	x = splbio();
	cmd->id = scsi_count++;

	switch (action) {
	case SCSI_QUEUE_TAIL:
	case SCSI_QUEUE_HEAD:
	case SCSI_QUEUE_ANYWHERE:
		/* check command length validity */
		switch (cmd->cdb.code.group) {

		case 0: cmd->cdb_len = 6;	/* 6 command bytes */
			break;

		case 1:
		case 2: cmd->cdb_len = 10;	/* 10 command bytes */
			break;

		case 3:                         /* reserved */
		case 4:				/* reserved */
		case 6:				/* vendor specific */
		case 7:				/* vendor specific */
			if (cmd->cdb_len == 0)
	                	panic("scsi_command: invalid command length\n");
			break;

		case 5:				/* 12 bytes */
			cmd->cdb_len = 12;
			break;
		}

		/* no linked commands in HPCS */
		if (cmd->cdb.command[cmd->cdb_len - 1] & 0x01)
			panic("scsi_command: linked commands not supported\n");

                TRACE_IT(cmd);
		break;

	case SCSI_FLUSH_ALL:
	case SCSI_FLUSH_COMMAND:
		break;
	default:
		panic("scsi_command: unrecognized action\n");
		break;
	}

	(*sc->command)(sc, cmd->device.target, action, cmd);
	
	if (action == SCSI_QUEUE_TAIL || 
	    action == SCSI_QUEUE_HEAD || 
	    action == SCSI_QUEUE_ANYWHERE) {
		cmd->flags |= SCSI_INPROGRESS;
		if (cmd->flags & SCSI_WAIT) {
#ifdef	MACH_KERNEL
			/*
			 * XXX if threads haven't been initialized
			 * (i.e. scsi commands issued at autoconfig time)
			 * we can't sleep.
			 */
			extern zone_t thread_zone;

			if (thread_zone == 0)
				while (cmd->flags & SCSI_INPROGRESS) {
					spl0();
					splbio();
				}
			else
#endif
			while (cmd->flags & SCSI_INPROGRESS)
				sleep((caddr_t)cmd, PRIBIO);
		}
	}
	splx(x);

	return(0);
}

void
scsi_error(dev, str, cmd)
dev_t dev;
char *str;
struct scsi_cmd *cmd;
{
#ifdef  DBG
        if (scsidebug)
                printf("scsi_error(0x%x, 0x%x, 0x%x)\n", dev, str, cmd);
#endif
	printf("scsi_error: %s bus=%d, target=%d lun=%d: ", str, cmd->device.bus, cmd->device.target, cmd->device.lun);
}

void
scsi_flush(sd)
struct scsi_device sd;
{
	struct scsi_controller	*sc;
	struct scsi_cmd cmd;
	int x;

	if ((sd.bus >= NSCSI) || !(sc = scsi_controller[sd.bus]))
		panic("scsi_flush: no controller");

	cmd.device = sd;

	x = splbio();
	cmd.id = scsi_count++;
	(*sc->command)(sc, sd.target, SCSI_FLUSH_ALL, &cmd);
	splx(x);
}

unsigned long
(*scsi_minphys(sd))()
struct scsi_device sd;
{
	struct scsi_controller	*sc;

	if ((sd.bus >= NSCSI) || !(sc = scsi_controller[sd.bus]))
		panic("scsi_minphys: no controller");

	return(sc->minphys);
}

int
scsi_open(dev, flag)
dev_t dev;
int   flag;
{
	struct scsi_controller	*sc;
	struct scsi_device sd;

	MKSCSIDEV(sd,dev);

	if ((sd.bus >= NSCSI) || !(sc = scsi_controller[sd.bus]))
		return(ENXIO);

	if (scsi_data[sd.bus].open)
		return(EBUSY);

	scsi_data[sd.bus].open = TRUE;
	return(0);
}

int
scsi_close(dev)
dev_t dev;
{
	struct scsi_controller	*sc;
	struct scsi_device sd;

	MKSCSIDEV(sd,dev);

	if ((sd.bus >= NSCSI) || !(sc = scsi_controller[sd.bus]))
		return(ENXIO);

	if (!scsi_data[sd.bus].open)
		return(EIO);

	scsi_data[sd.bus].open = FALSE;
	return(0);
}

#ifdef	MACH_KERNEL
/* XXX what do we need here to handle formatting functions? */
#else
int
scsi_ioctl(dev, op, arg, flag)
dev_t dev;
int op;
caddr_t arg;
int flag;
{
	struct scsi_cmd *cmd;
	struct scsi_controller *sc;
	struct scsi_inquiry *inquiry;
	struct scsi_device sd;
	int error = 0, s;

	MKSCSIDEV(sd,dev);

	if ((sd.bus >= NSCSI) || !(sc = scsi_controller[sd.bus]))
		return(ENXIO);

	switch (op) {

	case SIOC_GET_DEVICE:
		error = nsuser(u.u_procp->p_ucred, &u.u_acflag);
		if (error)
			return(error);
		*(struct scsi_device *)arg = sd;
		return (0);

	case SIOC_SET_DEVICE:
		error = nsuser(u.u_procp->p_ucred, &u.u_acflag);
		if (error)
			return(error);
		sd = *(struct scsi_device *)arg;
		return (0);
	
	case SIOC_ALLOCATE:
		error = nsuser(u.u_procp->p_ucred, &u.u_acflag);
		if (error)
			return(error);
		return (0);

	case SIOC_TEST_UNIT:
		cmd = scsi_command_alloc();
		SCSI_CMD6(cmd, SCSI_TEST_UNIT_READY);
		cmd->flags = SCSI_WAIT;
		cmd->device = sd;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
		if (cmd->scsi_status != SCSI_GOOD) {
			error = ENXIO;
			if (cmd->sense) {
				if (cmd->sense->sense_key == SCSI_SENSE_NOT_READY)
					error = EIO;
			}
		}
		if (cmd->sense) scsi_sense_free(cmd->sense);
		scsi_command_free(cmd);
		return (error);

	case SIOC_INQUIRY:
		cmd = scsi_command_alloc();
		inquiry = (struct scsi_inquiry *)arg;
		SCSI_CMD6_BUF(cmd, SCSI_INQUIRY, *inquiry);
		cmd->device = sd;
		cmd->flags = SCSI_DATA_IN | SCSI_WAIT;
		scsi_command(SCSI_QUEUE_TAIL, cmd);
		if (cmd->scsi_status != SCSI_GOOD) {
			error = ENXIO;
			if (cmd->sense) {
				error = EIO;
			}
		}
		if (cmd->sense) scsi_sense_free(cmd->sense);
		scsi_command_free(cmd);
		return (error);

	default:
		return(ENOTTY);
	}
}
#endif

#ifdef	MACH_KERNEL
#define SALLOC(z, v, type, panic_msg) {	\
	int s = splbio();		\
	ZGET((z), (v), type);		\
	if (!v)				\
		panic(panic_msg);	\
	splx(s);			\
}

#define SFREE(z, v) {			\
	int s = splbio();		\
	ZFREE((z), (vm_offset_t)(v));	\
	splx(s);			\
}
#else
#define SALLOC(z, v, type, panic_msg) {	\
	int s = splbio();		\
	v = (type)malloc(scsi_command_size, M_DEVBUF, M_NOWAIT); \
	if (!v)				\
		panic(panic_msg);	\
	splx(s);			\
}

#define SFREE(z, v) {			\
	int s = splbio();		\
	free((caddr_t)(v), M_DEVBUF);	\
	splx(s);			\
}

#define	ZFREE(z, v) {			\
	free((caddr_t)v, M_DEVBUF);	\
}
#endif

struct scsi_cmd *
scsi_command_alloc()
{
	struct scsi_cmd *cmd;
	SALLOC(scsi_command_zone, cmd, struct scsi_cmd *, "scsi_command_alloc: command zone empty");
	return(cmd);
}

struct scsi_request_sense *
scsi_sense_alloc()
{
	struct scsi_request_sense *sense;
	SALLOC(scsi_command_zone, sense, struct scsi_request_sense *, "scsi_sense_alloc: sense zone empty");
	return(sense);
}

void
scsi_command_free(cmd)
struct scsi_cmd *cmd;
{
	if (!cmd)
		panic("scsi_command_free: attempt to free NULL cmd");
	SFREE(scsi_command_zone, cmd);
}

void
scsi_sense_free(sense)
struct scsi_request_sense *sense;
{
	if (!sense)
		panic("scsi_sense_free: attempt to free NULL sense");
	SFREE(scsi_command_zone, sense);
}

#endif  /* NSCSI > 0 */
