/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * Common SCSI driver declarations
 */

#ifndef _HP_SCSI_COMMON_H_
#define _HP_SCSI_COMMON_H_ 1


#define SCSI_SELECT_TIMEOUT     250             /* SCSI selection timeout (ms)  */
#define SCSI_INITIATOR          7               /* Target ID of SCSI initiator  */
#define SCSI_MAX_MSG_LEN        12              /* Maximum message size accepted */

/* scsi bus phases */
#define SCSI_BP_DATA_OUT        0x00            /* data out phase */
#define SCSI_BP_DATA_OUT        0x00            /* data out phase */
#define SCSI_BP_DATA_IN         0x01            /* data in phase */
#define SCSI_BP_CMD             0x02            /* command phase */
#define SCSI_BP_STAT            0x03            /* status phase */
#define SCSI_BP_MSG_OUT         0x06            /* message out phase */
#define SCSI_BP_MSG_IN          0x07            /* message in phase */
                                                
/* Scsi messages */                             
#define SCSI_MSG_COMPLETE       0x00            /* Command complete messsage */
#define SCSI_MSG_EXTENDED       0x01            /* Extended message */
#define SCSI_MSG_SDP            0x02            /* Save data pointers */
#define SCSI_MSG_RDP            0x03            /* Restore data pointers */
#define SCSI_MSG_DIS            0x04            /* Disconnect */
#define SCSI_MSG_IDE            0x05            /* Initiator detected error */
#define SCSI_MSG_ABORT          0x06            /* Abort */
#define SCSI_MSG_REJ            0x07            /* Message reject */
#define SCSI_MSG_NOP            0x08            /* No operation */
#define SCSI_MSG_PERROR         0x09            /* Parity error */
#define SCSI_MSG_LCC            0x0a            /* Linked command complete */
#define SCSI_MSG_LCCF           0x0b            /* Linked command complete with flag */
#define SCSI_MSG_BDR            0x0c            /* Bus device reset */

#define SCSI_MSG_CLRQ           0x0e            /* Clear queue (queueing) */
#define SCSI_MSG_INITREC        0x0f            /* Initiate recovery (queueing) */
#define SCSI_MSG_RLSREC         0x10            /* Release recovery (queueing) */
#define SCSI_MSG_MUNDQTAG       0x20            /* Mundane/Simple queue tag (queueing) */
#define SCSI_MSG_HQTAG          0x21            /* Head of queue tag (queueing) */
#define SCSI_MSG_ORDQTAG        0x22            /* Ordered queue tag (queueing) */
#define SCSI_MSG_ABTAG          0x23            /* Abort tag (queueing) */

#define SCSI_MSG_ID_MASK        0x80            /* Identify message mask */
#define SCSI_MSG_ID_DISC_EN     0x40            /* Disconnect Enable */

/* Host identify message byte */
#define SCSI_HOST_ID_MESS       (SCSI_MSG_ID_MASK|SCSI_MSG_ID_DISC_EN)

/* Scsi extended messages */
#define SCSI_EMSG_SDTR          0x01            /* Synchronous data transfer request code */
#define SCSI_EMSG_WDTR          0x03            /* Wide data transfer request code */

#define SCSI_EMSG_ID            0               /* Symbols for emsg bytes       */
#define SCSI_EMSG_LEN           1
#define SCSI_EMSG_CODE          2
#define SCSI_EMSG_XFER_PERIOD   3 
#define SCSI_EMSG_XFER_WIDTH	3
#define SCSI_EMSG_REQ_ACK_OFF   4


/* Scsi manager software states - notice these are a superset of the scsi_bp* above */
#define SCSI_SS_DATA_OUT        0x00            /* Data out state */
#define SCSI_SS_DATA_IN         0x01            /* Data in state */
#define SCSI_SS_CMD             0x02            /* Command state */
#define SCSI_SS_STAT            0x03            /* Status state */
#define SCSI_SS_RES0            0x04            /* Dont mess */
#define SCSI_SS_RES1            0x05            /* with these */
#define SCSI_SS_MSG_OUT         0x06            /* Message out state */
#define SCSI_SS_MSG_IN          0x07            /* Message in state */
#define SCSI_SS_DIS             0x08            /* Disconnected */
#define SCSI_SS_ACC             0x09            /* Accepting message */
#define SCSI_SS_RES             0x0a            /* Reselected */
#define SCSI_SS_ASM             0x0b            /* Accepting status (command complete) message */
#define SCSI_SS_BUS_FREE        0x0c            /* Bus Free */
                                                                
/* Scsi manager transfer information states - used for establishing sync vs async data transfer */
#define SCSI_XT_UNKNOWN         0x00            /* Not known yet - must send sync data transfer request message */
#define SCSI_XT_WMSGSENT        0x01            /* We sent a wide data request message */
#define SCSI_XT_SMSGSENT        0x02            /* We sent a sync data request message */
#define SCSI_XT_SYNC            0x03            /* We are synchronous */
#define SCSI_XT_ASYNC           0x04            /* We are asynchrounous */

/* Scsi manager abort/reset states - used for handling abort and reset in a multitude of conditions */
#define SCSI_AR_NA              0x00            /* Not performing abort or reset */
#define SCSI_AR_SEL             0x01            /* Must select target to do abort or reset */
#define SCSI_AR_ATN             0x02            /* We want to send an abort or reset and we raised ATN */
#define SCSI_AR_SENT            0x03            /* We sent an abort or reset message */

/* Scsi manager command queueing states - used for handling SCSI-2 command queueing */
#define SCSI_CQ_UNKNOWN         0x00            /* Attempt to use command queueing */
#define SCSI_CQ_YES             0x01            /* Use command queueing message */
#define SCSI_CQ_NO              0x02            /* Not using command queueing */
#define SCSI_CQ_MSGSENT         0x04            /* We sent a queue message */

#endif /* _HP_SCSI_COMMON_H_ */
