/*
 * Copyright (c) 1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 *      Utah $Hdr: lrmsg.s 1.4 95/01/20$
 */
/*
 *	Fast migrating mach_rpc_trap path using registers
 *	(a.k.a. ``leaky-register RPC'').  On the PA we use:
 *
 *	arg0:	request port
 *	arg1:	ID
 *	arg2:	sequence number (from kernel)
 *	arg3:	XXX more activation flag (from kernel)
 *	r5-12:	8 simple arguments
 *	r13-14:	1 port argument (name, msgt_name)
 *	r15-r16:1 indirect (kernel-copied) memory argument (pointer, size)
 */

#define	DEBUGABORT	/* do a dump_ssp if we KERN_ABORTED for any reason */

#include <machine/asm.h>
#include <assym.s>
#include <machine/psl.h>

#define FR_SIZE		64	/* XXX */

	.space	$PRIVATE$
	.subspa	$DATA$

	.import cur_pcb,data
	.import need_ast,data
	.import istackptr,data/* XXX */
	.import kernel_port,data

	.space	$TEXT$
	.subspa $CODE$

/*
 * On entry:
 *	arg0-4	= message header
 *	r5-r16	= message arguments
 *	sr1     = kernel space
 *	pidr4   = kernel PID
 */
	.export	$lrrpc,entry
$lrrpc
	/*
	 * Find the current activation (t3) and PCB (ret1).
 	 */
	ldil	L%cur_pcb,r1
	ldw	R%cur_pcb(sr1,r1),ret1
	ldw	PCB_ACT(sr1,ret1),t3

	/*
	 * Save the relevant user state.
	 */
	stw	rp,SS_R2(sr1,ret1)
#ifdef R4_IS_SPECIAL
	stw	r4,SS_R4(sr1,ret1)
#endif
	stw	dp,SS_R27(sr1,ret1)
	stw	sp,SS_R30(sr1,ret1)
	stw	r31,SS_R31(sr1,ret1)

	/*
	 * Save the space registers.
	 * We need to save sr0 since it holds the space we need to return to.
	 * No need to save sr5 and sr6 as they are the same as sr4.
	 * No need to save sr7 because it is constant (kernel space).
	 */
	mfsp	sr0,t4
	stw	t4,SS_SR0(sr1,ret1)
	mfsp	sr4,t4
	stw	t4,SS_SR4(sr1,ret1)
#ifdef MULTIPLE_SPACES_PER_TASK
	mfsp	sr5,t4
	stw	t4,SS_SR5(sr1,ret1)
	mfsp	sr6,t4
	stw	t4,SS_SR6(sr1,ret1)
#endif

	/*
	 * Set the necessary space registers to point to the kernel.
	 * We no longer have to explicitly use sr1 when accessing data.
	 */
	mtsp	r0,sr4
	mtsp	r0,sr5
	mtsp	r0,sr6

	/*
	 * Find the current thread (r1) and IPC space (t4)
	 * XXX read lock the space
	 */
	ldw	ACT_TASK(t3),t4
	ldw	ACT_THREAD(t3),r1
	ldw	TASK_SPACE(t4),t4

	/*
	 * Lookup the destination port (t4).
	 * If it isn't in the table, we divert and call ipc_entry_lookup.
	 */
	copy	arg0,t1
	ldw	SPACE_TABLE_SIZE(t4),t2
	comb,>>=,n t1,t2,$divert_dest_lookup
	ldw	SPACE_TABLE(t4),t4
#if IPC_ENTRY_SIZE == 16
	shd	t1,r0,28,t1
	add	t1,t4,t2
#else
#if IPC_ENTRY_SIZE == 20
	sh2add	t1,t1,t1
	sh2add	t1,t4,t2
#else
#if IPC_ENTRY_SIZE == 24
	sh1add	t1,t1,t1
	sh3add	t1,t4,t2
#else
	b	$abort
	break	0,8
#endif
#endif
#endif
$divert_dest_lookup_back
	ldw	ENTRY_BITS(t2),t1
	ldw	ENTRY_OBJECT(t2),t4
	/*
	 * XXX lock the port (object), release the space lock
	 */
#ifdef DEBUGABORT
	bb,<,n	t1,MACH_PORT_TYPE_SEND_BIT,$a1
	break	0,8
	b,n	$abort
$a1
#else
	bb,>=,n t1,MACH_PORT_TYPE_SEND_BIT,$abort
#endif

	/*
	 * While we have a pointer to the port structure,
	 * save the port's receiver name in the header.
	 */
	ldw	IP_RECEIVER_NAME(t4),arg0

	/*
	 * XXX save a pointer to the sequence number for later
	 */
	ldo	IP_SEQNO(t4),t2
	stw	t2,ACT_SPARE+52(t3)

	/*
	 * XXX save reply buffer info for later
	 */
	stw	arg2,ACT_SPARE+56(t3)
	stw	arg3,ACT_SPARE+60(t3)

	/*
	 * XXX check for kernel object.  Ideally we would not have
	 * special case code, the appropriate actpool would be attached
	 * to the real port.
	 */
	ldw	IP_BITS(t4),t1
	extru	t1,31,16,t1
	comb,=,n r0,t1,$notkobject

	/*
	 * This port refers to a kernel object.
	 * Put this internal port pointer in the message.
	 * Also, make sure we grab an act from the kernel_port pool.
	 */
	ldil	L%kernel_port,t4
	ldw	R%kernel_port(t4),t4

$notkobject
	/*
	 * Find the port's current target (t4).
	 * It'll be either the one in itself or the one in its port set.
	 */
	ldw	IP_CUR_TARGET(t4),t4

	/*
	 * Grab an activation from the port's pool (t2).
	 * If the pool is empty, block waiting for one.
	 */
	ldw	IPT_ACTS(t4),t2
	comb,=,n t2,r0,$needact
	ldw	ACT_IPT_NEXT(t2),t1
	stw	t1,IPT_ACTS(t4)

$gotact
	/*
	 * Handle any necessary port translation in the message
	 */
	comb,=,n r0,r13,$noports
	/*
	 * kr = mach_port_rpc_copy(portp, sact, dact)
	 */
	.import	mach_port_rpc_copy,code
	ldil	L%mach_port_rpc_copy,ret0
	ldo	R%mach_port_rpc_copy(ret0),ret0
	stw	arg0,ACT_SPARE+0(t3)
	stw	arg1,ACT_SPARE+4(t3)
	stw	r13,ACT_SPARE+8(t3)
	stw	r14,ACT_SPARE+12(t3)
	ldo	ACT_SPARE+8(t3),arg0
	copy	t3,arg1
	bl	$lrrpc_kernel_call,rp
	copy	t2,arg2
	comb,<>	r0,ret0,$returnserveract
	ldw	ACT_PCB(t3),t1		/* non-nullified delay slot */
	ldw	ACT_SPARE+0(t3),arg0
	ldw	ACT_SPARE+4(t3),arg1
	ldw	ACT_SPARE+8(t3),r13
	ldw	ACT_SPARE+12(t3),r14

$noports
	/*
	 * Attach server activation (t2) to the current thread (r1).
	 * The old activation is still in t3.
	 */
	stw	t3,ACT_LOWER(t2)
	stw	t2,ACT_HIGHER(t3)
	stw	t2,THREAD_TOP_ACT(r1)
	stw	r1,ACT_THREAD(t2)
	stw	r0,ACT_USER_ACT(t2)

	/*
	 * Find the new activation's PCB (t1),
	 * and make it officially the current one.
	 */
	ldw	ACT_PCB(t2),t1
	ldil	L%cur_pcb,arg3
	stw	t1,R%cur_pcb(arg3)

	/* XXX refcount, alerts */

	/*
	 * Indicate that no floating-point state exists yet in the new PCB
	 * by stuffing an illegal value in the saved fr0.
	 * Then disable the FPU to force a trap on access in the new context.
	 */
	ldi	-1,arg3
	stw	arg3,PCB_FS+FS_FR0(t1)
	mtctl	r0,ccr

	/*
	 * Handle any indirect memory copies in the message.
	 *
	 * There must be both a non-zero client memory pointer (r15) and
	 * a non-zero server memory pointer (ACT_USER_RBUF).
	 *
	 * Note: we always load the server memory pointer for use in replies.
	 */
	movb,=	r15,t1,$nomem
	ldw	ACT_USER_RBUF(t2),r15	/* non-nullified delay slot */
	comb,=,n r0,r15,$nomem
	/*
	 * Check the copy size, if either the request (r16) or server
	 * buffer (ACT_USER_RBUF_SIZE) sizes are zero, we skip the copy.
	 */
	movb,=,n r16,arg2,$nomem
	ldw	ACT_USER_RBUF_SIZE(t2),r16
	comb,=,n r0,r16,$nomem
	/*
	 * Now take the minimum of the request and server buffer sizes
	 * and store in both in the message (r16) and the the argument to
	 * the copy routine (arg2).
	 */
	comclr,>= r16,arg2,r0
	or,tr	r16,r0,arg2		/* may be nullified */
	copy	arg2,r16		/* may be nullified */
	/*
	 * Finally, copy size (arg2) bytes from client mem (arg0) to
	 * server mem (arg1).
	 */
	stw	arg0,ACT_SPARE+0(t3)
	stw	arg1,ACT_SPARE+4(t3)
	copy	t1,arg0
	copy	r15,arg1
	bl	$lrrpc_memcopy,rp
	ldi	1,arg3
	comb,<>,n r0,ret0,$freeserveract
	ldw	ACT_SPARE+0(t3),arg0
	ldw	ACT_SPARE+4(t3),arg1

$nomem
	/*
	 * Grab and then update the receive port sequence number.
	 * XXX this should be done with the message queue lock held.
	 */
	ldw	ACT_SPARE+52(t3),t1
	ldw	0(t1),arg2
	ldo	1(arg2),arg3
	stw	arg3,0(t1)

#ifdef ACTCHECK
	ldw		ACT_RETURN_ROUTINE(t2),t1
	comb,=,n	r0,t1,$rtis0
	break		0x1f,0x1ffd
$rtis0
#endif
	/*
	 * Set the server's return_routine.
	 */
	ldil	L%$lrfast_return,rp
	ldo	R%$lrfast_return(rp),rp
	stw	rp,ACT_RETURN_ROUTINE(t2)

	/*
	 * Establish server PID.
	 */
	ldw	ACT_PID(t2),arg3
	mtctl	arg3,pidr1

	/*
	 * XXX experimental: if this was the last activation in the pool,
	 * piggyback an indication to the server via an additional argument
	 * to the entry routine.
	 */
	ldw	ACT_IPT_NEXT(t2),t1
	comclr,<> r0,t1,arg3
	ldi	1,arg3

	/*
	 * Set up the remaining user register state.
	 * The user SID and SP come from the target activation.
	 * The user data pointer and entry point come from the ipc_target.
	 */
	ldw	ACT_USER_STACK(t2),sp
	ldw	IPT_USER_DATA(t4),dp
	ldo	FR_SIZE(sp),sp
	ldw	ACT_SPACE(t2),t2
	ldw	IPT_USER_ENTRYPOINT(t4),r31

	/*
	 * Set space registers and protection IDs for the server.
	 * Note we must wait to set pidr4 until the last minute.
	 */
	mtsp	t2,sr4
	mtsp	t2,sr5
	mtsp	t2,sr6

	/*
	 * Finally invoke the server and set the protection ID correctly
	 */
	.call
	be	0(sr4,r31)
	mtctl	r0,pidr4

$lrfast_return
	/*
	 * Just came from the return path in the gateway page.
	 * t1 = current PCB, t2 = server activation.
	 */

	/*
	 * Find the thread (r1) and client activation (t3)
	 */
	ldw	ACT_THREAD(t2),r1
	ldw	ACT_LOWER(t2),t3

	/*
	 * Handle reply port translation in the message
	 */
	comb,=,n r0,r13,$noports2
	/*
	 * kr = mach_port_rpc_copy(portp, sact, dact)
	 */
	stw	ret0,ACT_SPARE+0(t3)
	stw	ret1,ACT_SPARE+4(t3)
	.import	mach_port_rpc_copy,code
	ldil	L%mach_port_rpc_copy,ret0
	ldo	R%mach_port_rpc_copy(ret0),ret0
	stw	r13,ACT_SPARE+8(t3)
	stw	r14,ACT_SPARE+12(t3)
	ldo	ACT_SPARE+8(t3),arg0
	copy	t2,arg1
	bl	$lrrpc_kernel_call,rp
	copy	t3,arg2
	comiclr,<> 0,ret0,r0	/* XXX if m_p_rpc_copy failed, return that */
	ldw	ACT_SPARE+0(t3),ret0
	ldw	ACT_SPARE+4(t3),ret1
	ldw	ACT_SPARE+8(t3),r13
	ldw	ACT_SPARE+12(t3),r14

$noports2
	/*
	 * Handle any indirect memory copies in the reply message.
	 *
	 * There must be both a non-zero server memory pointer (r15) and
	 * a non-zero client memory pointer (ACT_SPARE+56).
	 *
	 * Note: we always load the client's memory pointer.
	 */
	movb,=	r15,arg0,$nomem2
	ldw	ACT_SPARE+56(t3),r15	/* non-nullified delay slot */
	movb,=,n r15,arg1,$nomem2
	/*
	 * Check the copy size, if either the reply (r16) or client
	 * buffer (ACT_SPARE+60) sizes are zero, we skip the copy.
	 */
	movb,=,n r16,arg2,$nomem2
	ldw	ACT_SPARE+60(t3),r16
	comb,=,n r0,r16,$nomem2
	/*
	 * Now take the minimum of the reply and client buffer sizes
	 * and store in both in the message (r16) and the the argument to
	 * the copy routine (arg2).
	 */
	comclr,>= r16,arg2,r0
	or,tr	r16,r0,arg2		/* may be nullified */
	copy	arg2,r16		/* may be nullified */
	/*
	 * Finally, copy size (arg2) bytes from server mem (arg0) to
	 * client mem (arg1).
	 */
	stw	ret0,ACT_SPARE+0(t3)
	stw	ret1,ACT_SPARE+4(t3)
	bl	$lrrpc_memcopy,rp
	ldi	0,arg3
	comiclr,<> 0,ret0,r0	/* XXX if lrrpc_memcopy failed, return that */
	ldw	ACT_SPARE+0(t3),ret0
	ldw	ACT_SPARE+4(t3),ret1

$nomem2
	/*
	 * Restore the client state.
	 * PID is restored here, SID and registers restored in syscall_reentry.
	 */
	ldw	ACT_PID(t3),t1
	mtctl	t1,pidr1

$freeserveract
	/*
	 * Remove the server activation from the thread,
	 * making the client activation the current one.
	 * At this point:
	 *	r1 = current thread
	 *	t2 = server activation
	 *	t3 = user activation
	 */
	stw	r0,ACT_THREAD(t2)
	stw	t3,THREAD_TOP_ACT(r1)
	ldw	ACT_PCB(t3),t1
	ldil	L%cur_pcb,t4
	stw	t1,R%cur_pcb(t4)
#ifdef ACTCHECK
	stw	r0,ACT_LOWER(t2)
	stw	r0,ACT_HIGHER(t3)
#endif

$returnserveract
	/*
	 * Return the server activation to its ipc_target and wakeup anyone
	 * waiting for an activation from that target.  At this point:
	 *	t1 = current PCB
	 *	t2 = server activation
	 *	
	 */
	ldw	ACT_IPT(t2),arg0
	comb,=,n arg0,r0,$return_no_target
	ldw	IPT_ACTS(arg0),t4
	stw	t2,IPT_ACTS(arg0)
	stw	t4,ACT_IPT_NEXT(t2)
	ldw	IPT_WAITING(arg0),t4
	comb,<>,n r0,t4,$actwakeup

$return_no_target
	/*
	 * Return to syscall path for common exit.
	 * At this point, t1 should point to current PCB
	 */
	.import	$syscall_reentry,code
	ldil	L%$syscall_reentry,t2
	ldo	R%$syscall_reentry(t2),t2
	bv	0(t2)
	nop

/*
 * Following are out-of-line code fragments to perform various,
 * non-common-case actions.
 */

/*
 * RPC failed prior to entering server.
 * Load the PCB pointer back in t1 and return to syscall path with a failure.
 *
 * XXX should distinguish the various errors.
 */
$abort
	ldil	L%cur_pcb,t1
	ldw	R%cur_pcb(t1),t1
	.import	$syscall_reentry,code
	ldil	L%$syscall_reentry,t2
	ldo	R%$syscall_reentry(t2),t2
	bv	0(t2)
	ldi	KERN_ABORTED,ret0

/*
 * No server activation available.  If the port/pset supports migrating RPC,
 * block waiting for a free activation, otherwise fail.  At this point:
 *	t3:	client activation
 *	t4:	ipc_target
 * Returns with:
 *	t2:	server activation
 */
$needact
	/*
	 * See if target supports migrating RPCs
	 */
	ldw	IPT_TYPE(t4),t1
#ifdef DEBUGABORT
	bb,<,n	t1,IPT_TYPE_MIGRATE_RPC_BIT,$waitforact
	b	$abort
	break	0,8
$waitforact
#else
	bb,>=	t1,IPT_TYPE_MIGRATE_RPC_BIT,$abort /* note: backward branch */
	nop
#endif
	/*
	 * act = ipc_target_block(ipt)
	 */
	.import	ipc_target_block,code
	ldil	L%ipc_target_block,ret0
	ldo	R%ipc_target_block(ret0),ret0
	stw	arg0,ACT_SPARE+0(t3)
	stw	arg1,ACT_SPARE+4(t3)
	bl	$lrrpc_kernel_call,rp
	copy	t4,arg0
	copy	ret0,t2
	ldw	ACT_SPARE+0(t3),arg0
	b	$gotact
	ldw	ACT_SPARE+4(t3),arg1

/*
 * Wakeup anyone waiting for an activation from this target.  At this point:
 *	arg0:	ipc_target
 *	ret0:	RPC return code
 *	ret1:	RPC second return value (XXX not used)
 */
$actwakeup
	/*
	 * ipc_target_wakeup(ipt)
	 */
	copy	ret0,t2
	copy	ret1,t4
	.import	ipc_target_wakeup,code
	ldil	L%ipc_target_wakeup,ret0
	bl	$lrrpc_kernel_call,rp
	ldo	R%ipc_target_wakeup(ret0),ret0
	copy	t2,ret0
	b	$return_no_target
	copy	t4,ret1

/*
 * Need to call up to kernel to perform non-trivial name translation
 * on the destination port.  At this point:
 *	arg0:	name
 *	t4:	ipc_space
 * Returns with:
 *	t2:	ipc_entry
 */
$divert_dest_lookup
	/*
	 * entry = ipc_entry_lookup(space, name)
	 */
	.import	ipc_entry_lookup,code
	ldil	L%ipc_entry_lookup,ret0
	ldo	R%ipc_entry_lookup(ret0),ret0
	stw	arg0,ACT_SPARE+0(t3)
	stw	arg1,ACT_SPARE+4(t3)
	copy	arg0,arg1
	bl	$lrrpc_kernel_call,rp
	copy	t4,arg0
	ldw	ACT_SPARE+0(t3),arg0
	movb,<>	ret0,t2,$divert_dest_lookup_back
	ldw	ACT_SPARE+4(t3),arg1
#ifdef DEBUGABORT
	break	0,8
#endif
	b,n	$abort

/*
 * Copy from one address space to another.
 *
 * XXX always called in the "server" context;
 *     i.e. the destination during a request, the source during a reply.
 *
 * Passed in:
 *	arg0:	source offset
 *	arg1:	destination offset
 *	arg2:	size
 *	arg3:	1 if request, 0 if reply
 *	r1:	current thread
 * Uses:
 *	sr1:	source space
 *	sr2:	destination space
 *	pidr2:	source PID
 *	pidr3:	destination PID
 *	t1:	scratch, clobbered
 *	ret0-1:	scratch, clobbered
 *	r5-r8:	scratch, saved/restored as needed
 */
$lrrpc_memcopy
	/*
	 * Setup the source and destination space and PID registers
	 *	request: source is client, destination is server
	 *	reply:	 source is server, destination is client
	 */
	comb,=,n arg3,r0,$memcopy_isreply
	ldw	ACT_SPACE(t3),t1
	ldw	ACT_SPACE(t2),arg3
	mtsp	t1,sr1
	mtsp	arg3,sr2
	ldw	ACT_PID(t3),t1
	ldw	ACT_PID(t2),arg3
	mtctl	t1,pidr2
	mtctl	arg3,pidr3
	b,n	$memcopy_setupdone
$memcopy_isreply
	ldw	ACT_SPACE(t2),t1
	ldw	ACT_SPACE(t3),arg3
	mtsp	t1,sr1
	mtsp	arg3,sr2
	ldw	ACT_PID(t2),t1
	ldw	ACT_PID(t3),arg3
	mtctl	t1,pidr2
	mtctl	arg3,pidr3

$memcopy_setupdone
	/*
	 * Prepare to catch faults
	 */
	ldil	L%$memcopy_error,t1
	ldo	R%$memcopy_error(t1),t1
	stw	t1,THREAD_RECOVER(r1)

	/*
	 * XXX if either source or destination is not word aligned or the
	 * count is not a multiple of the word size, punt and copy bytes
	 */
	or	arg0,arg1,t1
	or	t1,arg2,t1
	extru,= t1,31,2,r0
	b,n	$memcopy_bytes

	/*
	 * If less than 4 words, skip block move
	 */
	addib,<,n -16,arg2,$memcopy_movewords

$memcopy_loop16
	/*
	 * We can move the data in 4 word moves.
	 * We'll use 4 registers to avoid interlock and pipeline stalls.
	 */
	stw	ret0,ACT_SPARE+16(t3)
	stw	ret1,ACT_SPARE+20(t3)
	ldwm	16(sr1,arg0),t1
	ldw	-12(sr1,arg0),arg3
	ldw	-8(sr1,arg0),ret0
	ldw	-4(sr1,arg0),ret1
	stwm	t1,16(sr2,arg1)
	stw	arg3,-12(sr2,arg1)
	stw	ret0,-8(sr2,arg1)
	addib,>= -16,arg2,$memcopy_loop16
	stw	ret1,-4(sr2,arg1)
	ldw	ACT_SPARE+16(t3),ret0
	ldw	ACT_SPARE+20(t3),ret1

$memcopy_movewords
	/*
	 * We have already decremented the count by 16, add 12 to it and then 
	 * we can test if there is at least 1 word left to move.
	 */
	addib,<,n 12,arg2,$memcopy_done

$memcopy_loop4
	/*
	 * Clean up any remaining words that were not moved in the 16 byte
	 * moves
	 */
	ldwm	4(sr1,arg0),t1
	addib,>= -4,arg2,$memcopy_loop4
	stwm	t1,4(sr2,arg1)

$memcopy_done
	/*
	 * Clear out pidrs and return
	 */
	mtctl	r0,pidr2
	mtctl	r0,pidr3
	stw	r0,THREAD_RECOVER(r1)
	bv	0(rp)
	ldi	0,ret0

$memcopy_bytes
        ldbs,ma	1(sr1,arg0),t1
        addib,> -1,arg2,$memcopy_bytes
        stbs,ma t1,1(sr2,arg1) 
	b,n	$memcopy_done

$memcopy_error
	/*
	 * Got an unresolved fault
	 */
	mtctl	r0,pidr2
	mtctl	r0,pidr3
	stw	r0,THREAD_RECOVER(r1)
	bv	0(rp)
	ldi	1,ret0

/*
 * Make an upcall to non-glue code.
 * Allocates a kernel activation and makes the call.
 *
 * Passed in:
 *	ret0: address of routine to call
 *	arg?: arguments for routine
 *	rp: return address
 * Returned:
 *	ret0-1: upcall return value(s)
 */
$lrrpc_kernel_call

	/*
	 * Get current PCB (ret1) and save caller-save registers
	 * (except for args and returns).
	 * XXX SS offset in PCB assumed to be 0
	 */
	ldil	L%cur_pcb,ret1
	ldw	R%cur_pcb(ret1),ret1
	stw	r1,SS_R1(ret1)
	/* rp cannot be saved in SS */
	stw	t4,SS_R19(ret1)
	stw	t3,SS_R20(ret1)
	stw	t2,SS_R21(ret1)
	stw	t1,SS_R22(ret1)
	/* do not save arg0-3 */
	/* do not save ret0-1 */
	/* r31 cannot be saved in SS */
#if 0
	/* sr0 cannot be saved in SS */
	mfsp	sr1,t1
	stw	t1,SS_SR1(ret1)
	mfsp	sr2,t1
	stw	t1,SS_SR2(ret1)
#endif

	/*
	 * Get current activation (t3) and thread (r1)
	 */
	ldw	PCB_ACT(ret1),t3
	ldw	ACT_THREAD(t3),r1

	/*
	 * Allocate a kernel activation (t1).
	 */
	ldil	L%kernel_port,t4
	ldw	R%kernel_port(t4),t4
	ldw	IPT_ACTS(t4),t1
	comb,<>,n t1,r0,$kact_available

	/*
	 * XXX block instead
	 */
	ldil	L%istackptr,t2
	ldw	R%istackptr(t2),sp
	stw	r0,R%istackptr(t2)
	.import panic_no_kstacks,code
	ldil	L%panic_no_kstacks,t2
	ldo	R%panic_no_kstacks(t2),t2
	bv	0(t2)
	ldo	FM_SIZE(sp),sp

$kact_available

	/*
	 * Remove it from the ipc_target.
	 */
	ldw	ACT_IPT_NEXT(t1),t2
	stw	t2,IPT_ACTS(t4)

	/*
	 * Attach the kernel activation (t1) to the current thread.
	 */
	stw	t3,ACT_LOWER(t1)
	stw	t1,ACT_HIGHER(t3)
	stw	t1,THREAD_TOP_ACT(r1)
	stw	r1,ACT_THREAD(t1)

	/*
	 * Take a reference on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t3),t4
	ldo	1(t4),t4
	stw	t4,ACT_REF_COUNT(t3)
	stw	t3,ACT_USER_ACT(t1)

	/*
	 * Make the new activation officially the current one.
	 */
	ldw	ACT_PCB(t1),ret1
	ldil	L%cur_pcb,t4
	stw	ret1,R%cur_pcb(t4)

	/*
	 * Get the new stack pointer and stash info we need later:
	 *	current sp (sp)
	 *	current dp (dp)
	 *	return address (rp)
	 *	kernel activation (t1)
	 *	current thread (r1)
	 *	caller activation (t3)
	 *	caller r31 (r31)
	 *	caller sr0 (sr0)
	 * We cannot use the activation saved state as value may have
	 * already been stored there.
	 */
	ldw	ACT_USER_STACK(t1),t2
	ldo	63(t2),t2
	depi	0,31,6,t2
	stw	sp,0(t2)
	stw	dp,4(t2)
	stw	rp,8(t2)
	stw	t1,12(t2)
	stw	r1,16(t2)
	stw	t3,20(t2)
	stw	r31,24(t2)
#if 0
	mfsp	sr0,t4
	stw	t4,28(t2)
#endif

	/*
	 * Initialize the kernel stack pointer and set the kernel data pointer.
	 */
	ldo	FR_SIZE+64(t2),sp
	ldil	L%$global$,dp
	ldo	R%$global$(dp),dp

	/*
	 * Make the call
	 */
	stw     r0,FM_PSP(sp)
	.call
	blr	r0,rp
	bv	0(ret0)
	nop

	/*
	 * Reload stashed state.
	 */
	ldo	-FR_SIZE-64(sp),t2
	ldw	0(t2),sp
	ldw	4(t2),dp
	ldw	8(t2),rp
	ldw	12(t2),t1
	ldw	16(t2),r1
	ldw	20(t2),t3
	ldw	24(t2),r31
#if 0
	ldw	28(t2),t4
	mtsp	t4,sr0
#endif

	/*
	 * Remove the kernel activation from the thread.
	 */
	stw	r0,ACT_THREAD(t1)
	stw	t3,THREAD_TOP_ACT(r1)

#ifdef ACTCHECK
	stw	r0,ACT_LOWER(t1)
	stw	r0,ACT_HIGHER(t3)
	stw	r0,ACT_USER_ACT(t1)
#endif

	/* 
	 * Drop the reference count on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t3),t2
	ldo	-1(t2),t2
	stw	t2,ACT_REF_COUNT(t3)

	/*
	 * Return the kernel activation to the kernel ipc_target.
	 */
	ldil	L%kernel_port,t2
	ldw	R%kernel_port(t2),t2
	ldw	IPT_ACTS(t2),t4
	stw	t1,IPT_ACTS(t2)
	stw	t4,ACT_IPT_NEXT(t1)

	/*
	 * Restore the caller activation.
	 */
	ldw	ACT_PCB(t3),arg0
	ldil	L%cur_pcb,t4
	stw	arg0,R%cur_pcb(t4)

	/*
	 * Restore the caller-save registers and return.
	 */
	ldw	SS_R1(arg0),r1
	/* rp already restored */
	ldw	SS_R19(arg0),t4
	ldw	SS_R20(arg0),t3
	ldw	SS_R21(arg0),t2
	/* do not restore arg0-3 */
	/* do not restore ret0-1 */
	/* r31 already restored */
#if 0
	/* sr0 already restored */
	ldw	SS_SR1(arg0),arg1
	mtsp	arg1,sr1
	ldw	SS_SR2(arg0),arg1
	mtsp	arg1,sr2
#endif
	bv	0(rp)
	ldw	SS_R22(arg0),t1

	.end
