/*
 *  (c) Copyright 1988 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * Copyright (c) 1990,1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * THE UNIVERSITY OF UTAH AND CSL PROVIDE THIS SOFTWARE IN ITS "AS IS"
 * CONDITION, AND DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM ITS USE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: model_dep.c 1.33 94/11/10$
 */

#include <kern/thread.h>
#include <kern/current.h>

#include <machine/pmap.h>
#include <machine/mach_param.h>
#include <machine/regs.h>
#include <machine/psl.h>
#include <machine/break.h>
#include <device/param.h>
#include <sys/reboot.h>
#include <mach/vm_param.h>
#include <mach/machine.h>
#include <sys/types.h>
#include <machine/pdc.h>
#include <machine/iomod.h>
#include <machine/machspl.h>
#include <machine/rpb.h>
#include <machine/nvm.h>

int boothowto = 0;
int bootdev   = 0;
#ifdef KGDB_LAN
int bootargc  = 0;
#endif
extern void	inittodr();
extern u_int mem_size;

void
machine_init()
{
	/*
	 * still needs a bit of work here :-)
	 * wait til you see configure...
	 */
	extern powerfail();
	(void)procitab(SPLPOWER, powerfail, 0, 0);

	/* XXX need to read timer before ASP reset in configure */
	inittodr();
	configure();

	get_root_device();
#ifdef	GPROF
	kmstartup();
#endif
}

configure()
{
	showbus();
	machine_conf();
	getdumpdev();
	pdc_console_init(0);
	autoconf();
}

machine_conf()
{
	/* XXX - just hacks up the table, doesn't really look at anything...*/
	machine_slot[0].is_cpu = TRUE;
	machine_slot[0].running = TRUE;

	machine_slot[0].cpu_type = 0;		/* XXX */
	machine_slot[0].cpu_subtype = 0;	/* XXX */

	machine_info.max_cpus = NCPUS;
	machine_info.avail_cpus = 1;
	machine_info.memory_size = mem_size;
}


/*
 * Really bogus dumpconf! We do this using a name, but we should eventually
 * change it to use config and the device number.
 * NOTE: savecore needs dumpdev, dumplo, dumpsize, and dumpmag.
 */

dev_t dumpdev = 0;
int   dumplo = 0;	    /* DEV_BSIZE offset from start of dump partition.*/
int   dumpsize = 0;	    /* NBPG size of dump. */
int   dumpoffset = -1;	    /* DEV_BSIZE offset from start of disk */
int   dumpmag = 0x8fca0101; /* Magic number for savecore */

#include <device/conf.h>
#include <machine/machparam.h>
#include "sd.h"
dev_ops_t dumpdev_ops;

getdumpdev()
{
	int		dev_minor;
	extern char 	*dump_name;

	/*
	 * Get the device and unit number from the name.
	 */
	if (!dev_name_lookup(dump_name, &dumpdev_ops, &dev_minor)) {
		printf("Can't find dump device entry: %s\n", dump_name);
		return;
	}
	dumpdev = dev_minor;
	dumpsize = btoc(mem_size);
}

#include <mach_glue.h>
#include <sys/ioctl.h>
#include <device/device_types.h>
#include <device/disk_status.h>

dumpconf()
{
	struct partinfo pi;

	/*
	 * The device should already be opened, and the dump info setup,
	 * but savecore needs a bit more.
	 */
	(*dumpdev_ops->d_setstat)(dumpdev, DIOCGPART, (int *)&pi, sizeof(pi));
	dumplo = pi.part->p_size - ctod(dumpsize);

	printf("dumpconf: ");
	if (dumpoffset == -1)
		printf("WARNING no dump partition found\n");
	else {
		printf("%d blocks starting at *disk* block %d\n",
		       ctod(dumpsize), dumpoffset);
		if (dumpsize != btoc(mem_size))
			printf("\tWARNING: dump partition is smaller than memory size\n");
	}
}

#include <mach/vm_prot.h>
#include <vm/pmap.h>
#include <mach/time_value.h>
#include <machine/machparam.h>	/* for btop */

/*ARGSUSED*/
timemmap(dev, off, prot)
	vm_prot_t prot;
{
	extern time_value_t *mtime;

	if (prot & VM_PROT_WRITE)
		return (-1);

	return (btop(pmap_extract(pmap_kernel(), (vm_offset_t) mtime)));
}

/*
 * XXX - Put this here until we get it into movc.s
 */

void
ovbcopy(src, dst, len)
	register char *src, *dst;
	register int len;
{
	if (len < 0)
		return;

	/* 
	 * use the lightning fast bcopy() if at all possible 
	 */

	if (src > dst || dst > src+len)
		bcopy(src, dst, len);
	else {
		/* 
		 * use our tremendously slow backward byte copy 
		 */

		dst += (len - 1);
		src += (len - 1);
		while (len-- > 0)
			*dst-- = *src--;
	}

	return;
}

int dofastboot = 0;		/* if set, always do a fastboot() */
int dotocdump = 1;		/* if set, do a crash dump on TOC's */
struct rpb rpb;			/* Used by doadump and crashdump */

/*
 * I changed the machine-independent code to pass the flags down. LBS
 */
halt_all_cpus(options)
	int options;
{
	/*
	 * If the (undocumented) 0x800 flag is set ("reboot -f"),
	 * we try to load and execute the boot program ourselves.
	 * This can greatly reduce kernel development time!
	 */
	if ((options & RB_FASTBOOT) || dofastboot) {
		fcacheall();
		fastboot((options & RB_ASKNAME) ? 1 : 0);
		/*NOTREACHED*/
	}
	if (options & RB_DUMP) {
		dumpsys();
		/*NOTREACHED*/
	}
	if (options & RB_HALT) {
		printf("CPU halted\n");
		while(1);
	} else {
		printf("Rebooting MACH...\n");
		(*(struct iomod *)LBCAST_ADDR).io_command = CMD_RESET;
	}
}

extern char  *panicstr;
extern int   dumplo, dumpsize, dumpoffset;
extern dev_t dumpdev;

halt_cpu()
{
        static int paniced = 0;
#ifdef  KGDB
#ifdef	KGDB_LAN
	kgdb_break();
#else
	{
		extern int kgdb_dev;

                if (kgdb_dev != -1) {
                        delay(1000000);
                        kgdb_break();
                }
	}
#endif
#endif	/* KGDB */
        if (panicstr && !paniced) {
                paniced = 1;
                savestk();
                dispstk();
#if 0
		/* done in halt_all_cpus */
		dumpsys();
#endif
        }
        halt_all_cpus(panicstr ? RB_DUMP : RB_AUTOBOOT);
}

dumpsys()
{
#ifdef PDCDUMP
	if (dumpoffset < 0) {
		printf("Sorry, not enough room on dumpdev for dump\n");
		return;
	}
	printf("\nPDC dumping %d blocks to device ", ctod(dumpsize));
	printf("0x%x at *disk* offset %d blocks\n", dumpdev, dumpoffset);
	doadump();
	/*NOTREACHED*/		
#else
	printf("Sorry, crash dumps not supported\n");
#endif
}

/*
 * Purge TLBs.
 * This is the architected algorithm.
 */
ptlball(pdc_cache)
        struct pdc_cache *pdc_cache;
{
        register pa_space_t sp;
        register caddr_t off;
	register int six, oix, lix;
	int sixend, oixend, lixend;

	/* instruction TLB */
        sixend = pdc_cache->it_sp_count;
        oixend = pdc_cache->it_off_count;
        lixend = pdc_cache->it_loop;
	sp = pdc_cache->it_sp_base;
	for (six = 0; six < sixend; six++) {
		off = (caddr_t) pdc_cache->it_off_base;
		for (oix = 0; oix < oixend; oix++) {
			for (lix = 0; lix < lixend; lix++)
				pitlbentry(sp, off);
			off += pdc_cache->it_off_stride;
		}
		sp += pdc_cache->it_sp_stride;
	}

	/* data TLB */
        sixend = pdc_cache->dt_sp_count;
        oixend = pdc_cache->dt_off_count;
        lixend = pdc_cache->dt_loop;
	sp = pdc_cache->dt_sp_base;
	for (six = 0; six < sixend; six++) {
		off = (caddr_t) pdc_cache->dt_off_base;
		for (oix = 0; oix < oixend; oix++) {
			for (lix = 0; lix < lixend; lix++)
				pdtlbentry(sp, off);
			off += pdc_cache->dt_off_stride;
		}
		sp += pdc_cache->dt_sp_stride;
	}
}

int itick_per_tick = 0;
u_int itick_per_n_usec = 0;
u_int n_usec = 0;

delay(usec)
{
	u_long start, end, mfctl();
	u_long max_n_usec, howlong;

	/* max # of `itimer_n_usec's before 32-bit register overflows */
	max_n_usec = (unsigned) 2000000000/itick_per_n_usec;	/* ~2^31 */

	do {
		/* determine how long to wait during this iteration */
		howlong = (usec > max_n_usec)? max_n_usec: usec;

		start = mfctl(CR_ITMR);
		end = start + ((itick_per_n_usec * howlong) / n_usec);

		/* N.B. Interval Timer may wrap around */
		if (end < start)
			while (mfctl(CR_ITMR) > end)
				;
		while (mfctl(CR_ITMR) < end)
			;

		usec -= howlong;
	} while (usec);
}

#include <machine/pdc.h>

/*
 * XXX: Stolen from HPUX.
 *
 * Pick good values of itick_per_n_usec and n_usec so that 
 * itick_per_n_usec/n_usec is a good approximation to the number
 * of iticks (determined by processor frequency) in a microsecond.
 * Previous use of 12 to approximate 12.5 caused gettimeofday() to
 * run backwards occasionally!
 * The numbers must be fairly small so that calculations with them
 * (e.g. converting iticks to usecs) won't cause integer overflow.
 */

delay_init()
{
    int min_err = 2000000;	/* a big initial value */
    int denom, num;
    int error;			/* current error */
    int max_denom;
    int	tick = 1000000 / HZ;

    itick_per_tick = (PAGE0->mem_10msec * 100) / HZ;

    /* Make sure that our denom won't cause overflow */
    max_denom = MIN(1000, 1000000000/itick_per_tick);

    /*
     * Compute good values of itick_per_n_usec and n_usec, such
     * that itick_per_n_usec/n_usec equals or is slightly greater than
     * itick_per_tick/10000 .  If we allowed it to be less than the
     * correct value, conversions from ticks to usecs would be slightly
     * high, and time might run backwards across 10ms ticks.  Catching
     * up slightly is preferable to that possibility.
     * 
     * The values produced by this code should produce conversions
     * accurate to about 1usec up to clock frequencies of 100MHz,
     * and smoothly degrade beyond that point.  One billion is used
     * below instead of maxint to allow for the possibility of a lost
     * clock tick.  In the worst case (denom = 1000), it would correctly
     * handle periods of up to 21.47 ms since the last tick.
     * 
     * Tick is defined to be 1000000/HZ, or the number of usecs per tick.
     * 
     */
    n_usec = 1;			/* must have an initial value */
    for (denom = 1; denom <= max_denom; denom++) {
	/* Given denom, compute corresponding value of num (Round up). */
	num = ((denom*itick_per_tick) + (tick -1)) / tick;
	error = num * tick - itick_per_tick * denom;
	if (error == 0) {	    /* Exact result; use it. */
	    n_usec = denom;
	    itick_per_n_usec = num;
	    break;
	}

	if (error < 0) {
	    panic("set_itick_per_n_usec: arithmetic doesn't work!");
	}

	/* Rounding up earlier guaranteed that error >= 0 . */
	/* Find smallest percent error, i.e. error/denom < min_err/n_usec */
	if (error * n_usec < min_err * denom) {
		min_err = error;
		n_usec = denom;
		itick_per_n_usec = num;
	}
    }
}

void
kern_print(type, ssp)
	int type;
	struct parisc_saved_state *ssp;
{
	char buffer[1024];
	int len;

	len = ssp->arg1;
	if (len < 0)
		cnputc(ssp->arg0);
	else if (len) {
		if (len >= sizeof(buffer))
			len = sizeof(buffer) - 1;
		copyin(ssp->arg0, buffer, len);
		buffer[len] = 0;
		printf("%s", buffer);
	}

	/*
	 * move the pc past the break instruction
	 */
	
	ssp->iioq_head = ssp->iioq_tail;
	ssp->iioq_tail = ssp->iioq_head + 4;
}

/*
 *  dopanic() will screw up the stack in such a way that adb will be
 *  unable to show us a stack back trace, so we do one here.  A trace
 *  of the current stack is stored in `stktrc', with the number of
 *  entries in `stkcnt'.
 */

#define	STKSIZ	64

struct {
	u_int	pc;
	u_int   fp;
	u_int	ssp;
} stktrc[STKSIZ] = { 0 };

u_int	stkcnt = 0;

#define	strippriv(pc)	(int *)((u_int)(pc) & ~3)	/* no priv bits */

savestk()
{
	register int *pc, *fp;				/* current pc & sp */
	int *getpc(), *getfp();
	struct parisc_saved_state *ssp;
#if 1
	int *high, *low;
#endif

	if (stkcnt >= STKSIZ)
		return;

	pc = strippriv(getpc());
	fp = strippriv(getfp());
#if 1
	high = fp;
	low = (int *)(trunc_page(fp) - 2 * PAGE_SIZE);
#endif

	stktrc[stkcnt].pc   = (u_int)pc;
	stktrc[stkcnt++].fp = (u_int)fp;
	while (stkcnt < STKSIZ) {
#if 1
		/*
		 * Don't switch stacks
		 */
		if (fp <= low || fp > high)
			break;
#else
		if (fp == 0) {
			/*
			 * Look at the saved state, and see if we can keep
			 * going by switching to another kernel stack.
			 */
			fp = (int *)stktrc[stkcnt-2].fp;
			ssp = (struct parisc_saved_state *) *(fp - 3);;
			if ((ssp->flags & SS_PSPKERNEL) == 0) {
				printf("savestk: no prev context (%x)\n", fp);
				break;
			}
			pc = strippriv(ssp->iioq_head);
			fp = (int *)ssp->r4;
			stktrc[stkcnt].ssp = (u_int)ssp;
		}
#endif
		else {
			/*
			 * Find the previous frame and return pointer.
			 */
			pc = strippriv(*(fp - 5));
			fp = (int *) *fp;
		}
		stktrc[stkcnt].pc   = (u_int)pc;
		stktrc[stkcnt++].fp = (u_int)fp;
	}
}

/*
 *  Display `stkcnt' elements of `stktrc' on the console.
 */

dispstk()
{
	static char hex[] = "0123456789abcdef";
	register char *cp1, *cp2;
	register int i, j;
	char stkstr[9];

	printf("\nStack Trace (depth=%d):", stkcnt);

	for (i = 0; i < stkcnt; i++) {

		/* format ethernet address into stkstr[] */
		cp1 = (char *)&stktrc[i].pc;
		cp2 = stkstr;
		for (j = 0; j < 4; j++) {
			*cp2++ = hex[*cp1 >> 4 & 0xf];
			*cp2++ = hex[*cp1++ & 0xf];
		}
		*cp2 = '\0';

		/* and splat it to the console */
		printf("%s\t0x%s", (i%4)? "": "\n", stkstr);
	}

	printf("\nEnd of Stack Trace\n");

	/* wait a couple seconds for messages to get out */
	delay(2000000);
}
