/* 
 * Copyright (c) 1990,1991,1992,1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: pcb.c 1.36 95/03/02$
 */

#include <mach_debug.h>
#include <kern/task.h>
#include <kern/thread.h>
#include <mach/thread_status.h>
#include <machine/machspl.h>
#include <machine/psl.h>
#include <vm/vm_kern.h>
#include <kern/mach_param.h>
#include <kern/current.h>

/*
 * These constants are dumb. They should not be in asm.h!
 */
#define ARG_SIZE	(12*4)
#define FM_SIZE		32

#ifdef MACH_COUNTERS
int   fpu_trap_count = 0;
int   fpu_switch_count = 0;
#endif



#include "ipc_space.h"
#include "ipc_port.h"

/* Pool of kernel activations.
   XXX shouldn't be global;
   at the very least, the pager should have a different pool.  */
struct ipc_port *kernel_port;

/* Physical address of current activation's PCB.
   This is the glue code's starting-point notion of self:
   from it can be found the activation, thread, task, etc.
   This should be moved to a control register (cr31 maybe) if one can be freed up.  */
struct PCB *cur_pcb;

/* This points to the PCB currently owning the FPU.
   If 0, then the FPU contains no valid state.  */
struct PCB *fpu_pcb;

#ifdef TRAP_STACK_CHECK
#define START_PAD	0x1000
#define END_PAD		0x1000
#else
#define START_PAD	0
#define END_PAD		0
#endif
#define PADDED_KERNEL_STACK_SIZE	(KERNEL_STACK_SIZE+START_PAD+END_PAD)

#define STACK_OFS	(START_PAD + FM_SIZE + 128) /* XXX ARG_SIZE */

#define KSTACKS 150
static char kstack_buf[KSTACKS*PADDED_KERNEL_STACK_SIZE+PARISC_PGBYTES];
static char *kstacks;
int kstacknum = 0;
#if 1
int watchacts;
#endif


/*
 *	The next field is at the base of the stack,
 *	so the low end is left unsullied.
 */

void thread_machine_init()
{
}

/* Initialize the machine-dependent state for a new thread.
   At this point, we know what the initial activation for the thread will be,
   but the activation has not yet been attached to the thread.  */
kern_return_t thread_machine_create(thread_t thread, Act *act, __dead void (*start_position)())
{
	if (start_position)
	{
		/* Starting a kernel thread */
		extern __dead void thread_machine_start_self();
		act->mact.pcb->ss.bsd_goto = (int)thread_machine_start_self;
		act->mact.pcb->ss.r10 = (int)start_position;
		act->mact.pcb->ss.r30 = act->mact.user_stack;
	}
	else
	{
		extern void thread_machine_start_self_user();

		/* Starting a user thread */
		act->mact.pcb->ss.bsd_goto = (int)thread_machine_start_self_user;
		*(int *)&act->mact.pcb->fs.fr0 = -1;
	}

	return KERN_SUCCESS;
}

/*
 * Machine-dependent cleanup prior to destroying a thread.
 */
void
thread_machine_destroy(thread_t thread)
{
	Act *tmp_act, *top_act = thread->top_act;

	/* A hack.  We don't even check that top_act is a kernel
	   activiation!  */
	tmp_act = ((struct ipc_target *)kernel_port)->ipt_acts;
	((struct ipc_target *)kernel_port)->ipt_acts = top_act;
	top_act->ipt_next = tmp_act;
}

/*
 * This is used to set the current thread,
 * when starting up a new processor.
 */
void
thread_machine_set_current(thread_t thread)
{
	cur_pcb = thread->top_act->mact.pcb;
}

kern_return_t
port_machine_set_rpcinfo(target, rpc_info, rpc_info_count)
	ipc_target_t target;
	rpc_info_t *rpc_info;
	unsigned int rpc_info_count;
{
	if (rpc_info_count != PARISC_RPC_INFO_COUNT)
		return KERN_INVALID_ARGUMENT;

#if 0
	printf("set port %08x entrypoint %08x dataptr %08x\n",
	       object, rpc_info->entry_pc, rpc_info->entry_dp);
#endif
	target->ipt_rpcinfo.entry_pc = rpc_info->entry_pc | PC_PRIV_USER;
	target->ipt_rpcinfo.entry_dp = rpc_info->entry_dp;

	return KERN_SUCCESS;
}

void
ipc_target_machine_init(target)
	ipc_target_t target;
{
	target->ipt_rpcinfo.entry_pc = 0;
	target->ipt_rpcinfo.entry_dp = 0;
}

/*
 *	stack_collect:
 *
 *	Free excess kernel stacks.
 *	May block.
 */
void stack_collect()
{
	/* XXX ??? */
}




#if	MACH_DEBUG
extern boolean_t stack_check_usage;

/*
 *	stack_statistics:
 *
 *	Return statistics on cached kernel stacks
 *	kept by this machine-dependent module.
 *	*maxusagep must be initialized by the caller.
 */
void stack_statistics(totalp, maxusagep)
	unsigned int *totalp;
	vm_size_t *maxusagep;
{
	*totalp = *maxusagep = 0;
	/*XXX*/
}
#endif	MACH_DEBUG



#ifdef DEBUG
dump_ss(struct parisc_saved_state *ss)
{
	printf("ss @ %8.8x:\n", ss);

	printf("rp  %8.8x	r12  %8.8x	sr3   %8.8x	cr0   %8.8x\n",
	       ss->r2, ss->r12, ss->sr3, ss->cr0);
	printf("r3  %8.8x	r13  %8.8x	sr4   %8.8x	pidr1 %8.8x\n",
	       ss->r3, ss->r13, ss->sr4, ss->cr8);
	printf("r4  %8.8x	r14  %8.8x	sr5   %8.8x	pidr2 %8.8x\n",
	       ss->r4, ss->r14, ss->sr5, ss->cr9);
	printf("r5  %8.8x	r15  %8.8x	sr6   %8.8x	sar   %8.8x\n",
	       ss->r5, ss->r15, ss->sr6, ss->cr11);
	printf("r6  %8.8x	r16  %8.8x	sr7   %8.8x	pidr3 %8.8x\n",
	       ss->r6, ss->r16, ss->sr7, ss->cr12);
	printf("r7  %8.8x	r17  %8.8x			pidr4 %8.8x\n",
	       ss->r7, ss->r17, ss->cr13);
	printf("r8  %8.8x	r18  %8.8x			eiem  %8.8x\n",
	       ss->r8, ss->r18, ss->cr15);
	printf("r9  %8.8x	dp   %8.8x\n",
	       ss->r9, ss->r27);
	printf("r10 %8.8x	sp   %8.8x			psw   %8.8x\n",
	       ss->r10, ss->r30, ss->ipsw);
	printf("r11 %8.8x					ksp   %8.8x\n",
	       ss->r11, 0xbadf00d);

}

dump_thread(th)
	thread_t th;
{
	panic("dump_thread");
#if 0
	printf(" thread @ 0x%x:\n", th);
 
	printf("  ss  %8.8x, state   %8.8x, ksp     %8.8x, ksp  %8.8x\n",
	       th->top_act->mact.ss, th->state, th->kernel_act.user_stack, th->thread.mthread.kernel_sp);

	printf("  message %8.8x, wevent  %8.8x\n",
	       th->thread.message, th->wait_event);
#endif
}
#endif


void kact_count()
{
	int i;
	Act *act;
	static int amin = KSTACKS;

	i = 0;
	for (act = kernel_port->ip_target.ipt_acts; act; act = act->ipt_next)
		i++;
	if (i < amin)
		amin = i;
	printf("%d of %d kernel activations in use, %d max\n",
	       KSTACKS-i, KSTACKS, KSTACKS-amin);
}

void act_machine_init()
{
	int i;
	Act *act;
	ipc_target_t target;
	void migr_rpc_entry();

	kernel_port = ipc_port_alloc_kernel();
	if (kernel_port == IP_NULL)
		panic("act_machine_init");

	kstacks = (char *) round_page(kstack_buf);
	target = &kernel_port->ip_target;

	for (i = 0; i < KSTACKS-25; i++) {
		if (act_create_kernel(&act))
			panic("can't create kernel activation");
		act_set_target(act, target);
	}
#ifdef USENRPC
	target->ipt_rpcinfo.entry_pc = (vm_offset_t) migr_rpc_entry;
	target->ipt_rpcinfo.entry_dp = getdp();
#endif

#ifdef TRAP_STACK_CHECK
	for (i = 0; i < KSTACKS*PADDED_KERNEL_STACK_SIZE/4; i++)
		*((int*)kstacks + i) = 0xc01df00d;
#endif

#if 0
	watchacts = 1;
#endif
}

kern_return_t
act_machine_create(task_t task, Act *act, vm_offset_t user_stack, vm_offset_t user_rbuf, vm_size_t user_rbuf_size)
{
	act->mact.space = task->map->pmap->space;
	act->mact.pid = task->map->pmap->pid;
#if 0
	printf("act_machine_create %08x space %08x pid %08x\n", act, act->mact.space, act->mact.pid);
#endif
	act->mact.pcb = &act->mact.xxx_pcb;
	act->mact.user_stack = (user_stack + 63) & ~63;
	act->mact.user_rbuf = user_rbuf;
	act->mact.user_rbuf_size = user_rbuf_size;

	/*
	 * Clear the PCB to avoid random values in the registers.
	 */
	bzero(act->mact.pcb, sizeof(struct PCB));

#if 0
	act->mact.return_routine = 0;
	act->mact.reentry_routine = 0;
#endif

	/*
	 * Initialize the back pointer to the activation.
	 */
	act->mact.pcb->ss.act = act;

	/*
	 * User threads will pull their context from the pcb when first
	 * returning to user mode, so fill in all the necessary values.
	 */
	act->mact.pcb->ss.flags = SS_INTRAP;
	act->mact.pcb->ss.sr4   = (int) act->mact.space;
	act->mact.pcb->ss.sr5   = (int) act->mact.space;
	act->mact.pcb->ss.sr6   = (int) act->mact.space;
	act->mact.pcb->ss.cr8   = act->mact.pid;

	act->mact.pcb->ss.cr15  = SPL0;
	act->mact.pcb->ss.cr22  = PSW_C | PSW_Q | PSW_P | PSW_D | PSW_I;

	act->mact.pcb->ss.iisq_head = (int) act->mact.space;
	act->mact.pcb->ss.iisq_tail = (int) act->mact.space;
	act->mact.pcb->ss.iioq_head = (int) PC_PRIV_USER;
	act->mact.pcb->ss.iioq_tail = (int) PC_PRIV_USER;

	return KERN_SUCCESS;
}

kern_return_t
act_create_kernel(Act **out_act)
{
	int rc;
	vm_offset_t rbuf;
	vm_size_t rbuf_size = 8192;	/* XXX */

	if (kstacknum >= KSTACKS)
		panic("out of kernel stacks");

	/* XXX hmm...can this always be pageable? */
	if (kmem_alloc_pageable(kernel_map, &rbuf, rbuf_size))
		panic("no kernel AS for activation!");

	if (rc = act_create(kernel_task,
		(vm_offset_t)kstacks
			+ PADDED_KERNEL_STACK_SIZE*kstacknum++
			+ START_PAD
			+ STACK_OFS, rbuf, rbuf_size, out_act))
		return rc;
#if 0
	if (watchacts)
		printf("a_c_k: %d in use\n", kstacknum);
#endif
	return KERN_SUCCESS;
}

void act_machine_destroy(Act *act)
{
	/* We don't need to do anything,
	   since we didn't allocate memory or anything like that.  */
}

__dead void act_machine_return(int code)
{
	Act *act = current_act();
	Act *tmp_act;

	act_deallocate(act);		/* XXX thread reference */
	act_deallocate(act);		/* XXX kernel activation reference */
	thread_terminate_self();	/* XXX */
}

/* Nudge the target activation,
   which may be on the same or a different thread.
   It cannot be the current (kernel) activation.
   Rolls the activation out of any two-way optimized RPC paths,
   allowing the thread's activation chain to be broken
   above this activation.
   Also ensures that any current returnhandlers will be seen
   before the target activation next gets a chance to execute.

   Already locked: Act */
void
act_machine_nudge(Act *act)
{
}

kern_return_t   act_machine_set_state(Act *act, int flavor, int *tstate, unsigned count)


{
	register struct parisc_thread_state   *state;
	register struct parisc_saved_state    *saved_state;

	if (flavor != PARISC_THREAD_STATE ||
	    count < PARISC_THREAD_STATE_COUNT)
		return KERN_INVALID_ARGUMENT;

	state = (struct parisc_thread_state *) tstate;
	saved_state = (struct parisc_saved_state *)&act->mact.pcb->ss;

	saved_state->flags = state->flags;

	/*
	 * allow the user to change any of the general registers
	 */

	saved_state->r1 = state->r1;
	saved_state->r2 = state->r2;
	saved_state->r3 = state->r3;
	saved_state->r4 = state->r4;
	saved_state->r5 = state->r5;
	saved_state->r6 = state->r6;
	saved_state->r7 = state->r7;
	saved_state->r8 = state->r8;
	saved_state->r9 = state->r9;
	saved_state->r10 = state->r10;
	saved_state->r11 = state->r11;
	saved_state->r12 = state->r12;
	saved_state->r13 = state->r13;
	saved_state->r14 = state->r14;
	saved_state->r15 = state->r15;
	saved_state->r16 = state->r16;
	saved_state->r17 = state->r17;
	saved_state->r18 = state->r18;
	saved_state->r19 = state->r19;
	saved_state->r20 = state->r20;
	saved_state->r21 = state->r21;
	saved_state->r22 = state->r22;
	saved_state->r23 = state->r23;
	saved_state->r24 = state->r24;
	saved_state->r25 = state->r25;
	saved_state->r26 = state->r26;
	saved_state->r27 = state->r27;
	saved_state->r28 = state->r28;
	saved_state->r29 = state->r29;
	saved_state->r30 = state->r30;
	saved_state->r31 = state->r31;

	saved_state->sr0 = state->sr0;
	saved_state->sr1 = state->sr1;
	saved_state->sr2 = state->sr2;
	saved_state->sr3 = state->sr3;
/*	saved_state->sr4 = state->sr4;*/
/* XXX This _should_ be user-modifiable, because the hardware register isn't even protected,
   but the Unix server currently depends on it _not_ being modifiable.  */

	saved_state->cr11 = state->cr11;

	/*
	 * Make sure IIOQ is in user mode
	 */
	saved_state->iioq_head = state->iioq_head | PC_PRIV_USER;
	saved_state->iioq_tail = state->iioq_tail | PC_PRIV_USER;

	/*
	 * Make sure only PSW_T, PSW_X, PSW_N, PSW_B, PSW_V and PSW_CB 
	 * bits are set by users.
	 */
	saved_state->cr22 = state->cr22 & 
		(PSW_T | PSW_X | PSW_N | PSW_B | PSW_V | PSW_CB);

	/*
	 * Software managed single stepping: We define a software bit in the
	 * PSW that allows a user process to single step. When that bit is
	 * set, we turn on the recovery counter, and set the count to 0
	 * so that 1 instruction is executed before trapping. We use space
	 * in the PCB to record that single step mode is on so that getstatus
	 * returns the software bit, and so that trap does not confuse it
	 * with an AST (temporary problem, I hope).
	 */
	if (state->cr22 & PSW_SS) {
#if 0
		printf("pcb: activating single-step\n");
#endif
		act->mact.sstep = 1;
		saved_state->cr0   = 0;
		saved_state->cr22  |= PSW_R;
	}
	else {
		act->mact.sstep = 0;
		saved_state->cr22  &= ~PSW_R;
	}

	/*
	 * Always make sure that PSW_C, PSW_Q, PSW_P, PSW_D are set. The PSW_I
	 * bit is set according to eiem at context switch.
	 */
	saved_state->cr22 |= PSW_C | PSW_Q | PSW_P | PSW_D | PSW_I;

	/*
	 * Copy FP state.  Use bcopy to avoid alignment issues.
	 */
	if (fpu_pcb == act->mact.pcb)
		fpu_pcb = 0;
	bcopy((char *)&state->fr0, (char *)&act->mact.pcb->fs,
	      sizeof(act->mact.pcb->fs));

	return KERN_SUCCESS;
}

kern_return_t   act_machine_get_state(Act *act, int flavor, int *tstate, unsigned *count)
{

	register struct parisc_thread_state   *state;
	register struct parisc_saved_state    *saved_state;

	switch (flavor) {

	case THREAD_STATE_FLAVOR_LIST:

		if (*count < 1)
			return KERN_INVALID_ARGUMENT;

		tstate[0] = PARISC_THREAD_STATE;
		*count = 1;

		break;

	case PARISC_THREAD_STATE:

		if (*count < PARISC_THREAD_STATE_COUNT)
			return KERN_INVALID_ARGUMENT;

		state = (struct parisc_thread_state *) tstate;
		saved_state = (struct parisc_saved_state *) &act->mact.pcb->ss;

		bzero((char *)state, sizeof(struct parisc_thread_state));
		state->flags = saved_state->flags;
		state->r1  = saved_state->r1;
		state->r2  = saved_state->r2;
		state->r3  = saved_state->r3;
		state->r4  = saved_state->r4;
		state->r5  = saved_state->r5;
		state->r6  = saved_state->r6;
		state->r7  = saved_state->r7;
		state->r8  = saved_state->r8;
		state->r9  = saved_state->r9;
		state->r10 = saved_state->r10;
		state->r11 = saved_state->r11;
		state->r12 = saved_state->r12;
		state->r13 = saved_state->r13;
		state->r14 = saved_state->r14;
		state->r15 = saved_state->r15;
		state->r16 = saved_state->r16;
		state->r17 = saved_state->r17;
		state->r18 = saved_state->r18;
		state->r19 = saved_state->r19;
		state->r20 = saved_state->r20;
		state->r21 = saved_state->r21;
		state->r22 = saved_state->r22;
		state->r23 = saved_state->r23;
		state->r24 = saved_state->r24;
		state->r25 = saved_state->r25;
		state->r26 = saved_state->r26;
		state->r27 = saved_state->r27;
		state->r28 = saved_state->r28;
		state->r29 = saved_state->r29;
		state->r30 = saved_state->r30;
		state->r31 = saved_state->r31;

		state->sr0 = saved_state->sr0;
		state->sr1 = saved_state->sr1;
		state->sr2 = saved_state->sr2;
		state->sr3 = saved_state->sr3;
		state->sr4 = saved_state->sr4;
		state->sr5 = saved_state->sr5;
		state->sr6 = saved_state->sr6;
		state->sr7 = saved_state->sr7;

		state->cr0 = saved_state->cr0;
		state->cr11 = saved_state->cr11;

		state->iioq_head = saved_state->iioq_head;
		state->iisq_head = saved_state->iisq_head;
		state->iioq_tail = saved_state->iioq_tail;
		state->iisq_tail = saved_state->iisq_tail;

		/* 
		 * Only export the meaningful bits of the psw
		 */
		state->cr22 = saved_state->cr22 & 
			(PSW_X | PSW_T | PSW_N | PSW_B | PSW_V | PSW_CB);

		if (act->mact.sstep)
			state->cr22 |= PSW_SS;
		/*
		 * Copy FP state.  Use bcopy to avoid alignment issues.
		 * If this act's state is in the FPU, flush it first.
		 * If this act's state is undefined, just supply zeros.
		 */
		if (fpu_pcb == act->mact.pcb)
			fpu_flush();
		if (*((int*)&act->mact.pcb->fs.fr0) == -1)
			bzero((char*)&state->fr0, sizeof(double) * 32);
		else
			bcopy((char *)&act->mact.pcb->fs, (char *)&state->fr0,
			      sizeof(double) * 32);

		*count = PARISC_THREAD_STATE_COUNT;
		break;

	default:
		return KERN_INVALID_ARGUMENT;

	}

	return KERN_SUCCESS;
}

void panic_no_kstacks()
{
	/*XXX*/
	dump_act(cur_pcb->ss.act);
	dump_ssp(&cur_pcb->ss);
	panic("no kernel stacks available on trap\n");
}

#ifdef TRAP_STACK_CHECK
static int max_trap_stack_size = 0;

int trap_stack_check(int ret)
{
	vm_offset_t stack = cur_pcb->ss.act->mact.user_stack - STACK_OFS;
	vm_offset_t sp, p, q;

	/* Find the current stack pointer.  */
	asm("copy %%r30,%0" : "=r" (sp));

	/* Make sure it's in range.  */
	if ((sp < stack + STACK_OFS) || (sp >= stack+PADDED_KERNEL_STACK_SIZE-END_PAD))
		panic("stack pointer %08x out of bounds in trap_stack_check", sp);

	/* Search backward for the last dirtied word in the stack.  */
	for (p = stack+PADDED_KERNEL_STACK_SIZE-4; p >= stack; p -= 4)
		if (*((int*)p) != 0xc01df00d)
			break;
#if 0
	printf("trap_stack_check: stack %08x sp %08x used %08x max %08x\n",
		stack, sp, p, stack+KERNEL_STACK_SIZE);
#endif
	if (p - stack - START_PAD > max_trap_stack_size)
	{
		max_trap_stack_size = p - stack - START_PAD;
#if 0
		printf("trap_stack_check: new max stack size: %08x (%d)\n",
			max_trap_stack_size, max_trap_stack_size);
#endif
	}
	if (p >= stack+PADDED_KERNEL_STACK_SIZE-4)
		panic("trap_stack_check: MAJOR stack overflow");
	if (p >= stack+PADDED_KERNEL_STACK_SIZE-END_PAD-4)
		panic("trap_stack_check: stack overflow");

	/* Reinitialize the rest of the stack for the next time around.  */
	for (q = sp; q <= p; q += 4)
		*((int*)q) = 0xc01df00d;

	/* Make sure the start padding didn't get changed.  */
	for (p = stack; p < START_PAD; p += 4)
		if (*((int*)p) != 0xc01df00d)
			panic("trap_stack_check: stack underflow");

	return ret;
}
#endif

#ifdef USENRPC

#include <mach/mig_errors.h>

void
kern_act_entry(InP, OutP)
	rpc_message_t *InP, *OutP;
{
	mig_routine_t routine;
	extern mig_routine_t LRmach_server_routine(),
			     LRmach_port_server_routine();

#if 1
	static int count = 50;
	if (--count > 0)
		printf("%d: k_a_e(%x): id=%d, port=0x%x, seq=0x%x\n",
		       count, InP,
		       InP->Head.rpch_id,
		       InP->Head.rpch_port,
		       InP->Head.rpch_seqno);
#endif

	/* Default values for outgoing message */
	OutP->Head.rpch_port = MACH_PORT_NULL;
	OutP->Head.rpch_id = InP->Head.rpch_id + 100;
	OutP->Head.rpch_arg2 = 0;
	OutP->Head.rpch_arg3 = 0;
	OutP->PortName = MACH_PORT_NULL;
	OutP->MemAddr = InP->MemAddr;
	OutP->MemSize = 0;

	if ((routine = LRmach_server_routine(&InP->Head)) != 0
	 || (routine = LRmach_port_server_routine(&InP->Head)) != 0) {
		(*routine)(InP, OutP);
	} else {
		printf("k_a_e(%x): id=%d, port=0x%x, seq=0x%x: failed!\n",
		       InP,
		       InP->Head.rpch_id,
		       InP->Head.rpch_port,
		       InP->Head.rpch_seqno);
		OutP->Direct[0] = MIG_BAD_ID;
	}
}
#endif
