/* 
 * Copyright (c) 1990, 1991, 1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: pdc_console.c 1.5 94/12/14$
 */

/*
 *	Routines to use as a console device the PDC left over
 *	from the boot program.
 */

#include <machine/pdc.h>
#include <machine/iodc.h>
#include <machine/iomod.h>

extern int vm_on;
extern int pdc_alive;

/*
 * All of these pointers are based on what the boot program left around for
 * us to use.
 */
#define	KY_HPA		PAGE0->mem_kbd.pz_hpa
#define	KY_SPA		PAGE0->mem_kbd.pz_spa
#define	KY_LAYER	PAGE0->mem_kbd.pz_layers
#define	KY_IODC		PAGE0->mem_kbd.pz_iodc_io
#define	KY_CLASS	PAGE0->mem_kbd.pz_class

#define	CN_HPA		PAGE0->mem_cons.pz_hpa
#define	CN_SPA		PAGE0->mem_cons.pz_spa
#define	CN_LAYER	PAGE0->mem_cons.pz_layers
#define	CN_IODC		PAGE0->mem_cons.pz_iodc_io
#define	CN_CLASS	PAGE0->mem_cons.pz_class

#define	MINIOSIZ	64	        /* minimum buffer size for IODC call */
static char cnbuf[MINIOSIZ * 2];
static char *cndata = 0;

static int *pdcret = 0;
static int pdcbuf[33];
static int (*cniodc)() = (int (*)()) 0;	/* console IODC entry point */
static int (*kyiodc)() = (int (*)()) 0;	/* console IODC entry point */

void
pdc_console_init(turnon)
	int turnon;
{
	extern void pdcputc(), (*romputc)();
	extern int pdcgetc(), (*romgetc)();

	if (!turnon) {
		romputc = 0;
		romgetc = 0;
		return;
	}

	pdcret = (int *)((((int) pdcbuf) + 7) & ~7);
	cndata = (char *) ((((int) &cnbuf[0]) + MINIOSIZ-1) & ~(MINIOSIZ-1));

	/*
	 * The boot program loaded the IODC code for the console is at mem_free.
	 */
	cniodc = (int (*)()) PAGE0->mem_free;
	kyiodc = (int (*)()) PAGE0->mem_free;

	romputc = pdcputc;
	romgetc = pdcgetc;
}

/*
 * If virtual memory is off then we can just call the routines directly.
 * When VM is on, we have to use the routine pdc_iodc() which turns off
 * VM and puts us in a state that we can make the iodc call.
 */
int
pdcgetc(wait)
	int wait;
{
	int err;

	if (!vm_on)
		err = (*kyiodc)(KY_HPA, IODC_IO_CONSIN, KY_SPA, KY_LAYER,
				pdcret, 0, cndata, 1, 0);
	else
		err = pdc_iodc(kyiodc, 0, KY_HPA, IODC_IO_CONSIN, KY_SPA, 
			       KY_LAYER, pdcret, 0, cndata, 1, 0);

	if (err < 0)
		printf("KBD input error: %d\n", err);

	if (pdcret[0] == 0)
		return (0);

	if (pdcret[0] > 1)
		printf("KBD input got too much (%d)\n", pdcret[0]);

	return ((int)*cndata);
}

void
pdcputc(c)
	int c;
{
	*cndata = c;

	if (!vm_on)
		(void) (*cniodc)(CN_HPA, IODC_IO_CONSOUT, CN_SPA, CN_LAYER,
				 pdcret, 0, cndata, 1, 0);
	else
		pdc_iodc(cniodc, 0, CN_HPA, IODC_IO_CONSOUT, CN_SPA, CN_LAYER,
			 pdcret, 0, cndata, 1, 0);
}
