/* 
 * Copyright (c) 1990,1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: pim.h 1.4 94/12/14$
 *	Author: Jeff Forys, University of Utah CSL
 */

#ifndef	_PIM_
#define	_PIM_

/*
 * Processor Internal Memory structure.
 */

struct mchk_type {	/* reason for machine check */
	unsigned int	cache:	1,	/* cache system error */
			tlb:	1,	/* TLB system error */
			bus:	1,	/* bus transation error */
			assist:	1,	/* coprocessor or SFU error */
			proc:	1,	/* processor internal error */
			resv:	26,	/* (reserved) */
			advise:	1;	/* PDC advisory shutdown */
};

struct mchk {		/* High/Low Priority Machine Check information */
	struct mchk_type type;		/* general location of error */
	unsigned int	cpustate;	/* validity of processor state */
	unsigned int	detector;	/* type of instr that detected error */
	unsigned int	cache;		/* breakdown of `type.cache' */
	unsigned int	tlb;		/* breakdown of `type.tlb' */
	unsigned int	bus;		/* breakdown of `type.bus' */
	unsigned int	assist;		/* breakdown of `type.assist' */
	unsigned int	proc;		/* breakdown of `type.proc' */
	unsigned int	assist_id;	/* unit id of failing coproc or SFU */
	unsigned int	slave;		/* slave bus addr (iff `type.bus') */
	unsigned int	master;		/* master bus addr (iff `type.bus') */
	unsigned int	path_info;	/* transaction path (iff `type.bus') */
};

struct pim {		/* Processor Internal Memory */
	unsigned int	pim_gr[32];	/* 32 general registers (0x0) */
	unsigned int	pim_cr[32];	/* 32 control registers (0x80) */
	unsigned int	pim_sr[8];	/* 8 space registers (0x100) */
	unsigned int	pim_pcsqt;	/* pc space queue (tail) (0x120) */
	unsigned int	pim_pcoqt;	/* pc offset queue (tail) (0x124) */
	struct   mchk	pim_mchk;	/* HPMC or LPMC information (0x128) */
	double		pim_fp[32];	/* 32 floating pt registers (0x158) */
					/* End of PIM (0x258) */
};

#endif	/* _PIM_ */
