/* 
 * Copyright (c) 1990,1991,1992,1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: syscall.s 1.35 95/01/20$
 *	Author: Bob Wheeler, University of Utah CSL
 */

#include <machine/asm.h>
#include <assym.s>
#include <kern/syscall_emulation.h>
#include <mach/exception.h>
#include <machine/act.h>
#include <machine/pmap.h>
#include <machine/psl.h>
#include <machine/machspl.h>
#include <mach/machine/vm_param.h>


	.space	$PRIVATE$
	.subspa	$DATA$

	.import $global$,data
	.import	cur_pcb,data
	.import mach_trap_count,data
	.import mach_trap_table,data
	.import need_ast,data
	.import istackptr,data /* XXX */
	.import kernel_port,data

$syscall_panic
        .string "argument count for system call greater than 12\x00"

#ifdef SYSCALLSTAT
	.export syscallstats,data
	.align 32
syscallstats
	.word	0	/* +00 syscalls */
	.word	0	/* +04 emulated */
	.word	0	/* +08 non-mach_msg */
	.word	0	/* +12 mach_msg */
	.word	0		/* +16 bad option */
	.word	0		/* +20 bad bits */
	.word	0		/* +24 bad reply port */
	.word	0		/* +28 bad name */
	.word	0		/* +32 bad rights */
	.word	0		/* +36 no actpool */
	.word	0		/* +40 no activation */
	.word	0		/* +44 UX server exit calls */
	.word	0		/* +48 RPC */
	.word	0		/* +52 complex reply */
	.word	0		/* +56 complex request */
#endif

	.space	$TEXT$
	.subspa $CODE$

/*
 * syscall
 * 
 * Determine what kind of a system call we are dealing with and either handle
 * it internally or invoke an emulator for this call.
 */

	.align  PARISC_PGBYTES
	.export $syscall,entry
$syscall

#ifdef	KGDB
	/*
	 * get a stack for KGDB to run on to debug the syscall reflection code
	 */
	copy	sp,r29
	.import panic_stack,data
	ldil	L%panic_stack,t2
	ldo	R%panic_stack(t2),sp

	.export	kgdb_syscall_break,entry
kgdb_syscall_break
#endif	KGDB
	/*
	 * We can trash any caller saved register except r22 (t1) which is used
	 * to pass the system call number. We will need to save r31 and sr0
	 * before we call anything since they are caller saves registers.
	 */

#ifdef SYSCALLSTAT
	ldil	L%syscallstats,r1
	ldo	R%syscallstats(r1),r1
	ldw	0(sr1,r1),t2
	ldo	1(t2),t2
	stw	t2,0(sr1,r1)
#endif

#ifdef SYSCALL_PRINT
	.import last_syscall_num,data
	ldil	L%last_syscall_num,r1
	stw	r22,R%last_syscall_num(sr1,r1)
#endif

	/*
	 * Find the current activation (t2).
	 * Until we get the space registers set up we need to explicitly use sr1.
 	 */
	ldil	L%cur_pcb,r1
	ldw	R%cur_pcb(sr1,r1),t2
	ldw	PCB_ACT(sr1,t2),t2

#ifdef ACTWATCH
	/*
	 * Check activation at the entry to any syscall
	 * (emulated, message, trap, or migration).
	 * XXX stash syscall # so we can reload and dump later.
	 */
	stw	r22,ACT_SPARE+40(sr1,t2)
	break	0,11
#endif

	/*
	 * get the task that this thread belongs to and then get the pointer
	 * to the emulator dispatch table for the task
	 */
	ldw	ACT_TASK(sr1,t2),t2
	ldw	TASK_EML_DISPATCH(sr1,t2),t2

	/*
 	 * If the emulator dispatch table pointer is NULL then it's a Mach
	 * task; it can't call an emulator...
	 */
	comb,=,n r0,t2,$noemulation

	/*
	 * See if the system call number is less than the emulator DISP_MIN
	 * value. If so, its not emulated. Put the value r22 - DISP_MIN into
	 * t3. Thus t3 is the system call number relative to the smallest
	 * real system call allowed.
	 */
	ldw	DISP_MIN(sr1,t2),t4

	sub,>=	r22,t4,t3
	b,n	$noemulation

	/*
	 * get the count of system call vectors for this emulator. If the
	 * system call number we have is greater than the emulator's count
	 * don't emulate it
	 */
	ldw	DISP_COUNT(sr1,t2),t4
	comb,>,n t3,t4,$noemulation

	/*
	 * now check the emulator's vector table for a valid address. First 
	 * shift the value in t3 left by 2 to get a word count then use it
	 * as the index into the vector.
	 */
	ldo	DISP_VECTOR(t2),t4
	shd	t3,r0,30,t3
	add	t3,t4,t3

	/*
	 * get the contents of the vector location and if it is zero then
	 * this system call isn't emulated
 	 */
	ldw	0(sr1,t3),t4

	comb,=,n t4,r0,$noemulation

#ifdef SYSCALLSTAT
	ldil	L%syscallstats,r1
	ldo	R%syscallstats(r1),r1
	ldw	4(sr1,r1),t2
	ldo	1(t2),t2
	stw	t2,4(sr1,r1)
#endif
	/*
	 * get rid of the kernel PID from pidr4 and
	 * do a branch back to the user space only go to the place
	 * that the emulation vector is pointing to. sr0 and r31 point to
	 * the return point for the emulator library to use. We need to 
	 * set the privilege level to be user mode in t4.
	 */

	depi	PC_PRIV_USER,31,2,t4

#ifdef	KGDB
	copy	r29,sp
#endif	KGDB

	be	0(sr0,t4)
	mtctl	r0,pidr4


$noemulation
#ifdef	KGDB
	copy	r29,sp
#endif	KGDB

	/*
	 * This is a Mach system call;
	 * we need to handle it internally in the micro-kernel.
	 */

	/*
	 * Find the current PCB (ret1).
	 */
	ldil	L%cur_pcb,ret1
	ldw	R%cur_pcb(sr1,ret1),ret1

	/*
	 * Save the relevant user state.
	 */
	stw	r2,SS_R2(sr1,ret1)
#ifdef R4_IS_SPECIAL
	stw	r4,SS_R4(sr1,ret1)
#endif
	stw	r27,SS_R27(sr1,ret1)
	stw	r30,SS_R30(sr1,ret1)
	stw	r31,SS_R31(sr1,ret1)

	/*
	 * Set the syscall flag.
	 * This is used by various applications (via the get_state
	 * interface) to determine the correct "PC" (either iioqh
	 * or r31).
	 */
	ldi	SS_INSYSCALL,t2
	stw	t2,SS_FLAGS(sr1,ret1)

	/*
	 * Save the space registers so that we can get them pointing to the kernel.
	 * We need to save sr0 since it is a caller saves registers and it
	 * holds the space we need to return to.
	 * No need to save sr6-sr7 because they are constant and unchangeable to user code.
	 * sr5 must be saved since we change it.
	 */
	mfsp	sr0,t4
	stw	t4,SS_SR0(sr1,ret1)
	mfsp	sr4,t4
	stw	t4,SS_SR4(sr1,ret1)
#ifdef MULTIPLE_SPACES_PER_TASK
	mfsp	sr5,t4
	stw	t4,SS_SR5(sr1,ret1)
	mfsp	sr6,t4
	stw	t4,SS_SR6(sr1,ret1)
#endif

	/*
	 * This is the base interruption and encapsulation rollback point for system calls.
	 * Interruption rollback before this rolls forward to this point.
	 */

	/*
	 * Set the space registers to point to the kernel.
	 * We no longer have to explicitly use sr1 when accessing data.
	 */
	mtsp	r0,sr4
	mtsp	r0,sr5
	mtsp	r0,sr6

	/*
	 * Find the user activation (t3).
	 */
	ldw	PCB_ACT(ret1),t3

#ifdef ACTCHECK
	ldw		ACT_THREAD(t3),t4
	comb,<>,n	r0,t4,$tok
	break		0x1f,0x1
$tok
#endif

	/*
	 * Check for migrating RPC through mach_msg_trap.
	 * If it is, divert directly to the fast message path in msg.s.
	 */
	.import $mach_msg_trap_fast,code
	ldi	-25,t2
	comb,<>,n t2,t1,$not_mach_msg_trap
	ldil	L%$mach_msg_trap_fast,t2
	ldo	R%$mach_msg_trap_fast(t2),t2
	bv	0(t2)
	nop
$not_mach_msg_trap

#ifdef SYSCALLSTAT
	ldil	L%syscallstats,r1
	ldo	R%syscallstats(r1),r1
	ldw	8(r1),t2
	ldo	1(t2),t2
	stw	t2,8(r1)
#endif

	.export	$mach_msg_trap_fast_abort,entry
$mach_msg_trap_fast_abort

	/*
	 * Find a kernel activation to run on (t2).
	 */
	ldil	L%kernel_port,t4
	ldw	R%kernel_port(t4),t4
	ldw	IPT_ACTS(t4),t2
	comb,<>,n t2,r0,$kact_available

	/* XXX block */
	ldil	L%istackptr,t3
	ldw	R%istackptr(t3),sp
	stw	r0,R%istackptr(t3)
	.import panic_no_kstacks,code
	ldil	L%panic_no_kstacks,t2
	ldo	R%panic_no_kstacks(t2),t2
	bv	0(t2)
	ldo	FM_SIZE(sp),sp

$kact_available

	/*
	 * Remove it from the port.
	 */
	ldw	ACT_IPT_NEXT(t2),ret0
	stw	ret0,IPT_ACTS(t4)

	/*
	 * Attach the kernel activation (t2) to the current thread (now r1).
	 * The user activation is still in t3.
	 *
	 * XXX refcounts for the kernel activation.
	 */
	ldw	ACT_THREAD(t3),r1
	stw	t3,ACT_LOWER(t2)
	stw	t2,ACT_HIGHER(t3)
	stw	t2,THREAD_TOP_ACT(r1)
	stw	r1,ACT_THREAD(t2)

	/*
	 * Take a reference on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t3),t4
	ldo	1(t4),t4
	stw	t4,ACT_REF_COUNT(t3)
	stw	t3,ACT_USER_ACT(t2)

#ifdef ACTCHECK
	ldw		ACT_REENTRY_ROUTINE(t3),t4
	comb,=,n	r0,t4,$reis0
	ldw		ACT_PCB(t2),t3
	ldil		L%cur_pcb,t4
	stw		t3,R%cur_pcb(t4)
	ldw		ACT_USER_STACK(t2),sp
	break		0x1f,0x2
$reis0
#endif
	/*
	 * Set the reentry routine in the user activation
	 * to point to the syscall reentry point.
	 */
	ldil	L%$syscall_reentry,t4
	ldo	R%$syscall_reentry(t4),t4
	stw	t4,ACT_REENTRY_ROUTINE(t3)

	/*
	 * Make the new activation officially the current one.
	 */
	ldw	ACT_PCB(t2),t3
	ldil	L%cur_pcb,t4
	stw	t3,R%cur_pcb(t4)

	/*
	 * At this point,
	 * the new activation is attached to and owned by the thread.
	 * We are done mucking in the old activation.
	 * Therefore, this is the new base for interruption rollback.
	 */

	/*
	 * Initialize the kernel stack pointer from the new activation.
	 */
	ldw	ACT_USER_STACK(t2),sp

#if DEBUG
	/* 
	 * Clear out the previous stack pointer word.
	 */
	stw	0,FM_PSP(sp)
#endif

	/*
	 * Set the data pointer for the kernel.
	 */
	ldil	L%$global$,dp
	ldo	R%$global$(dp),dp

	/*
	 * Negate the system call code in r22 (t1).
	 * If the original code was positive then it's probably a dynamic interface call.
	 */
	sub,>	r0,t1,t2
	b,n	$mach_call_range

	/*
	 * get the maximum mach trap number and see if this is at or below it
	 */
	ldil	L%mach_trap_count,t1
	ldw	R%mach_trap_count(t1),t1

	comb,>,n t2,t1,$mach_call_range

#ifdef ACTWATCH
	/*
	 * Now that kernel activation is attached check again
	 * (mostly to catch problems from the fast path).
	 * Abs value of mach syscall # is in t2.
	 */
	break	0,11
#endif

	/*
	 * get the mach trap table
	 */
	ldil	L%mach_trap_table,t1
	ldo	R%mach_trap_table(t1),t1

	/*
	 * Now shift the syscall code to the left 4 bits to get an index into
	 * the mach_trap_table. The first word of each entry is the argument
	 * count, the next word is the offset for the routine.
	 */
	shd	t2,r0,28,t2
	add	t1,t2,t1

	/*
	 * t1 now points to the entry in mach_trap_table for this system call.
	 * get the argument count for this call
	 */
	ldw	0(t1),t2

	/*
	 * If the argument count is less than or equal to 4 then all the 
	 * arguments are already in the registers. In that case just call 
	 * the routine.
	 */
	comib,>=,n 4,t2,$mach_call_call

	/*
	 * argh... some of the arguments that we need are back on the caller's
	 * frame on the user's stack. Let's go pick them up and move them into
	 * our frame. We have room for NARGS arguments. 4 in registers and the
	 * rest in the stack frame. Set t2 to be the number of arguments that
	 * we don't have to move. At the same time check if we need to move
	 * more than NARGS registers.
	 */
#ifndef DEBUG
	subi	NARGS,t2,t2
#else
	subi,<	NARGS,t2,t2
	b,n	$mach_call_cont

	/*
	 * panic! Someone has modified kern/syscall_sw.c and not made the 
	 * corrections to this file.
	 */
	ldil	L%$syscall_panic,arg0
	ldo	R%$syscall_panic(arg0),arg0

	.import	panic,code
	ldil	L%panic,t1
	ldo	R%panic(t1),t1
	.call 
	blr     r0,rp
	bv      0(t1)
	nop

	/*
	 * never returns...
	 */

$mach_call_cont
#endif

#ifndef ROLLBACK
	/*
	 * Set up the thread recovery pointer to catch faults.
	 */
	.import	$mach_call_badaddr,code
	ldil	L%$mach_call_badaddr,t4
	ldo	R%$mach_call_badaddr(t4),t4
	stw	t4,THREAD_RECOVER(r1)
#endif

#ifdef MULTIPLE_SPACES_PER_TASK
	/*
	 * Get the user's data space pointer for explicit user space accesses.
	 */
	ldw	SS_SR5(ret1),t4
#else
	/*
	 * Get the user's data space pointer for explicit user space accesses.
	 * XXX sr4 - sr6 are all the same, so we only save sr4 in the saved
	 * state.  Hence, we must use it.
	 */
	ldw	SS_SR4(ret1),t4
#endif
	mtsp	t4,sr1

	/*
	 * next get the user's stack pointer into t4
	 */
	ldw	SS_R30(ret1),t4

	/*
	 * now move the correct number of arguments over from the user's stack
	 * to our stack. This is a really specialized form of copyin!
	 */
	blr,n	t2,r0
	nop

	ldw	VA_ARG11(sr1,t4),t2
	stw	t2,VA_ARG11(sp)

	ldw	VA_ARG10(sr1,t4),t2
	stw	t2,VA_ARG10(sp)

	ldw	VA_ARG9(sr1,t4),t2
	stw	t2,VA_ARG9(sp)

	ldw	VA_ARG8(sr1,t4),t2
	stw	t2,VA_ARG8(sp)

	ldw	VA_ARG7(sr1,t4),t2
	stw	t2,VA_ARG7(sp)

	ldw	VA_ARG6(sr1,t4),t2
	stw	t2,VA_ARG6(sp)

	ldw	VA_ARG5(sr1,t4),t2
	stw	t2,VA_ARG5(sp)

	ldw	VA_ARG4(sr1,t4),t2
	stw	t2,VA_ARG4(sp)

#ifndef ROLLBACK
	stw	r0,THREAD_RECOVER(r1)
#endif

$mach_call_call
	/*
	 * get the address of the routine to call
	 */
	ldw	4(t1),t2
#ifdef ACTCHECK
	comb,<>,n	r0,t2,$saok
	break		0x1f,0x3
$saok
#endif

#ifdef SYSCALL_PRINT
	.import syscall_print_enable,data
	ldil	L%syscall_print_enable,t1
	ldw	R%syscall_print_enable(t1),t1
	comb,=,n t1,r0,$dont_pre_print
	.import syscall_pre_print
	ldil	L%syscall_pre_print,t1
	ldo	R%syscall_pre_print(t1),t1
	.call
	blr     r0,rp
	bv	0(t1)
	nop
	copy	ret0,t2
$dont_pre_print
#endif

#ifdef ACTWATCH
	/*
	 * Check one last time (pick up errors from stack access faults).
	 * XXX original syscall # in t1
	 */
	ldw	PCB_ACT(t3),t1
	ldw	ACT_SPARE+40(t1),t1
	break	0,11
#endif

	/*
	 * call the system routine
	 */
	.call
	blr     r0,rp
	bv	0(t2)
	nop


#ifdef SYSCALL_PRINT
	.import syscall_print_enable,data
	ldil	L%syscall_print_enable,t1
	ldw	R%syscall_print_enable(t1),t1
	comb,=,n t1,r0,$dont_print
	.import syscall_print
	ldil	L%syscall_print,t2
	ldo	R%syscall_print(t2),t2
	copy	ret0,arg0
	.call
	blr     r0,rp
	bv	0(t2)
	nop
$dont_print
#endif

#ifdef TRAP_STACK_CHECK
	.import trap_stack_check
	ldil	L%trap_stack_check,t2
	ldo	R%trap_stack_check(t2),t2
	copy	ret0,arg0
	.call
	blr     r0,rp
	bv	0(t2)
	nop
#endif

$syscall_return
	/*
	 * Find the thread (r1), kernel activation (t2), and user activation (t3).
	 */
	ldil	L%cur_pcb,t4
	ldw	R%cur_pcb(t4),t2
	ldw	PCB_ACT(t2),t2
	ldw	ACT_LOWER(t2),t3
	ldw	ACT_THREAD(t2),r1

#ifdef ACTWATCH
	/*
	 * Back from syscall, check again
	 * XXX load original syscall # in t1
	 */
	ldw	ACT_SPARE+40(t3),t1
	break	0,11
#endif

	/*
	 * Call any waiting returnhandlers
	 */
	ldw	ACT_HANDLERS(t3),t1
	comb,<>,n t1,r0,$syscall_call_handlers

	/*
	 * Jump to the user activation's return transition routine.
	 * This could in fact be the null_act's thread-destruction routine.
	 */

#ifdef ACTCHECK
	ldw		ACT_REENTRY_ROUTINE(t3),arg3
	ldil		L%$syscall_reentry,t1
	ldo		R%$syscall_reentry(t1),t1
	comb,=,n	t1,arg3,$rerok
	break		0x1f,0x4
$rerok
#endif

	/*
	 * Remove the kernel activation from the thread.
	 */
	stw	t3,THREAD_TOP_ACT(r1)
	stw	r0,ACT_THREAD(t2)

#ifdef ACTCHECK
	stw	r0,ACT_LOWER(t2)
	stw	r0,ACT_HIGHER(t3)
	stw	r0,ACT_USER_ACT(t2)
#endif

	/* 
	 * Drop the reference count on the user activation. XXX
	 */
	ldw	ACT_REF_COUNT(t3),t1
	ldo	-1(t1),t1
	stw	t1,ACT_REF_COUNT(t3)

	/*
	 * Return the kernel activation to the kernel port.
	 */
	ldil	L%kernel_port,t1
	ldw	R%kernel_port(t1),t1
	ldw	IPT_ACTS(t1),arg3
	stw	t2,IPT_ACTS(t1)
	stw	arg3,ACT_IPT_NEXT(t2)

	/*
	 * Return to the user activation.
	 * This marks the end of the (very simple) return transition.
	 */
	ldw	ACT_PCB(t3),t1
	stw	t1,R%cur_pcb(t4)

	/*
	 * Invoke the user activation's return reentry routine.
	 * This could be the null_act's thread-destruction routine
	 * or an encapsulated activation's APC routine.
	 */
	ldw	ACT_REENTRY_ROUTINE(t3),t4
#ifdef ACTCHECK
	stw		r0,ACT_REENTRY_ROUTINE(t3)
#if 0
	comb,<>,n	r0,t4,$rerok2
#else
	ldil		L%$syscall_reentry,t2
	ldo		R%$syscall_reentry(t2),t2
	comb,=,n	t2,t4,$rerok2
#endif
	break		0x1f,0x4
$rerok2
#endif
	bv,n	0(t4)

	.export	$syscall_reentry,entry /* XXX??? */
$syscall_reentry
	/*
	 * t1 should contain the current (user) PCB
	 */

#ifdef ACTCHECK
	ldw		PCB_ACT(t1),t2
	ldw		ACT_THREAD(t2),t2
	comb,<>,n	r0,t2,$tok2
	break		0x1f,0x5
$tok2
#endif

	/*
	 * Disable the FPU so the user code will trap if it tries to use it.
	 */
	mtctl	r0,ccr

	/*
	 * Check for AST delivery.
	 * If need_ast is not zero then we need to load the lower privilege
	 * transfer trap bit in the PSW.  This will mean turning off the
	 * Q-bit to load the PC queues so that we can do an RFI and set
	 * the L bit.
	 */
	ldil	L%need_ast,t2
	ldw	R%need_ast(t2),t2
	comb,<>,n r0,t2,$syscall_ast

$syscall_ast_back
	/*
	 * restore the user's registers
	 */
	ldw	SS_R31(t1),r31
	ldw	SS_R30(t1),r30
	ldw	SS_R27(t1),r27
#ifdef R4_IS_SPECIAL
	ldw	SS_R4(t1),r4
#endif
	ldw	SS_R2(t1),r2

	/*
	 * restore space registers; this means that we need to start explicitly
	 * saying sr1.
	 */
	mtsp	r0,sr1

	ldw	SS_SR4(sr1,t1),t2
	mtsp	t2,sr4
#ifdef MULTIPLE_SPACES_PER_TASK
	ldw	SS_SR5(sr1,t1),t2
#endif
	mtsp	t2,sr5
#ifdef MULTIPLE_SPACES_PER_TASK
	ldw	SS_SR6(sr1,t1),t2
#endif
	mtsp	t2,sr6
	ldw	SS_SR0(sr1,t1),t2
	mtsp	t2,sr0

#ifdef ACTWATCH
	/*
	 * Returning to caller, check one last time.
	 * XXX original syscall # in t1
	 */
	ldw	PCB_ACT(sr1,t1),t2
	ldw	ACT_SPARE+40(sr1,t2),t1
	break	0,11
#endif

	/*
	 * finally return to the caller and remove kernel protection.
	 */
	.call
	be	0(sr0,r31)
	mtctl	r0,pidr4

$syscall_ast
	/*
	 * We need to make sure that we don't take a fault or interrupt while
	 * we do this. This means that all this code has to fit in one page.
	 */

	/*
	 * save the current system mask and clear all of it
	 */
	rsm	RESET_PSW,t2

	/*
	 * now add the lower privilege transfer bit and move to ipsw
	 */
	depi	1,PSW_L_P,1,t2
	depi	1,PSW_C_P,1,t2
	mtctl	t2,ipsw

	/*
	 * load the space and offset queues
	 */
	mtctl	r0,pcsq
	mtctl	r0,pcsq

	ldil	L%$syscall_ast_back,t2
	ldo	R%$syscall_ast_back(t2),t2

	mtctl	t2,pcoq
	ldo	4(t2),t2
	mtctl	t2,pcoq

	rfi
	nop

$mach_call_range
	/*
	 * the system call number was out of range
	 * call syscall_error(exception, code, subcode, saved_state *)
	 */
	ldi	EXC_BAD_INSTRUCTION,arg0
	ldi	EXC_PARISC_INVALID_SYSCALL,arg1

	/*
	 * subcode is the original value of the system call code
	 */
	sub	r0,r22,arg2
	copy	t3,arg3

$mach_call_error
	.import	syscall_error,code
	ldil	L%syscall_error,t1
	ldo	R%syscall_error(t1),t1
	.call
	blr     r0,rp
	bv,n    0(t1)
	nop

	/*
	 * just handle this like a normal return from a system call...
	 */
	b,n	$syscall_return

#ifndef ROLLBACK
/*
 * Come here on longjmp from trap code on an unresolved page fault while
 * doing copyin of user arguments.
 */
$mach_call_badaddr
	stw	r0,THREAD_RECOVER(r1)

	/*
	 * syscall_error(exception, code, subcode, saved_state *)
	 *	subcode is the faulting address which was put in ret0
	 *	by the trap code
	 * We already had a frame ready to call the system call, just use it.
	 */
	ldi	EXC_BAD_ACCESS,arg0
	ldi	EXC_PARISC_VM_PROT_READ,arg1
	copy	ret0,arg2
	b	$mach_call_error
	copy	t3,arg3
#endif


/* XXX become encapsulation code */
/*
 * Detected waiting returnhandlers coming out of the syscall.
 * Transform the syscall stack frame into a trap-style stack frame
 * and return through the trap path, since the user state could be changed arbitrarily.
 *
 * Thread (r1), kernel activation (t2), and user activation (t3).
 */
$syscall_call_handlers
	/*
	 * Find the user PCB (t1).
	 */
	ldw	ACT_PCB(t3),t1

	/*
	 * Clear the caller-save registers which are undefined because we entered by a syscall.
	 * XXX Should we really do this, or would it be better to just leave the old stale values?
	 */
	stw	r0,SS_R1(t1)
	stw	r0,SS_R19(t1)
	stw	r0,SS_R20(t1)
	stw	r0,SS_R21(t1)
	stw	r0,SS_R22(t1)
	stw	r0,SS_R23(t1)
	stw	r0,SS_R24(t1)
	stw	r0,SS_R25(t1)
	stw	r0,SS_R26(t1)
	stw	r0,SS_R28(t1)
	stw	r0,SS_R29(t1)
	stw	r0,SS_SR1(t1)
	stw	r0,SS_SR2(t1)

	/*
	 * Setup the instruction queue to mimic the 'be' instruction
	 * with which syscall would have returned.
	 */
	ldw	SS_SR0(t1),t4
	stw	t4,SS_IISQH(t1)
	stw	t4,SS_IISQT(t1)
	ldw	SS_R31(t1),t4
	depi	3,31,2,t4
	stw	t4,SS_IIOQH(t1)
	ldo	4(t4),t4
	stw	t4,SS_IIOQT(t1)

	/*
	 * Save other registers normally saved in a trap but not in a syscall.
	 */
	mfctl	eiem,t4
	stw	t4,SS_CR15(t1)
	mfctl	sar,t4
	stw	t4,SS_CR11(t1)
	mfctl	pidr1,t4
	stw	t4,SS_CR8(t1)
	mfctl	rctr,t4
	stw	t4,SS_CR0(t1)
	ssm	0,t4
	stw	t4,SS_CR22(t1)

	/*
	 * Indicate that this state was saved from a syscall.
	 */
	ldi	SS_INSYSCALL,t2
	stw	t2,SS_FLAGS(t1)

	/*
	 * Jump to the trap returncall handler,
	 * which will save the callee-save general registers, execute the returnhandlers,
	 * and eventually return through the trap path.
	 */
	.import	$call_handlers,code
	ldil	L%$call_handlers,t1
	ldo	R%$call_handlers(t1),t1
	bv	0(t1)
	nop

	.end

