/* 
 * Copyright (c) 1990, 1991, 1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: busconf.c 1.5 92/05/22$
 *	Author: Jeff Forys, Bob Wheeler, University of Utah CSL
 */

/*
 *	Find and partially configure devices on HP800 series machines.
 *	Most of the configuration of a particular device is left for the
 *	specific device's attach routine.
 */

#include <sys/types.h>
#include <machine/iomod.h>
#include <machine/pdc.h>

extern	int (*pdc)();		/* PDC entry point */

u_int	spa_size = 0;		/* amount of SPA space used by mem_size */
u_int	mem_size = 0;		/* bottom of memory (alloc up) */
u_int	io_size = FP_ADDR;	/* top of IO space (alloc down) */

struct modtab modtab[MODTABSIZ] = { &modtab[0] };
static	struct modtab *mptr = modtab;
static  struct modtab *last_mptr = (struct modtab *)0;
struct modtab *smb_bus = modtab;

#define	SPA_SAFE  PAGE0->imm_spa_size

#define SMB_FLEX_ADDRESS	((struct iomod *) 0xfff80000)
#define SMB_LBA_ADDRESS		((struct iomod *) 0xfffc0000)
#define SMB_GBA_ADDRESS		((struct iomod *) 0xfffe0000)

busconf()
{
	mem_size = PAGE0->imm_max_mem;
	spa_size = PAGE0->imm_spa_size;

	/*
	 * Broadcast the SMB flex address for cards on the main system bus
	 */
	SMB_LBA_ADDRESS->io_flex =
		(caddr_t)((u_int) SMB_FLEX_ADDRESS | DMA_ENABLE);

	/*
	 * scan the SMB bus
	 */
	configure_bus(SMB_FLEX_ADDRESS, (struct modtab *)0);
}

configure_bus(flex, upper_port)
	struct iomod  *flex;		/* flex address of bus */
	struct modtab *upper_port;	/* upper port of bus convertor */
{	
	struct pdc_iodc_read ignore;
	struct modtab *bus = mptr;

	/*
	 * find out what is on this bus...
	 */
	scan_bus(flex, upper_port);

	/*
	 * allocate soft physical address space for this bus
	 */
	allocate_spa(bus);

	/*
	 * initialize the devices on the bus
	 */
	initialize_bus(bus);

	/*
	 * configure remote busses below us
	 */
	configure_remote_bus(bus);
}

scan_bus(hpa, upper_port)
	struct iomod  *hpa;		/* initial hpa of bus */
	struct modtab *upper_port;	/* upper port of bus convertor */
{
	struct pdc_iodc_read ignore;
	register int fixed;
	register u_int spasiz;

	/*
	 * For each module on the bus, ask the PDC what it is; if we
	 * support the module, add it to our module table.
	 */
	for (fixed = 0; fixed < MAXMODBUS; fixed++, hpa++) {
		if ((*pdc)(PDC_IODC, PDC_IODC_READ, &ignore, hpa,
			   PDC_IODC_INDEX_DATA,&mptr->m_type,
			   sizeof(mptr->m_type)) < 0)
			continue;

		mptr->m_fixed = fixed;
		mptr->m_hpa = hpa;
		mptr->m_spa = mptr->m_spasize = 0;

		/*
		 * If this bus is below a bus convertor then record the
		 * first thing found on the bus.
		 */
		if (upper_port != (struct modtab *)0 && 
		    upper_port->m_nextbus == (struct modtab *)0)
			upper_port->m_nextbus = mptr;

		switch(mptr->m_type.iodc_type) {
		case IODC_TP_NPROC:            /* native processor */
			if (upper_port != (struct modtab *)0) {
				printf("processor below SMB bus ignored\n");
				continue;
			}
			break;

		case IODC_TP_MEMORY:	       /* memory module */
			if (upper_port != (struct modtab *)0) {
				printf("memory below SMB bus ignored\n");
				continue;
			}
			if (hpa == PAGE0->imm_hpa) {
				mptr->m_spasize = PAGE0->imm_spa_size;
				mptr->m_memsize = PAGE0->imm_max_mem;
				break;
			}

			/*
			 * initialize and test memory modules
			 */
			switch (mptr->m_type.iodc_sv_model) {
			case SVMOD_MEM_ARCH: {	/* architected */
				u_int spa, spaincr, spamax;
				volatile int junk;

				/*
				 * Send a reset command to the memory board.
				 * Give this memory module a fake SPA to 
				 * enable it. Prepare for memtest by trapping 
				 * certain errors.
				 */
				if (io_command(hpa, CMD_RESET)) {
					config_error("can't reset memory",
						     mptr);
					continue;
				}
				hpa->io_spa = 
					IOSPA(SPA_SAFE,mptr->m_type);
				hpa->io_control = IO_CTL_MEMINIT;
				mptr->m_spasize =
					1 << mptr->m_type.iodc_spa_shift;

				/*
				 * We have "mptr->m_spasize" bytes of memory; 
				 * testing all of it will take too long!  
				 * Our memory boards are 16MB. If we assume 
				 * that the smallest memory board is 256K 
				 * (given the boot size restriction),
				 * lets check every 128K or so.
				 */
				spaincr = mptr->m_spasize >> 7;
				spamax = SPA_SAFE + mptr->m_spasize;
				for (spa = SPA_SAFE; spa < spamax; 
				     spa += spaincr) {
					junk = *(int *)spa;
					if (IO_ERR_VAL(hpa->io_status)
					    ==IO_ERR_SPA)
						break;
				}


				mptr->m_memsize = spa - SPA_SAFE;
				if (io_command(hpa, CMD_CLEAR)) {
					config_error("can't clear memory",
						     mptr);
					continue;
				}
				if (mptr->m_memsize != mptr->m_spasize) {
					config_error("bad memory module",
						     mptr);
					continue;
				}

				break;
			}

			case SVMOD_MEM_PDEP: {	/* processor-dependent */
				struct pdc_iodc_minit minit;

				/*
				 * All we can do here is contact the PDC to 
				 * find out how much memory this module has.
				 */
				(*pdc)(PDC_IODC, PDC_IODC_NINIT, &minit, 
				       hpa, SPA_SAFE);
				mptr->m_memsize = minit.max_mem;
				mptr->m_spasize = minit.max_spa;
				break;
			}

			default:
				config_error("unsupported memory", mptr);
				continue;
			}
			
			hpa->io_spa = 0;	/* disable */
			break;

		case IODC_TP_BCPORT:	/* Bus Convertor */
			mptr->m_islower = hpa->io_status &IO_ERR_BC_LP;
			if (!mptr->m_islower)
				break;

			mptr->m_otherport = upper_port;
			upper_port->m_otherport = mptr;
			if (io_command(hpa, CMD_RESET)) {
				config_error("can't reset bus conv", mptr);
				continue;
			}
			break;

		case IODC_TP_CIO:	/* CIO */
		case IODC_TP_B_DMA:	/* Type-B DMA (NIO Transit, Parallel)*/
		case IODC_TP_B_DIRECT:	/* Type-B Direct */
		case IODC_TP_A_DMA:	/* Type-A DMA (NIO HPIB, LAN, ... ) */
		case IODC_TP_A_DIRECT:  /* Type-A Direct (RS232, HIL, ... ) */
			if (mptr->m_type.iodc_spa_shift) {
				spasiz = 1 << mptr->m_type.iodc_spa_shift;
				mptr->m_spasize = (spasiz > 1)? spasiz: 0;
			} else
				mptr->m_spasize = 0;
			break;

		case IODC_TP_OTHER:	         /* monitor */
		case IODC_TP_CONSOLE:	         /* console */
			break;

		default:			/* unsupported module  */
			config_error("unsupported module", mptr);
			continue;
		}

		if (last_mptr != (struct modtab *)0)
			last_mptr->m_next = mptr;
		mptr->m_next = (struct modtab *)0;
		last_mptr = mptr;
		mptr++;
	}

	last_mptr = (struct modtab *)0;
}

/*
 * Allocate SPA space for bus
 */
allocate_spa(bus)
	register struct modtab *bus;
{
	register struct modtab *best;
	register u_int largest;
	register struct modtab *m;

	/*
	 * allocate SPA space for any memory modules on this bus
	 */
	while(1) {
		largest = 0;
		for (m = bus; m != (struct modtab *)0; m = m->m_next) {
			if (m->m_type.iodc_type != IODC_TP_MEMORY
			    || m->m_hpa == PAGE0->imm_hpa
			    || m->m_spasize == 0
			    || m->m_spa != 0
			    || m->m_type.iodc_spa_io == 1)
				continue;

			if (m->m_spasize > largest) {
				best = m;
				largest = m->m_spasize;
			}
		}

		if (largest == 0)
			break;

		best->m_spa = spa_size;
		spa_size += best->m_spasize;
		mem_size += best->m_memsize;
	}

	/*
	 * allocate SPA space for I/O devices on this bus. Priority goes to
	 * modules with the correct alignment. In case of a tie, allocate the
	 * one with the largest SPA need. If no module meets the current 
	 * alignment then then find the largest SPA and increase io_size to get
	 * the proper alignment.
	 */
	while(1) {
		/*
		 * only look at modules that meet the current alignment of 
		 * io_size and find the largest SPA needed.
		 */
		largest = 0;
		for (m = bus; m != (struct modtab *)0; m = m->m_next) {
			if (m->m_type.iodc_type == IODC_TP_MEMORY
			    || m->m_spasize == 0
			    || m->m_spa != (caddr_t)0
			    || m->m_type.iodc_spa_io == 0)
				continue;

			if (io_size & (m->m_spasize - 1)) {
				if (m->m_spasize > largest) {
					best = m;
					largest = m->m_spasize;
				}
			}
		}

		if (largest == 0) {
			/*
			 * didn't find anything with the correct alignment,
			 * get the largest SPA need
			 */ 
			for (m = bus; m != (struct modtab *)0; 
			     m = m->m_next) {
				if (m->m_type.iodc_type == IODC_TP_MEMORY
				    || m->m_spasize == 0
				    || m->m_spa != (caddr_t)0
				    || m->m_type.iodc_spa_io == 0)
					continue;

				if (m->m_spasize > largest) {
					best = m;
					largest = m->m_spasize;
				}
			}
		}

		if (largest == 0)
			break;

		/*
		 * align io_size and then allocate spa space
		 */
		io_size &= ~(best->m_spasize - 1);
		io_size -= best->m_spasize;
		best->m_spa = io_size;

		/*
		 * If this was our boot or console device SPA, update PAGE0 
		 * appropriately.
		 */
		if (best->m_hpa == PAGE0->mem_boot.pz_hpa)
			PAGE0->mem_boot.pz_spa = (caddr_t) io_size;
		else if (best->m_hpa == PAGE0->mem_cons.pz_hpa)
			PAGE0->mem_cons.pz_spa = (caddr_t) io_size;
		else if (best->m_hpa == PAGE0->mem_kbd.pz_hpa)
			PAGE0->mem_kbd.pz_spa = (caddr_t) io_size;
	}

}

initialize_bus(bus)
	register struct modtab *bus;
{
	register struct modtab *m;

	/*
	 * Disable all SPA space.
	 */
	for (m = bus; m != (struct modtab *)0; m = m->m_next)
		if (m->m_spa && m->m_hpa != PAGE0->mem_cons.pz_hpa)
			m->m_hpa->io_spa = 0;

	/*
	 * Do module specific reconfiguration.
	 */
	for (m = bus; m != (struct modtab *)0; m = m->m_next) {
		switch(m->m_type.iodc_type) {
		case IODC_TP_NPROC:            /* native processor */
			if (m->m_hpa == PAGE0->mem_hpa)
				m->m_alive = 1;
			break;

		case IODC_TP_MEMORY:
			m->m_alive = 1;
			if (m->m_hpa == PAGE0->imm_hpa)
				break;

			if (m->m_spa == 0)
				break;

			if (m->m_type.iodc_sv_model == SVMOD_MEM_ARCH) {
				m->m_hpa->io_command = CMD_RESET;
				m->m_hpa->io_spa = IOSPA(m->m_spa,m->m_type);
				m->m_hpa->io_control = IO_CTL_MEMOKAY;
			} else if (m->m_type.iodc_sv_model == SVMOD_MEM_PDEP) {
				struct pdc_iodc_minit minit;

				/* tell module IODC to initialize memory */
				(*pdc)(PDC_IODC, PDC_IODC_NINIT, &minit,
				       m->m_hpa, m->m_spa);
			}
			break;

		case IODC_TP_BCPORT:	       /* Bus Convertor */
			break;

		case IODC_TP_CIO:	/* CIO */
		case IODC_TP_B_DMA:	/* Type-B DMA (NIO Transit, Parallel)*/
		case IODC_TP_B_DIRECT:	/* Type-B Direct */
		case IODC_TP_A_DMA:	/* Type-A DMA (NIO HPIB, LAN, ... ) */
		case IODC_TP_A_DIRECT:  /* Type-A Direct (RS232, HIL, ... ) */
		case IODC_TP_OTHER:
		case IODC_TP_CONSOLE:	         /* console */
			break;

		}
	}
}

configure_remote_bus(bus)
	struct modtab *bus;
{
	register struct modtab *m;
	struct iomod *upper_hpa;
	struct iomod *lower_hpa;
	struct iomod *lower_flex;

	for (m = bus; m != (struct modtab *)0; m = m->m_next) {
		if (m->m_type.iodc_type != IODC_TP_BCPORT || m->m_islower)
			continue;

		upper_hpa = m->m_hpa;

		io_size &= BCONV_ALIGN;
		m->m_io_high = (u_int) io_size;

		/*
		 * get new HPA for this bus, need 256K (0x40000) for HPA region
		 */
		lower_flex = (struct iomod *) ((io_size - 0x40000) &FLEX_MASK);

		if (io_command(upper_hpa, CMD_RESET)) {
			config_error("can't reset bus convertor", m);
			continue;
		}

		if (io_command(upper_hpa, CMD_CLEAR)) {
			config_error("can't clear bus convertor", m);
			continue;
		}

		if (io_command(upper_hpa, CMD_PORT)) {
			config_error("can't set port on bus convertor", m);
			continue;
		}

		/*
		 * set address range that this bus convertor should pass on
		 */
		upper_hpa->io_io_low  = (u_int) IO_ADDR;
		upper_hpa->io_io_high = m->m_io_high;

		/*
		 * put upper port in peek mode and broadcast the flex address
		 */
		upper_hpa->io_control = IO_CTL_BC_PEEK;
		lower_flex->io_flex = (caddr_t) lower_flex;
		upper_hpa->io_control = IO_CTL_BC_INCLUDE;

		/*
		 * allocate HPA space and scan and configure lower busses
		 */
		io_size = (u_int) lower_flex;
		scan_bus(lower_flex, m);
		allocate_spa(m->m_nextbus);
		configure_remote_bus(m->m_nextbus);

		/*
		 * align io_size to 64K for bus convertor and set low address
		 */
		io_size &= BCONV_ALIGN;
		m->m_io_low = (u_int) io_size;

		/*
		 * set up other side of the bus convertor
		 */
		m->m_otherport->m_io_low = m->m_io_low;
		m->m_otherport->m_io_high = m->m_io_high;

		lower_hpa = m->m_otherport->m_hpa;

		/*
		 * now reset both sides of the convertor
		 */
		if (io_command(upper_hpa, CMD_RESET)) {
			config_error("can't reset bus convertor", m);
			continue;
		}

		/*
		 * set I/O range on upper port
		 */
		upper_hpa->io_io_low  = m->m_io_low;
		upper_hpa->io_io_high = m->m_io_high;

		/*
		 * broadcast flex address to lower bus
		 */
		upper_hpa->io_control = IO_CTL_BC_PEEK;
		lower_flex->io_flex = 
			(caddr_t) ((u_int) lower_flex | DMA_ENABLE);
		upper_hpa->io_control = IO_CTL_BC_INCLUDE;

		/*
		 * clear upper port
		 */
		if (io_command(upper_hpa, CMD_CLEAR)) {
			config_error("can't clear bus convertor", m);
			continue;
		}

		/*
		 * set range of lower port
		 */
		lower_hpa->io_io_low  = m->m_otherport->m_io_low;
		lower_hpa->io_io_high = m->m_otherport->m_io_high;

		/*
		 * put lower port in exclude mode
		 */
		lower_hpa->io_command = IO_CTL_BC_EXCLUDE;

		/*
		 * clear lower port
		 */
		if (io_command(lower_hpa, CMD_CLEAR)) {
			config_error("can't clear bus convertor", m);
			continue;
		}
	}
}


/*
 * We had an error with a module. Print as much about it and then return.
 */
config_error(what, where)
	char *what;
	struct modtab *where;
{
	printf("Warning: %s at %x (type:%x svers:%x hvers:%x)\n",
	       what, where->m_hpa, where->m_type.iodc_type,
	       where->m_type.iodc_sv_model, where->m_type.iodc_model);
}

/*
 * io_command
 *
 * issue a command to a device waiting for ready bit in the status register
 */
io_command(hpa, command)
	struct iomod *hpa;
	u_int command;
{
	/*
	 * Bus convertors must repond within 5 seconds, so we will wait that
	 * if we have to...
	 */

	int tries = 0;

	while(tries < 5 && (hpa->io_status & IO_ERR_MEM_RY) == 0) {
		delay(1000000);
		tries++;
	}

	if ((hpa->io_status & IO_ERR_MEM_RY) == 0) 
		return -1;
	else
		hpa->io_command = command;
	return 0;
}

dump_spa()
{
	struct modtab *m;

	for (m = modtab; m != (struct modtab *)0; m = m->m_next)
		printf("module slot = %x, type = %x, HPA/SPA/SPASIZE = %x/%x/%x\n",
		       m->m_fixed, m->m_type.iodc_type, m->m_hpa, m->m_spa, 
		       m->m_spasize);

	printf("io_size = %x, spa_size = %x\n", io_size, spa_size);
}

/*
 * Do lights.
 *	`ons' is a mask of LEDs to turn on,
 *	`offs' is a mask of LEDs to turn off,
 *	`togs' is a mask of LEDs to toggle.
 * XXX not very efficient, maybe do in assembly language?
 */
ledcontrol(ons, offs, togs)
	int ons, offs, togs;
{
}
