/* 
 * Copyright (c) 1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: hpib.c 1.5 94/11/10$
 *	Author: Bob Wheeler, University of Utah CSL
 */

/*
 * 	HP27110B CIO HPIB driver
 */

int dump_chains = 0;

#include <hpib.h>
#if NHPIB > 0

#include <sys/types.h>
#include <device/param.h>
#include <device/errno.h>
#include <device/buf.h>
#include <device/conf.h>
#include <machine/machparam.h>
#include <machine/spl.h>

#include <hpcio/device.h>
#include <hpcio/cioreg.h>
#include <hpcio/hpibreg.h>
#include <hpcio/cs80.h>

#define purge_data_cache(address, count) pdcache(0, (address), (count))

char odd_parity[] = {
        0x80, 0x01, 0x02, 0x83, 0x04, 0x85, 0x86, 0x07,
        0x08, 0x89, 0x8a, 0x0b, 0x8c, 0x0d, 0x0e, 0x8f,
        0x10, 0x91, 0x92, 0x13, 0x94, 0x15, 0x16, 0x97,
        0x98, 0x19, 0x1a, 0x9b, 0x1c, 0x9d, 0x9e, 0x1f,
        0x20, 0xa1, 0xa2, 0x23, 0xa4, 0x25, 0x26, 0xa7,
        0xa8, 0x29, 0x2a, 0xab, 0x2c, 0xad, 0xae, 0x2f,
        0xb0, 0x31, 0x32, 0xb3, 0x34, 0xb5, 0xb6, 0x37,
        0x38, 0xb9, 0xba, 0x3b, 0xbc, 0x3d, 0x3e, 0xbf,
        0x40, 0xc1, 0xc2, 0x43, 0xc4, 0x45, 0x46, 0xc7,
        0xc8, 0x49, 0x4a, 0xcb, 0x4c, 0xcd, 0xce, 0x4f,
        0xd0, 0x51, 0x52, 0xd3, 0x54, 0xd5, 0xd6, 0x57,
        0x58, 0xd9, 0xda, 0x5b, 0xdc, 0x5d, 0x5e, 0xdf,
        0xe0, 0x61, 0x62, 0xe3, 0x64, 0xe5, 0xe6, 0x67,
        0x68, 0xe9, 0xea, 0x6b, 0xec, 0x6d, 0x6e, 0xef,
        0x70, 0xf1, 0xf2, 0x73, 0xf4, 0x75, 0x76, 0xf7,
        0xf8, 0x79, 0x7a, 0xfb, 0x7c, 0xfd, 0xfe, 0x7f
	};

void	hpibintr();

int	hpibattach(), hpibprobe(), hpibscan(), hpiblayer();
void	hpibdisplay();
struct driver hpibdriver = {
	"hpib", hpibattach, hpibprobe, hpibscan, hpibdisplay, hpiblayer
};

struct hpib_softc {
	struct cio_chan	    *cc;	/* pointer to cio channel struct */
	struct dmaquad	    *cmdq;	/* command quad */
	struct dmaquad      *data;	/* read/write data quad */
	struct dmaquad	    *qstat;     /* cs/80 status quad */
	struct linkquad     *link;	/* link quad */
	struct statquad	    *stat;	/* status quad for DMA chain */
	struct semaquad     *sema;	/* semaphore quad */
	struct dmaquad      *quads;	/* extra quads for scatter/gather */
	struct hpib_request *head;	/* queue for hpib requests */
	struct hpib_request *tail;	/* queue for hpib requests */
	struct cio_request  *cr;	/* cio request packet */
	struct hpib_request *hr;	/* hpib request for internal use */
	int		    active;	/* requests are being processed */
	int                 alive;	/* alive and configured */
};

struct hpib_softc hpib_softc[NHPIB];
struct cio_request hpibcr[NHPIB];
struct hpib_request internal_hr[NHPIB];
struct hpib_dev hpib_dev[NHPIB][MAX_HPIB_SLAVES];

extern caddr_t cio_malloc();

/*
 * test for HPIB controller
 */
hpibprobe(cc)
	struct cio_chan *cc;
{
	return (cc->type == CIOIDY_HPIB) ? 0 : ENXIO;
}

/*
 * configure and initialize HPIB
 */
hpibattach(cc, unit, flags)
	struct cio_chan *cc;
{
	register struct hpib_softc *hs;
	struct hpib_request *hr;
	struct dmaquad *q;
	char *bp;

	if (cc->alive)
		return 0;

	if (unit > NHPIB)
		return ENXIO;

	hs = &hpib_softc[unit];

	/*
	 * remember the channel that we are assigned on the CIO
	 */
	hs->cc = cc;

	/*
	 * initialize the request queue
	 */
	hs->head = (struct hpib_request *) &hs->head;
	hs->tail = hs->head;

	/* 
	 * Allocate buffers and DMA quads, we will allow a maximum raw transfer
	 * of 64K bytes. This means that we need 31 additional scatter/gather 
	 * dmaquads. All together we need 37 quads.
	 */
	q = (struct dmaquad *) cio_malloc(sizeof(struct dmaquad) * 37, 
					 CIO_QUAD_ALIGN);

	hs->cmdq  = q++;
	hs->data  = q++;
	hs->qstat = q++;
	hs->link  = (struct linkquad *) q++;
	hs->stat  = (struct statquad *) q++;
	hs->sema  = (struct semaquad *) q++;
	hs->quads = q;

	/*
	 * some of the quad structure can be filled in once and for all...
	 */
	hs->qstat->link	   = (struct dmaquad *) hs->link;
	hs->link->link     = CIO_EOC(hs->link);
	hs->link->sema     = hs->sema;
	hs->link->stat     = hs->stat;

	/*
	 * fill in the internal synchronous HPIB requests
	 */
	hs->hr = hr  = &internal_hr[unit];
	hr->command  = cio_malloc(CIO_MIN_BUFFER, CIO_BUFFER_ALIGN);
	hr->buffer   = cio_malloc(0x2000, CIO_BUFFER_ALIGN);
	hr->qstatus  = cio_malloc(CIO_MIN_BUFFER, CIO_BUFFER_ALIGN);
	hr->qstatcnt = 0x20;  /* XXX - symbolic? */
	hr->unit     = unit;
	hr->hpibunit = unit;  /* XXX - This is confused! */

	/*
	 * now fill in the CIO request packet
	 */
	hs->cr          = &hpibcr[unit];
	hs->cr->cmdq    = hs->cmdq;
	hs->cr->stat    = hs->stat;
	hs->cr->quad    = hs->quads;
	hs->cr->ciounit = cc->ciounit;
	hs->cr->channel = cc->channel;
	hs->cr->polled  = 0;

	/*
	 * finally install the interrupt service routine
	 */
	cc->int_level = SPL_IBIT_CIOHPIB;
	(void)procitab(SPLCIOHPIB, hpibintr, 0, 0);

	cc->alive = 1;
	hs->alive = 1;

	/*
	 * HP-IB Initialize
	 */
	bp = (char *) hr->command;
	hpib_initialize(bp);
	hr->cmdcnt = bp - (char *) hr->command;
	hr->bufcnt = 0;

	if (hpibsync(hr, 0)) {
		printf("hpibattach: HP-IB initialize error, hpib%d", unit);
		cc->alive = 0;
		hs->alive = 0;
		return EIO;
	}

	/*
	 * clear the controller
	 */
	bp = (char *) hr->command;
	tcs80_nodata(bp, 0, 4);
	chan_indep_clear(bp, 0x0f);
	hr->cmdcnt = bp - (char *) hr->command;
	hr->bufcnt = 0;
	if (hpibsync(hr, 1)) {
		printf("hpibattach: channel clear error on hpib%d", unit);
		cc->alive = 0;
		hs->alive = 0;
		return EIO;
	}

	/*
	 * set release timeout to suppress release timeout and release 
	 * automatically during idle time
	 */
	bp = (char *) hr->command;
	cs80_nodata(bp, 0, 10);
	set_unit(bp, 0xF);
	set_release(bp, 1, 1);
	hr->cmdcnt = bp - (char *) hr->command;
	hr->bufcnt = 0;
	if (hpibsync(hr, 1)) {
		printf("hpibattach: set release error on hpib%d", unit);
		cc->alive = 0;
		hs->alive = 0;
		return EIO;
	}

	return 0;
}

/*
 * scan the HPIB for devices on it and start filling in tables
 */
hpibscan(unit)
{
	char *bp;
	int slave;
	struct hpib_softc *hs = &hpib_softc[unit];
	struct hpib_request *hr = hs->hr;
	struct cio_chan *cc = hs->cc;
	struct hpibbus *hi;
	struct hpib_dev	*hd;
	extern struct bus bus_devices[];
	struct bus *device;

	if (!hs->alive) 
		return ENXIO;

	/*
	 * scan the HPIB bus and get IDs for slaves
	 */
	hr->bufcnt = 2;
	hr->read = 1;

	for (slave = 0; slave < MAX_HPIB_SLAVES; slave++) {
		hd = &hpib_dev[unit][slave];
		hd->slave     = slave;
		hd->hpibunit  = unit;

		/*
		 * get amigo identification
		 */
		bp = (char *) hr->command;
		amigo_identify(bp, slave);
		hr->cmdcnt = bp - (char *) hr->command;
		if (hpibsync(hr, 0))
			hd->hwid = 0;
		else {
			bp = (char *) hr->buffer;
			if (*(char *) hr->qstatus == 0)
				hd->hwid = bp[0] << 8 | bp[1];
		}
	}


	/*
	 * look in the configuration table and try to assign units to slaves
	 */
	for (device = bus_devices; device->driver; device++) {
		/*
		 * if this device isn't on an HPIB then skip it
		 */
		if (device->bustype != &hpibdriver)
			continue;

		/*
		 * see if this is the right unit or if it's a wildcard 
		 * HPIB unit
		 */
		if (device->busunit != -1 && device->busunit != unit)
			continue;

		/*
		 * This is the right HPIB unit search all hpib slaves for
		 * something that looks like what we want...
		 */
		for (slave = 0; slave < MAX_HPIB_SLAVES; slave++) {
			hd = &hpib_dev[unit][slave];

			/*
			 * see if this device has already been configured 
			 * doesn't exist or doesn't have the right slave number
			 */
			if (hd->alive || !hd->hwid ||
			    (device->unit != slave && device->unit != -1))
				continue;

			/*
			 * see if we recognize this device as what we want...
			 */
			if ((*device->driver->probe)(hd) != 0)
				continue;

			/*
			 * this looks like it... attach it and associate it 
			 * with the right unit
			 */
			if ((*device->driver->attach)
			    (hd, device->unit,device->flags) != 0)
				continue;

			hd->alive = 1;
			hd->device = device;
			break;
		}

	}

	return 0;
}

void
hpibdisplay(unit)
{
	int slave;
	struct hpib_dev	*hd;
	extern int indent;

	for (slave = 0; slave < MAX_HPIB_SLAVES; slave++) {
		hd = &hpib_dev[unit][slave];
		if (!hd->hwid)
			continue;

		if (hd->alive) {
			iprintf("%s%d on hpib%d slave %d\n", 
				hd->device->driver->name,
				hd->device->unit, unit, slave);
			indent += 5;
			(*hd->device->driver->display)();
			indent -= 5;
		} else
			iprintf("unknown device id 0x%x on hpib%d slave %d\n",
				hd->hwid, unit, slave);
	}
}

hpiblayer(unit, layer, devname)
	int *layer;
	char *devname;
{
	struct hpib_dev	*hd;
	hd = &hpib_dev[unit][*layer];

	if (*layer >= MAX_HPIB_SLAVES)
		return ENXIO;

	if (!hd->hwid)
		return ENXIO;

	if (hd->alive)
		return (*hd->device->driver->layer)
			(hd->device->unit, layer + 1, devname);

	return ENXIO;
}


/*
 * queue an HPIB request
 */
hpibgo(hr)
	struct hpib_request *hr;
{
	struct hpib_softc *hs = &hpib_softc[hr->hpibunit];
	int s;

	if (hr->hpibunit > NHPIB) {
		printf("hpibgo: bad unit number\n");
		return ENXIO;
	}

	if (!hs->alive) {
		printf("hpibgo: unit not alive\n");
		return ENXIO;
	}

	s = splbio();
	/*
	 * insert this request in the queue for this HPIB controller
	 */
	insque(hr, hs->tail);

	/*
	 * see if the controller is active, if not then start it
	 */
	if (hs->active == 0) {
		hs->active = 1;
		hpibstart(hr->hpibunit);
	}
	splx(s);
	return 0;
}


/*
 * take the HPIB request on the top of the queue and fill in the necessary
 * quads and start the CIO
 */
hpibstart(hpibunit)
{
	struct hpib_softc   *hs = &hpib_softc[hpibunit];
	struct hpib_request *hr = hs->head;
	extern int noisy;

	while(hs->head != (struct hpib_request *) &hs->head) {

		/*
		 * get the next hpib request
		 */
		hr = hs->head;

		/*
		 * fill in the quads
		 */
		hs->cmdq->command  = CIODMA_WC;
		hs->cmdq->address  = hr->command;
		hs->cmdq->count    = hr->cmdcnt;

		/*
		 * if there is a data transaction then link in a data quad
		 */
		if (hr->bufcnt == 0)
			hs->cmdq->link = hs->qstat;
		else {
			hs->cmdq->link = hs->data;
			
			hs->data->command = (hr->read) ? CIODMA_RD : CIODMA_WD;
			hs->data->address = hr->buffer;
			hs->data->count   = hr->bufcnt;
			hs->data->link    = hs->qstat;
		}
		
		/*
		 * link in the HPIB QSTAT quad
		 */
		hs->qstat->command = CIODMA_RS;
		hs->qstat->address = hr->qstatus;
		hs->qstat->count   = hr->qstatcnt;

		hs->link->command  = CIODMA_LNKST | CIO_INTR;

		if ((hr->status = ciogo(hs->cr)) != 0) {
			printf("hpibstart: hpib%d failed ciogo\n", hpibunit);
			remque(hs->head);
			(*hr->done)(hr->unit);
			continue;
		} else
			return;
	}

	/*
	 * error with packet and nothing else to do
	 */
	hs->active = 0;
}


/*
 * take the HPIB request on the top of the queue and fill in the necessary
 * quads and start the CIO, returns synchronously when done. This routine
 * should only be used at bootstrap time.
 */
hpibsync(hr, cs80)
	struct hpib_request *hr;
{
	struct hpib_softc *hs = &hpib_softc[hr->hpibunit];

	if (hr->hpibunit > NHPIB) {
		printf("hpibsync: bad unit number\n");
		return ENXIO;
	}

	if (!hs->alive) {
		printf("hpibsync: unit not alive\n");
		return ENXIO;
	}

	/*
	 * fill in the quads
	 */
	hs->cmdq->command  = CIODMA_WC;
	hs->cmdq->address  = hr->command;
	hs->cmdq->count    = hr->cmdcnt;

	/*
	 * if there is a data transaction then link in a data quad
	 */
	if (hr->bufcnt == 0)
		hs->cmdq->link = hs->qstat;
	else {
		hs->cmdq->link = hs->data;

		hs->data->command = (hr->read) ? CIODMA_RD : CIODMA_WD;
		hs->data->address = hr->buffer;
		hs->data->count   = hr->bufcnt;
		hs->data->link    = hs->qstat;
	}

	/*
	 * link in the HPIB QSTAT quad
	 */
	hs->qstat->command = CIODMA_RS;
	hs->qstat->address = hr->qstatus;
	hs->qstat->count   = hr->qstatcnt;

	hs->link->command  = CIODMA_LNKST;

	if ((hr->status = ciosync(hs->cr)) != 0) {
		printf("hpibsync : failed ciosync\n");
		return hr->status;
	}

	if ((hr->status = hpib_error(hr->qstatus, cs80)) != 0)
		return hr->status;

	return 0;
}

void
hpibintr()
{
	struct hpib_softc *hs;
	int unit;
	int s;

	/*
	 * we can get an interrupt for both a transaction completion and a
	 * module error. First check for any transactions that have completed
	 */
	for (unit = 0; unit < NHPIB; unit++) {
		struct hpib_softc   *hs = &hpib_softc[unit];
		struct hpib_request *hr = hs->head;

		if (!hs->active)
			continue;

		/*
		 * we know there is a transaction in progress, check the
		 * CIO status link to see if it is finished.
		 */
		s = splbio();
		if (!hs->cr->stat->link) {
			pdcache(0, hs->cr->stat, sizeof(struct statquad));
			splx(s);
			continue;
		}
		splx(s);

		dequeue_cio_request(hs->cc);

		/*
		 * remove this request from the queue
		 */
		remque(hs->head);
	
		/*
		 * call the appropriate routine to signal this request is done
		 */
		(*hr->done)(hr->unit);

		/*
		 * see if there are any more requests to start
		 */
		if (hs->head != (struct hpib_request *) &hs->head)
			hpibstart(unit);
		else
			hs->active = 0;
	}

	/*
	 * now see if there are any errors on the channel
	 */
	for (unit = 0; unit < NHPIB; unit++) {
		struct hpib_softc   *hs = &hpib_softc[unit];

		if (!test_cio_are(hs->cc))
			continue;

		/*
		 * I don't think we should see interrupts here. Since we only 
		 * do CS/80 transactions all errors should be in CS/80 status 
		 * buffers. Minimally tell the hpibdone routine that this
		 * transaction is shot.
		 */
		if (hs->head != (struct hpib_request *) &hs->head) {
			printf("hpibintr: I/O request killed\n");
			hs->head->status = EIO;
			(*hs->head->done)(hs->head->unit);
		}

		clear_cio_are(hs->cc);

		printf("hpib%d interrupt detected...\n", unit);
	}
}

/*
 * this routine analyzes a status buffer for errors and passes back an error
 * code.
 */
hpib_error(stat, cs80)
	struct cs80_status *stat;
{
	int i;
	
	/*
	 * first look at the tstat field...
	 */
	if (stat->tstat > 4)
		return EIO;

	/*
	 * if this is a CS/80 packet then check the qstat
	 */
	if (cs80 && stat->qstat) {
		printf("hpib_error : qstat = %x\n", stat->qstat);

		printf("stat->tstat   = %x\n", stat->tstat & 0xff);
		printf("stat->qstat   = %x\n", stat->qstat & 0xff);
		printf("stat->volunit = %x\n", stat->volunit & 0xff);
		printf("stat->pending = %x\n", stat->pending & 0xff);
		printf("stat->reject  = %x\n", stat->reject & 0xffff);
		printf("stat->fault   = %x\n", stat->fault & 0xffff);
		printf("stat->access  = %x\n", stat->access & 0xffff);
		printf("stat->info    = %x\n", stat->info & 0xffff);
		printf("stat->params  = ");
		for (i = 0; i < 10; i++)
			printf("%x ", stat->params[i]);
		printf("\n");

		return EIO;
	}

	return 0;
}

dump_hr(hr)
	struct hpib_request *hr;
{
	char *b;

	printf("hr : %x\n", hr);
	printf("hr->next     = %x\n", hr->next);
	printf("hr->prev     = %x\n", hr->prev);
	printf("hr->read     = %x\n", hr->read);
	printf("hr->command  = %x\n", hr->command);
	printf("hr->cmdcnt   = %x\n", hr->cmdcnt);
	dump_buffer(hr->command, hr->cmdcnt);
	printf("hr->buffer   = %x\n", hr->buffer);
	printf("hr->bufcnt   = %x\n", hr->bufcnt);
	printf("hr->qstatus  = %x\n", hr->qstatus);
	printf("hr->qstatcnt = %x\n", hr->qstatcnt);
	printf("hr->unit     = %x\n", hr->unit);
	printf("hr->done     = %x\n", hr->done);
	printf("hr->status   = %x\n\n", hr->status);
}
#endif NHPIB > 0
