/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * ISA & EISA bus manager
 */

#include "eisa.h"
#if NEISA > 0

#include <sys/param.h>
#include <sys/syslog.h>
#include <hpsgc/bsd_glue.h>
#include <hpsgc/atbus.h>
#include <hpsgc/iomap.h>
#include <hpsgc/scsi.h>
#include <hpsgc/atbus_common.h>
#include <machine/asp.h>
#define	assert(bool)

/* 128k for devices that use DMA channels */
#define IOMAP_DMA_SIZE		0x020000
#define IOMAP_8_BIT_SIZE	0x010000
#define IOMAP_16_BIT_SIZE	0x020000

#define ATBUS_CSR_SIZE		8			/* Size of an atbus csr for mapping */

#define ATBUS_CSR_LOW_IOVA	((vm_offset_t)0x100)
#define ATBUS_CSR_HIGH_IOVA	((vm_offset_t)0x400)
#define EISA_CSR_LOW_IOVA	((vm_offset_t)0x1000)
#define EISA_CSR_HIGH_IOVA	((vm_offset_t)0x10000)

typedef struct {
	unsigned long	ent;
	unsigned char	pad[0x1000-sizeof(unsigned long)];
} iomap_entry_t;

static unsigned long invalid_entry = 0;         	/* 0 will cause ref to go nowhere */

#define IOMAP_IOADDR_LOW	(0x000100000)
#define IOMAP_IOADDR_HIGH	(0x000500000)

#define ATBUS_LOW_IOVA		((vm_offset_t)0x00080000)
#define ATBUS_HIGH_IOVA		((vm_offset_t)0x10000000)

#define ATBUS_CSR_MAPPING_SIZE	0x1000

/* IOMAP index layout */
#define IOMAP_LOW_MASTER	((IOMAP_IOADDR_LOW-IOMAP_IOADDR_LOW) >> ATBUS_IOMAP_PAGE_SHIFT)
#define IOMAP_HIGH_MASTER	((IOMAP_IOADDR_HIGH-IOMAP_IOADDR_LOW-IOMAP_DMA_SIZE) >> ATBUS_IOMAP_PAGE_SHIFT)
#define IOMAP_LOW_8_BIT		IOMAP_HIGH_MASTER
#define IOMAP_HIGH_8_BIT	(IOMAP_LOW_8_BIT+(IOMAP_8_BIT_SIZE >> ATBUS_IOMAP_PAGE_SHIFT))
#define IOMAP_LOW_16_BIT	IOMAP_HIGH_MASTER
#define IOMAP_HIGH_16_BIT	(IOMAP_LOW_16_BIT+(IOMAP_16_BIT_SIZE >> ATBUS_IOMAP_PAGE_SHIFT))
#define IOMAP_LOW_32_BIT	IOMAP_LOW_MASTER
#define IOMAP_HIGH_32_BIT	IOMAP_HIGH_MASTER

#define IOMAP_NENTRIES		((IOMAP_IOADDR_HIGH-IOMAP_IOADDR_LOW) >> ATBUS_IOMAP_PAGE_SHIFT)

/* convert Physical address to IOMAP page quantity */
#define PAGE(x)		((unsigned)(x) >> ATBUS_IOMAP_PAGE_SHIFT)
/* convert IOVA to 0 based iomap index */
#define INDEX(x)	(((unsigned)(x)-IOMAP_IOADDR_LOW) >> ATBUS_IOMAP_PAGE_SHIFT)
/* convert 0 based iomap index to IOVA */
#define ADDR(x)		(((vm_offset_t)((x) << ATBUS_IOMAP_PAGE_SHIFT) + IOMAP_IOADDR_LOW))

#define ATBUS_NINTS		16		/* number of interrupts */
/* page register index values */
static	u_char	prindex[] = { 7, 3, 1, 2, 0, 11, 9, 10 };
#define PRSET(chan)	atbusinfo[0].control->isa.pr.all[prindex[(chan)]] = ((unsigned)ioaddr)>>16;

/* EISA/ATBUS adapter specific control */
#define IO_EISA_ADAP_PHYS	0xfc010000
#define IO_EISA_ADAP_LEN	0x1000

struct at_adap_control {
	unsigned char	pad;
	unsigned char	lock;
	unsigned char	liowait;
	unsigned char	speed_jumper;
#define ADAP_25	0x00
#define ADAP_33	0x01
};

/* only bit 1 valid */
#define ADAP_MASK(x)	((x) & 1)

union eisa_control {
	struct at_control isa;
	struct {
		struct {
			char pad[0x4d0];
			volatile unsigned char pic0;
#define EISA_INTCTL0_ALL		0xf8		/* 0, 1, 2 edge, level for the all others */
			volatile unsigned char pic1;
#define EISA_INTCTL1_ALL		0xdf		/* 13 edge, level for all others */ 
		} intctl;
		struct {
			char pad[0x460];
			volatile unsigned char nmi;
			volatile unsigned char nmi_ext;
#define EISA_NMI_BUSRESET		0x01
#define EISA_NMI_IOPORT_ENAB		0x02
#define EISA_NMI_NMI_ENAB		0x04
#define EISA_NMI_MASTER_TIMEOUT_ENAB	0x08
#define EISA_NMI_RES4			0x10
#define EISA_NMI_IOPORT_INT		0x20
#define EISA_NMI_MASTER_INT		0x40
#define EISA_NMI_INT			0x80
		} nmi;
	} eisa;
};

/*
 * Per-bus info
 */
struct atbusinfo {
	int		configured;
	int		debug_trace;	/* -1 traces all interrupts, bit N traces interrupts on IRQ N */
	int		spurcnt;
	int 		active[ATBUS_NINTS];
	union eisa_control *control;
	union dma0	*dma0;
	union dma1	*dma1;
	int		dma0_tc;
	int		dma1_tc;
	iomap_entry_t	*iomap;
	char		iomap_allocated[IOMAP_NENTRIES];
	struct at_adap_control *adap;
	unsigned char	*iack;
} atbusinfo[NEISA];

char atbus_canshare[ATBUS_NINTS] = {
	0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
};

/*
 * DMA support
 */

void
#ifndef	_NO_PROTO
atbus_dma_start(int chan, vm_offset_t ioaddr, unsigned long count, int opts, int dir)
#else
atbus_dma_start(chan, ioaddr, count, opts, dir)
	int chan;
	vm_offset_t ioaddr;
	unsigned long count;
	int opts, dir;
#endif
{
	unsigned char mode;
	int	subchan = chan & 0x3;

	if (chan == 4 || ((unsigned)chan) >= ATBUS_NCHANS)
		panic("atbus_dma_start: invalid channel\n");

	/* channel number in chip, single byte transfer */
	mode = MODE(chan);

	switch (dir) {
	case ATBUS_READ:
		mode |= MODE_WRITE;	/* reading a device means writing memory */
		break;
	case ATBUS_WRITE:
		mode |= MODE_READ;
		break;
	default:
		panic("atbus_dma_start: invalid direction");
		break;
	}

	if (opts & ATBUS_AUTOINIT)
		mode |= MODE_AUTOINIT;
	if (opts & ATBUS_DECREMENT)
		mode |= MODE_DECREMENT;
	if (opts & ATBUS_CASCADE)
		mode |= MODE_CASCADE;
	else
		mode |= MODE_SINGLE;

	/* eisa opts */
	if (opts & ATBUS_TYPEA)
		;
	if (opts & ATBUS_TYPEB)
		;
	if (opts & ATBUS_BURST)
		;
	if (opts & ATBUS_BYTE)
		;
	if (opts & ATBUS_WORD)
		;
	if (opts & ATBUS_LONG)
		;

/* I hate writing/maintaining things twice */
#define DMASET(chip) { \
	/* mask this channel just in case */	\
	(chip)->write.mask_channel = 0x4 | subchan;			delay(1); \
	(chip)->write.clear_byte_pointer = 0;				delay(1); \
	(chip)->write.channel[subchan].address = (unsigned)ioaddr;	delay(1); \
	(chip)->write.channel[subchan].address = (unsigned)ioaddr>>8;	delay(1); \
	PRSET(chan);							delay(1); \
	count--;								 \
	(chip)->write.clear_byte_pointer = 0;				delay(1); \
	(chip)->write.channel[subchan].count = count;			delay(1); \
	(chip)->write.channel[subchan].count = count>>8;		delay(1); \
	(chip)->write.mode = mode;					delay(1); \
	}

	if (chan < 4) {
		DMASET(atbusinfo[0].dma0);
	}
	else {
		DMASET(atbusinfo[0].dma1);
	}
}

unsigned long
#ifndef _NO_PROTO
atbus_dma_stop(int chan)
#else
atbus_dma_stop(chan)
	int chan;
#endif
{
	int subchan = chan & 0x3;
	unsigned long resid;

#define DMARMASK(x)	((x)&0xffff)
#define DMADONE(chip, tc) {	\
	(chip)->write.mask_channel = 0x4 | subchan;	delay(1); \
	(tc) |= (chip)->read.status;		delay(1); \
	if ((tc) & (1<<subchan)) {		\
		(tc) &= ~(1<<subchan);		\
		return (0);				\
	}						\
	resid = (chip)->read.channel[subchan].count;		delay(1); \
	resid |= (chip)->read.channel[subchan].count << 8;	delay(1); \
	}

	if (chan == 4 || ((unsigned)chan) >= ATBUS_NCHANS)
		panic("atbus_dma_stop: invalid channel\n");

	if (chan < 4) {
		DMADONE(atbusinfo[0].dma0, atbusinfo[0].dma0_tc);
	}
	else {
		DMADONE(atbusinfo[0].dma1, atbusinfo[0].dma1_tc);
	}

	if (DMARMASK(resid) == DMARMASK(-1))
		return (0);

	return (++resid);
}

/*
 * IOMAP support
 */

void
#ifndef _NO_PROTO
atbus_iomap_map(vm_offset_t iova, vm_offset_t *phys, int npages)
#else
atbus_iomap_map(iova, phys, npages)
	vm_offset_t iova;
	vm_offset_t *phys;
	int npages;
#endif
{
	int iomap_index;

	iomap_index = INDEX(iova);
	while (npages--)
		atbusinfo[0].iomap[iomap_index++].ent = PAGE(*phys++);
}

void
#ifndef _NO_PROTO
atbus_iomap_unmap(vm_offset_t iova, int npages)
#else
atbus_iomap_unmap(iova, npages)
	vm_offset_t iova;
	int npages;
#endif
{
	int iomap_index;

	iomap_index = INDEX(iova);
	while (npages--)
		atbusinfo[0].iomap[iomap_index++].ent = invalid_entry;
}

vm_offset_t
#ifndef _NO_PROTO
atbus_iomap_allocate(int type, vm_offset_t addr, int npages)
#else
atbus_iomap_allocate(type, addr, npages)
	int type;
	vm_offset_t addr;
	int npages;
#endif
{
	int iomap_index, base, target;

	/* if we know where we want it see if there is room */
	if (addr != ATBUS_IOMAP_ANYWHERE) {
		target = INDEX(addr);
		base = target;
	}
	else switch (type) {

	case ATBUS_IOMAP_MASTER:
		base = IOMAP_LOW_MASTER;
		target = IOMAP_HIGH_MASTER;
		break;
	
	case ATBUS_IOMAP_8_BIT:
		base = IOMAP_LOW_8_BIT;
		target = IOMAP_HIGH_8_BIT;
		break;

	case ATBUS_IOMAP_16_BIT:
		base = IOMAP_LOW_16_BIT;
		target = IOMAP_HIGH_16_BIT;
		break;

	case ATBUS_IOMAP_32_BIT:
		base = IOMAP_LOW_32_BIT;
		target = IOMAP_HIGH_32_BIT;
		break;

	default:
		panic("atbus_iomap_allocate: unknown type requested\n");
		return(ATBUS_IOMAP_NOSPACE);
	}

	assert(target <= IOMAP_NENTRIES);

	iomap_index = iomap_allocate(atbusinfo[0].iomap_allocated, base, target, npages);
	if (iomap_index == IOMAP_NOSPACE)
		return(ATBUS_IOMAP_NOSPACE);
	else
		return(ADDR(iomap_index));
}

vm_offset_t
atbus_iomap_allocate_wait(type, addr, npages)
int		type;
vm_offset_t	addr;
int		npages;
{
	vm_offset_t iova;

	while ((iova = atbus_iomap_allocate(type, addr, npages)) == ATBUS_IOMAP_NOSPACE) {
		sleep((vm_offset_t)atbusinfo[0].iomap_allocated, PRIBIO);
	}

	return(iova);
}

void
#ifndef _NO_PROTO
atbus_iomap_free(vm_offset_t iova, int npages)
#else
atbus_iomap_free(iova, npages)
	vm_offset_t iova;
	int npages;
#endif
{
	assert(iova >= IOMAP_IOADDR_LOW && iova < IOMAP_IOADDR_HIGH);
	iomap_free(atbusinfo[0].iomap_allocated, INDEX(iova), npages);
	wakeup((vm_offset_t)atbusinfo[0].iomap_allocated);
}

/*
 * Interrupt support
 */
static void
#ifndef _NO_PROTO
atbus_intr_disable(int busnum, int id)	/* XXX needs rewrite for multiple AT buses */
#else
atbus_intr_disable(busnum, id)	/* XXX needs rewrite for multiple AT buses */
	int busnum, id;
#endif
{
	union pic	*pic;
	u_char		piclev;

#ifdef DBG
	if (((unsigned)id) >= ATBUS_NINTS || !atbusinfo[0].active)
		panic("atbus_intr_disable: bad id or attempt to disable undefined interrupt");
#endif
	if (id < ATBUS_PIC_NINTS) {
		pic = &atbusinfo[0].control->isa.pic0;
		piclev = id;
	}
	else {
		pic = &atbusinfo[0].control->isa.pic1;
		piclev = id - ATBUS_PIC_NINTS;
	}
	pic->p.imr |= (1<<piclev);		/* disable: add to interrupt mask */
}

static void
#ifndef _NO_PROTO
atbus_intr_enable(int busnum, int id)  /* XXX needs rewrite for multiple AT buses */
#else
atbus_intr_enable(busnum, id)  /* XXX needs rewrite for multiple AT buses */
	int busnum, id;
#endif
{
	union pic	*pic;
	u_char		piclev;

#ifdef DBG
	if (((unsigned)id) >= ATBUS_NINTS || !atbusinfo[0].active)
		panic("atbus_intr_enable: bad id or attempt to enable undefined interrupt");
#endif
	if (id < ATBUS_PIC_NINTS) {
		pic = &atbusinfo[0].control->isa.pic0;
		piclev = id;
	}
	else {
		pic = &atbusinfo[0].control->isa.pic1;
		piclev = id - ATBUS_PIC_NINTS;
	}
	pic->p.imr &= ~(1<<piclev);		/* enable: remove from interrupt mask */
}

/*
 * atbus_resolver - sets up interrupt handler
 *                      called from platform-independent 'handler_add()'
 */

atbus_resolver(action, bus, level, type)
	int action, bus, level, type;
{
        struct atbusinfo *info;

#ifdef DBG
        if (((unsigned)bus) >= NEISA)
                panic("(atbus_resolver): bad AT bus index");
	if (((unsigned)level) >= ATBUS_NINTS)
		panic("(atbus_resolver): bad level");
#endif
        info = &atbusinfo[bus];
	if (!info->configured) {
		return(1);
	}

        switch (action) {
	case RESOLVER_ADD_HANDLER:
	        if (!info->active[level]) {
                        info->active[level]++;
                        atbus_intr_enable(bus, level);
	        }
	        else {
                        /* already have at least one handler registered */
#ifdef DBG
                        if (!atbus_canshare[level])
	                        panic("(atbus_resolver): attempt to overwrite used interrupt level");
#endif /* DBG */
                        info->active[level]++;
	        }
		if (type & ATBUS_RESOLVER_LEVEL_INT) {
			/* level sensitive interrupt line */
			if (level < ATBUS_PIC_NINTS)
				info->control->eisa.intctl.pic0 |= (1 << level);
                        else
				info->control->eisa.intctl.pic1 |= (1 << (level-ATBUS_PIC_NINTS));
		}
		else {
			/* edge sensitive interrupt line */
			if (level < ATBUS_PIC_NINTS)
				info->control->eisa.intctl.pic0 &= ~(1 << level);
                        else
				info->control->eisa.intctl.pic1 &= ~(1 << (level-ATBUS_PIC_NINTS));
		}
                break;

	case RESOLVER_DEL_HANDLER:
#ifdef DBG
                if (!info->active[level])
                        panic("(atbus_resolver): deleting non-configured interrupt level");
#endif
                if (--info->active[level] == 0)
                        atbus_intr_disable(bus, level);
                break;

        default:
                panic("(atbus_resolver): unrecognized action");
        }

        return(0);
}

atbus_nmi(bus, ef)
	int bus;
        void *ef;
{
	printf("atbus_nmi: got EISA NMI!  (help me, help me)\n");
}

/* This is called as an interrupt handler for the SYSBUS */
int
atbus_intr(bus)
	int bus;
{

	struct scsi_controller  *scp;
        struct atbusinfo *info;
	union pic *pic, *master_pic;
	unsigned int piclev, id, ints;

	info = &atbusinfo[bus];
	master_pic = &info->control->isa.pic0;

	id = *info->iack;			/* fetch interrupting level and iack */
	if (id >= ATBUS_NINTS)
                panic("atbus_intr: invalid interrupt id");

#ifdef notyet
	/* check if this is us interrupting */
	master_pic->r.r0 = ATBUS_PIC_READ_IREQ;	/* read master IRR to see if there are any unmasked interrupts */
	ints = master_pic->p.irr;
	if ((ints & ~master_pic->imr) == 0) {
		return(0);
	}
#endif

	/* check to see if this is a spurious interrupt */
	master_pic->r.r0 = ATBUS_PIC_READ_INSERV;
	ints = master_pic->p.irr;
        if (id >= ATBUS_PIC_NINTS) {
                /* slave pic on master, this is OK */
                piclev = 2;
        }
        else {
                piclev = id;
        }

        if ((ints & (1<<piclev)) == 0) {
                switch (piclev) {
		case 7:
		case 15:
			/* spurious bus interrupts */
			if ((++info->spurcnt & (256-1)) == 0) {
#ifdef notyet
				log(LOG_WARNING, "atbus_intr: spurious interrupt atbus %d count %d\n",
					ih->ih_busnum, info->spurcnt);
#endif
			}
			break;
		default:
			log(LOG_WARNING, "atbus_intr: stray interrupt atbus %d\n", bus);
		}
		return(1);			/* we did interrupt, but board is gone */
	}

        if (!info->active[id]) {
                printf("atbus_intr: interrupt on inactive line\n");
                return(1);
        }

	if (id < ATBUS_PIC_NINTS) {
		pic = master_pic;
		piclev = id;
	}
	else {
		pic = &info->control->isa.pic1;
		piclev = id - ATBUS_PIC_NINTS;
		pic->r.r0 = ATBUS_PIC_EOI;	/* eoi slave before master */
	}

	master_pic->r.r0 = ATBUS_PIC_EOI;		/* eoi master always */

	pic->p.imr |= (1<<piclev);		/* disable: add to interrupt mask */
	{
		extern struct scsi_controller ncr53c700_controller[];
		extern int eisa_irqs[], eisa_scno[], eisa_msiz;
		int i;

		scp = NULL;
		for (i = 1; i < eisa_msiz; i++)
			if (id == eisa_irqs[i] && eisa_scno[i] != -1) {
				scp = &ncr53c700_controller[eisa_scno[i]];
				break;
			}
		if (scp == NULL) {
                	printf("atbus_intr: no handler for intr %d\n", id);
			return(1);
		}
		(void) ncr53c700_intr(scp);
	}
	pic->p.imr &= ~(1<<piclev);		/* enable: turn off mask */

	return(1);
}

vm_offset_t
#ifndef _NO_PROTO
atbus_xlate_iova(vm_offset_t iova, vm_offset_t basepaddr)
#else
atbus_xlate_iova(iova, basepaddr)
	vm_offset_t iova;
	vm_offset_t basepaddr;
#endif
{
	if (iova >= ATBUS_CSR_LOW_IOVA && iova < ATBUS_CSR_HIGH_IOVA)
		return(basepaddr+(((vm_offset_t)iova/ATBUS_CSR_SIZE) * ATBUS_CSR_MAPPING_SIZE));
	else if (iova >= EISA_CSR_LOW_IOVA && iova < EISA_CSR_HIGH_IOVA)
		return(basepaddr+(vm_offset_t)iova);
	else if (iova < ATBUS_HIGH_IOVA)
		return(basepaddr+(vm_offset_t)iova);
	else
		panic("atbus_xlate_iova: bad io address\n");
}

/*
 * atbus_probe - check for existance of bus, and return indication
 */
#ifndef _NO_PROTO
atbus_probe(hp_ctlrinfo_t *ctlrinfop)
#else
atbus_probe(ctlrinfop)
	hp_ctlrinfo_t *ctlrinfop;
#endif
{
	/* check bus validity */
	if (((unsigned int)ctlrinfop->ctlrnum) >= NEISA) {
		printf("atbus_probe: invalid AT bus index\n");
		return(0);
	}

/* XXX */
	atbus_attach(ctlrinfop);
	return(1);
}


/*
 * atbus_attach - actually configure AT bus
 */

#ifndef _NO_PROTO
atbus_attach(hp_ctlrinfo_t *ctlrinfop)
#else
atbus_attach(ctlrinfop)
	hp_ctlrinfo_t *ctlrinfop;
#endif
{
	int ii;
        struct atbusinfo *info;
	union pic *pic, *master_pic;
	unsigned int piclev, id, ints;

	/* figure out which EISA bus we need to look at */
	info = &atbusinfo[ctlrinfop->ctlrnum];

        if (info->configured)
		return(0);

	/* initialize the hardware */
        info->configured = TRUE;

	info->control = (union eisa_control *)ctlrinfop->vcsrs[0];

	/* map iack register space */
	info->iack = (unsigned char *)io_map(IO_EISA_IACK_PHYS, IO_EISA_IACK_LEN, "eisa/atbus iack");
	if (!info->iack)
		panic("atbus_attach: unable to map eisa/atbus iack page");

	/* map & initialize iomap */
	info->iomap = (iomap_entry_t *)io_map(atbus_xlate_iova((vm_offset_t)IOMAP_IOADDR_LOW, ctlrinfop->csrs[0]), IOMAP_NENTRIES*sizeof(iomap_entry_t), "eisa/atbus iomap");
	if (!info->iomap)
		panic("atbus_attach: unable to map eisa/atbus iomap");

	atbus_iomap_unmap((vm_offset_t)IOMAP_IOADDR_LOW, IOMAP_NENTRIES);
	bzero((void *)info->iomap_allocated, sizeof(info->iomap_allocated));

	/* map adapter specific regs */
	info->adap = (struct at_adap_control *)io_map(IO_EISA_ADAP_PHYS, sizeof(*info->adap), "mongoose regs");
	if (!info->adap)
		panic("atbus_attach: unable to map eisa/atbus adapter control page");

	printf("eisa%d: Adapter speed %s Mhz.\n", ctlrinfop->ctlrnum, ADAP_MASK(info->adap->speed_jumper) == ADAP_25 ? "25" : "33");
	info->adap->liowait = 1;			/* de-assert liowait */
	info->adap->lock = 1;				/* unlock bus */

	/* issue bus reset */
	info->control->eisa.nmi.nmi_ext = EISA_NMI_BUSRESET;
	DELAY(1);
	info->control->eisa.nmi.nmi_ext = 0;
	/* wait a bit for reset to settle down */
	DELAY(100);

	/* initialize pics */
	info->control->isa.pic0.r.r0 = 0x11;		/* init config data: edge triggered, icw4 needed */
	info->control->isa.pic0.r.r1 = 0x00;		/* no interrupt vector */
	info->control->isa.pic0.r.r1 = 0x04;		/* icw3: master (slave on irq2) */
	info->control->isa.pic0.r.r1 = 0x11;		/* icw4: mode is fully nested and normal eoi */
	info->control->isa.pic0.r.r1 = 0xff;		/* ocw1: disable all interrupts */

	info->control->isa.pic1.r.r0 = 0x11;		/* init config data: edge triggered, icw4 needed */
	info->control->isa.pic1.r.r1 = 0x08;		/* interrupt vector */
	info->control->isa.pic1.r.r1 = 0x02;		/* icw3: slave's interrupt request id */
	info->control->isa.pic1.r.r1 = 0x11;		/* icw4: mode is fully nested and normal eoi */
	info->control->isa.pic1.r.r1 = 0xff;		/* ocw1: disable all interrupts */

	/* edge sensitive interrupts for all bus available lines default. */
        info->control->eisa.intctl.pic0 = 0;
        info->control->eisa.intctl.pic1 = 0;

	/* perform interrupt acknowledge to get the bus ready */
	ii = 0;
	while (ii < 3) {
		id = *info->iack;
		if (id == 7)
			ii++;
		if (id >= ATBUS_PIC_NINTS)
			/* eoi slave before master */
			info->control->isa.pic1.r.r0 = ATBUS_PIC_EOI;
		/* eoi master always */
		info->control->isa.pic0.r.r0 = ATBUS_PIC_EOI;
	}

	/* refresh disabled, timers off */
	info->control->isa.timers[0].control = 0x50;
	info->control->isa.timers[0].control = 0x10;
	info->control->isa.timers[1].control = 0x10;
	info->control->isa.timers[1].control = 0x90;

	/* initialize dma */
	info->dma0 = &info->control->isa.dma0;
	info->dma1 = &info->control->isa.dma1;

	/* master clear the chips */
	info->dma0->write.master_clear = 0;			delay(1);
	info->dma1->write.master_clear = 0;			delay(1);

	/* 
	 * Set the command registers to 0 this means:
	 * memory to memory disabled
	 * channel 0 hold disabled
	 * controller enabled
	 * normal timing
	 * fixed priority
	 * late write selection
	 * DREQ active high
	 * DACK active low
	 */
	info->dma0->write.command = 0;			delay(1);
	info->dma1->write.command = 0;			delay(1);

	/* mask all channels */
	info->dma0->write.mask_write = 0xff;			delay(1);
	info->dma1->write.mask_write = 0xff;			delay(1);

	/* set channels 0-3 , 5-7 to single xfer, 4 to cascade */
	info->dma0->write.mode = MODE(0)|MODE_SINGLE;		delay(1);
	info->dma0->write.mode = MODE(1)|MODE_SINGLE;		delay(1);
	info->dma0->write.mode = MODE(2)|MODE_SINGLE;		delay(1);
	info->dma0->write.mode = MODE(3)|MODE_SINGLE;		delay(1);

	info->dma1->write.mode = MODE(0)|MODE_CASCADE;	delay(1);
	info->dma1->write.mode = MODE(1)|MODE_SINGLE;		delay(1);
	info->dma1->write.mode = MODE(2)|MODE_SINGLE;		delay(1);
	info->dma1->write.mode = MODE(3)|MODE_SINGLE;		delay(1);

	/* enable channel 0 on second controller for cascade */
	info->dma1->write.mask_channel = 0;			delay(1);

	aspitab(INT_EISA_NMI, SPLVM, atbus_nmi, (int)ctlrinfop->ctlrnum, 0);
	aspitab(INT_EISA, SPLEISA, atbus_intr, (int)ctlrinfop->ctlrnum, 0);
	return(1);
}

atbus_slave()
{
	panic("atbus_slave: called!");
}

#endif	/* NEISA > 0 */
