/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

#ifndef	_HP_ATBUS_H_
#define _HP_ATBUS_H_

#ifdef OSF1_KERNEL
#include <hp/businfo.h>
#include <hp/handler.h>
#endif

/*
 * DMA support
 */

/* opts */
#define ATBUS_AUTOINIT	0x0001		/* channel auto initializes */
#define ATBUS_DECREMENT	0x0002		/* addresses decrement */
#define ATBUS_CASCADE	0x0004		/* cascade: bus mastership */

#define ATBUS_TYPEA	0x0008		/* eisa typeA dma */
#define ATBUS_TYPEB	0x0010		/* eisa typeB dma */
#define ATBUS_BURST	0x0020		/* eisa burst mode dma */

#define ATBUS_BYTE	0x0040		/* eisa channels can select size */
#define ATBUS_WORD	0x0080
#define ATBUS_LONG	0x0100

/* dir */
#define ATBUS_READ	0
#define ATBUS_WRITE	1

#ifndef _NO_PROTO
void
atbus_dma_start(int chan, vm_offset_t ioaddr, unsigned long count, int opts, int dir);

unsigned long
atbus_dma_stop(int chan);		/* returns residual count */
#endif


/*
 * IOMAP support
 */

#define ATBUS_IOMAP_PAGE_SHIFT		12              /* (2 ^ 12) = 4096 */
#define ATBUS_IOMAP_PAGE_SIZE		4096
#define ATBUS_IOMAP_PAGE_MASK		(ATBUS_IOMAP_PAGE_SIZE - 1)

#define ATBUS_IOMAP_MASTER		1
#define ATBUS_IOMAP_8_BIT		2
#define ATBUS_IOMAP_16_BIT		3
#define ATBUS_IOMAP_32_BIT		4

#define ATBUS_IOMAP_ANYWHERE		((vm_offset_t)(-1))
#define ATBUS_IOMAP_NOSPACE		((vm_offset_t)(-1))

#ifndef _NO_PROTO
vm_offset_t atbus_iomap_allocate(int type, vm_offset_t addr, int npages);
vm_offset_t atbus_iomap_allocate_wait(int type, vm_offset_t addr, int npages);
void atbus_iomap_free(vm_offset_t iova, int npages);

void atbus_iomap_map(vm_offset_t iova, vm_offset_t *phys, int npages);
void atbus_iomap_unmap(vm_offset_t iova, int npages);
#endif

/*
 * ATBUS (ISA/EISA) interrupt support routines
 */

#ifndef _NO_PROTO
int atbus_resolver(int action, ihandler_t *ih);
#endif

/* resolver flags (passed in rparam) */
#define ATBUS_RESOLVER_LEVEL_INT	0x01		/* level sensitive interrupt line, default is edge */

#ifndef _NO_PROTO
/* On sau13 atbus_intr() is an interrupt handler on the sysbus */
int atbus_intr(ihandler_t *ih, void *ef);

vm_offset_t
atbus_xlate_iova(vm_offset_t iova, vm_offset_t baseaddr);	/* given atbus iova, return physical address */
#endif

#endif	/* _HP_ATBUS_H_ */
