/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * Copyright (c) 1992 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * THE UNIVERSITY OF UTAH AND CSL PROVIDE THIS SOFTWARE IN ITS "AS IS"
 * CONDITION, AND DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM ITS USE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: alignment.c 1.5 94/11/10$
 */

#ifdef	HPOSFCOMPAT
#include <kern/thread.h>
#include <mach/exception.h>
#include <mach/vm_param.h>
#include <machine/trap.h>
#include <machine/psl.h>
#include <machine/regs.h>
#include <machine/spl.h>
#include <machine/break.h>
#include <mach_kdb.h>
#include <machine/spl.h>
#include <machine/pdc.h>
#include <machine/iomod.h>

static int align_ldstx[] = {
	0x0,		/* no alignment	    */
	0x1,		/* 2-byte alignment */
	0x3,		/* 4-byte alignment */
	0xf		/* f-byte alignment */
};

/*
 * aligment_fault - Decipher aligment fault.
 *
 * Data memory protection traps and unaligned reference traps come
 * in on the same trap number.  This routine is passed the save state
 * and figures out if an aligment fault occurred.
 */
boolean_t
alignment_fault(ssp) 
	struct hp800_saved_state *ssp;
{
        boolean_t result = FALSE;

        if ((OPCODE_IS_STORE(ssp->iir)) || (OPCODE_IS_LOAD(ssp->iir))) {
                switch (OPCODE(ssp->iir)) {

                case OP_LDH:
                case OP_STH:
                        if (ssp->ior & 0x1) {
                                result = TRUE;
                        }
                        break;

                case OP_LDW:
                case OP_LDWM:
                case OP_STW:
                case OP_STWM:
                        if (ssp->ior & 0x3) {
                                result = TRUE;
                        }
                        break;

		case OP_LDSTX:
        		if (ssp->ior & align_ldstx[SIZE_OP_LDSTX(ssp->iir)]) {
                                result = TRUE;
                        }
                        break;
			
		}
        }
        return(result);
}

/*
 * handle_alignment_fault - Handle user alignment fault.
 */
int
handle_alignment_fault(ssp)
	struct hp800_saved_state *ssp;
{
	int		result;
        unsigned long	opcode = ssp->iir;

	if (!alignment_fault(ssp))
		return(0);
	
        switch (OPCODE(opcode)) {
		
        case OP_LDW:
                result = align_load(opcode, ssp, 4);
                break;
		
        case OP_LDH:
                result = align_load(opcode, ssp, 2);
                break;
		
        case OP_LDWM:
                result = align_load_modify(opcode, ssp);
		break;

        case OP_STW:
                result = align_store(opcode, ssp, 4);
                break;
		
        case OP_STH:
                result = align_store(opcode, ssp, 2);
                break;

        case OP_STWM:
                result = align_store_modify(opcode, ssp);
		break;

        case OP_LDSTX:
                result = align_load_indexed(opcode, ssp);
		break;
	}
	
	if (!result) {
		ssp->ipsw |= PSW_N;
		return(1);
	}
        return(0);
}

/*
 * trap_instr_imm - Convert 2s complement (with sign bit on rhs) into integer.
 *
 * Mask specifies the size of the immediate.
 */
unsigned long
trap_instr_imm(i, mask)
	unsigned long i;
	unsigned long mask;
{
	if (i & 0x1) 
                return(-(((~(i >> 1)) + 1) & mask));
	else 
		return((i >> 1) & mask);
}

/* 
 * align_load - Handle regular load alignment fault.
 */
align_load(opcode, ssp, size)
	unsigned long	opcode;
	unsigned long	*ssp;		/* XXX */
	int		size;
{
	unsigned char *reg_ptr;

	reg_ptr = (unsigned char *) &ssp[OPCODE_LMEM_TARGET(opcode)];
	
	if (size == 2) {
		/*
	 	 * Halfword load zeros upper 16 bits.
		 */
		bzero(reg_ptr, 4);
		reg_ptr += 2;
	}
        return(copyin((ssp[OPCODE_LMEM_BASE(opcode)] +
		       trap_instr_imm(OPCODE_LMEM_IMM(opcode), OPCODE_L_IMASK)),
		      reg_ptr, size));
}

/* 
 * align_load_modify - Handle modified load alignment fault.
 */
align_load_modify(opcode, ssp)
	unsigned long	opcode;
	unsigned long	*ssp;		/* XXX */
{
        int		imm, result;
	unsigned long	addr;

	imm  = trap_instr_imm(OPCODE_LMEM_IMM(opcode), OPCODE_L_IMASK);
	addr = ssp[OPCODE_LMEM_BASE(opcode)];

	if (imm < 0)		/* Pre-decrement */
		addr += imm;

	result = copyin(addr, &ssp[OPCODE_LMEM_TARGET(opcode)], 4);

	if (imm > 0)		/* Post-increment */
		addr += imm;

	ssp[OPCODE_LMEM_BASE(opcode)] = addr;
	return(result);
}

/* 
 * align_store_modify - Handle modified store alignment fault.
 */
align_store_modify(opcode, ssp)
	unsigned long	opcode;
	unsigned long	*ssp;		/* XXX */
{
        int		imm, result;
	unsigned long	addr;

	imm  = trap_instr_imm(OPCODE_LMEM_IMM(opcode), OPCODE_L_IMASK);
	addr = ssp[OPCODE_LMEM_BASE(opcode)];
	
	if (imm < 0)		/* Pre-decrement */
		addr += imm;

	result = copyout(&ssp[OPCODE_LMEM_SOURCE(opcode)], addr, 4);
	
	if (imm > 0)		/* Post-increment */
		addr += imm;

	ssp[OPCODE_LMEM_BASE(opcode)] = addr;
	return(result);
}

/* 
 * trap_align_store - Handle regular store alignment fault.
 */
align_store(opcode, ssp, size)
	unsigned long 	opcode;
	unsigned long	*ssp;		/* XXX */
	int		size;
{
        return(copyout(((unsigned long)(&ssp[OPCODE_LMEM_SOURCE(opcode)]) +
			(4 - size)),
		       (ssp[OPCODE_LMEM_BASE(opcode)] +
			trap_instr_imm(OPCODE_LMEM_IMM(opcode),
				       OPCODE_L_IMASK)),
		       size));
}

align_load_indexed(opcode, ssp)
	unsigned long 	opcode;
	unsigned long	*ssp;		/* XXX */
{
	int 	result;
	
	if (BIT_19(opcode)) {
		switch (OP_EXT4_TYPE(opcode)) {
			
		case OP_EXT4_LDWS:
			result = align_short_load(opcode, ssp, 4);
			break;
			
		case OP_EXT4_LDHS:
			result = align_short_load(opcode, ssp, 2);
			break;
			
		case OP_EXT4_STWS :
			result = align_short_store(opcode, ssp, 4);
			break;
			
		case OP_EXT4_STHS :
			result = align_short_store(opcode, ssp, 2);
			break;
		}
	} else {
		switch (OP_EXT4_TYPE(opcode)) {
		case OP_EXT4_LDWX :
			result = align_indexed_load(opcode, ssp, 2);
			break;
		case OP_EXT4_LDHX :
			result = align_indexed_load(opcode, ssp, 1);
			break;
		} 
	}
	return(result);
}

/*
 * trap_align_short_load - Handle short load alignment fault.
 */
align_short_load(opcode, ssp, size)
	unsigned long	opcode;
	unsigned long	*ssp;		/* XXX */
	int 		size;
{
        int 		imm, result;
        unsigned long 	from;
	unsigned char 	*reg_ptr;

        imm  = trap_instr_imm(OPCODE_SLMEM_IMM(opcode), OPCODE_S_IMASK);
        from = ssp[OPCODE_SLMEM_BASE(opcode)];
	
	/* Modify base register before */
        if ((OPCODE_M_UPDATE(opcode) && OPCODE_A_UPDATE(opcode)) ||
	    (!OPCODE_M_UPDATE(opcode)))
                from += imm;

	reg_ptr = (unsigned char *)&ssp[OPCODE_SLMEM_TARGET(opcode)];

	if (size == 2) {
		/*
	 	 * Halfword load zeros upper 16 bits.
		 */
		bzero(reg_ptr, 4);
		reg_ptr += 2;
	}
	result = copyin(from, reg_ptr, size);
	
        if (!result && OPCODE_M_UPDATE(opcode)) {
		/* Modify base register after */		
                if (!OPCODE_A_UPDATE(opcode))
                        from += imm;
                ssp[OPCODE_SLMEM_BASE(opcode)] = from;
        }
        return(result);
}

/*
 * trap_align_short_store - Handle short store alignment fault.
 */
align_short_store(opcode, ssp, size)
	unsigned long	opcode;
	unsigned long	*ssp;		/* XXX */
	int 		size;
{
        int 		imm, result;
        unsigned long 	to;

        imm = trap_instr_imm(OPCODE_SSMEM_IMM(opcode), OPCODE_S_IMASK);
        to  = ssp[OPCODE_SSMEM_BASE(opcode)];
	
	/* Modify base register before */
        if ((OPCODE_M_UPDATE(opcode) && OPCODE_A_UPDATE(opcode)) ||
	    (!OPCODE_M_UPDATE(opcode)))
                to += imm;

	result = copyout(((unsigned long)(&ssp[OPCODE_SSMEM_SOURCE(opcode)]) +
			  (4 - size)),
			 to, size);

	if (!result && OPCODE_M_UPDATE(opcode)) {
		/* Modify base register after */		
                if (!OPCODE_A_UPDATE(opcode))
                        to += imm;
                ssp[OPCODE_SSMEM_BASE(opcode)] = to;
        }
        return(result);
}

/*
 * trap_align_indexed_load - Align indexed load fault.
 */
align_indexed_load(opcode, ssp, size)
	unsigned long	opcode;
	unsigned long	*ssp;		/* XXX */
	int 		size;
{
        unsigned long	index, from;
        int 		result;
	unsigned char 	*reg_ptr;

        index = ssp[OPCODE_IMEM_X(opcode)];
        if (OPCODE_U_UPDATE(opcode))
                index <<= size;
	
        from = ssp[OPCODE_IMEM_BASE(opcode)];
	if (!OPCODE_M_UPDATE(opcode))		/* Pre-modify */
		from += index;

	reg_ptr = (unsigned char *)&ssp[OPCODE_IMEM_TARGET(opcode)];
	
	if (size == 1) {
		/*
	 	 * Halfword load zeros upper 16 bits.
		 */
		bzero(reg_ptr, 4);
		reg_ptr += 2;
	}
        result = copyin(from, reg_ptr, (1 << size));

        if (!result && OPCODE_M_UPDATE(opcode))	/* Post-modify and update */
                ssp[OPCODE_IMEM_BASE(opcode)] = from + index;

        return(result);
}
#endif
