/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

#include <machine/db_machdep.h>


#define GET_W(ins)   ((ins) & 0x00000001)
#define GET_W1(ins) (((ins) & 0x001f0000) >> 16)
#define GET_W2(ins) (((ins) & 0x00001ffc) >> 2)
#define GET_X(ins)  GET_W1(ins)
#define GET_B(ins)  (((ins) & 0x02e00000) >> 21)

#define GET_GR(regs, n) (*(((unsigned *)regs) + (((n)+1)*4)))


static unsigned
assemble_12(x,y)
     unsigned x, y; /* x is 11 bits; y is 1 bit */
{
  return ((y << 11) | ((x & 0x00000001) << 10) | (x >> 1));
}


static unsigned 
assemble_17(x, y, z)
     unsigned x, y, z; /* x is 5 bits, y is 11, and z is 1 */
{
  return ((z << 16) | (x << 15) | ((y & 0x00000001) << 10) | (y >> 1));
}


static unsigned
sign_ext(x, n)
     unsigned x, n;
{
  return (unsigned)((int)(x << (32 - n)) >> (32 - n));
}

unsigned
branch_taken(ins, addr, getreg, regs)
     unsigned ins;
     db_addr_t addr;
     unsigned (* getreg)();
     db_regs_t *regs;
{
  unsigned disp;

  switch (ins & 0xfc00e000)
    {
    case /* BL */   0xe8000000:
      disp = 
	(sign_ext(assemble_17(GET_W1(ins), GET_W2(ins), GET_W(ins)), 17)) << 2;
      return (regs->iioq_head + disp + 8);

    case /* GATE */ 0xe8002000:
      disp = 
	(sign_ext(assemble_17(GET_W1(ins), GET_W2(ins), GET_W(ins)), 17)) << 2;
      return (regs->iioq_head + disp + 8);
	
    case /* BLR */  0xe8004000:
      disp = ((GET_GR(regs, GET_B(ins))) << 3);

      return (regs->iioq_head + disp + 8);

    case /* BV */   0xe800c000:
      disp = 
	(GET_GR(regs, GET_B(ins)) + (GET_GR(regs, GET_X(ins)) << 3))
	  & 0xfffffffc;

      if ((regs->iioq_head & 0x00000003) 
	  < (GET_GR(regs, GET_B(ins)) & 0x00000003))
	return disp | (GET_GR(regs, GET_B(ins)) &0x00000003);
      else
	return disp | (regs->iioq_head & 0x00000003);

    case /* BE */   0xe0000000:
    case /* BLE */  0xe4000000:
      disp = 
	(sign_ext(assemble_17(GET_W1(ins), GET_W2(ins), GET_W(ins)), 17)) << 2;
      disp = (GET_GR(regs, GET_B(ins)) + disp) & 0xfffffffc;
      if ((regs->iioq_head & 0x00000003) 
	  < (GET_GR(regs, GET_B(ins)) & 0x00000003))
	return disp | (GET_GR(regs, GET_B(ins)) &0x00000003);
      else
	return disp | (regs->iioq_head & 0x00000003);

    case /* ADDBF */  0xa8000000:
    case /* ADDBT */  0xa0000000:
    case /* ADDIBF */ 0xac000000:
    case /* ADDIBT */ 0xa4000000:
    case /* COMBF */  0x88000000:
    case /* COMBT */  0x80000000:
    case /* COMIBF */ 0x8c000000:
    case /* COMIBT */ 0x84000000:
    case /* MOVB */   0xc8000000:
    case /* MOVIB */  0xcc000000:
    case /* BB */     0xc4000000:
    case /* BVB */    0xc0000000:
      disp = (sign_ext(assemble_12(GET_W1(ins), GET_W(ins)), 12)) << 2;
      return (regs->iioq_head + disp + 8);

    default:
      printf("Error: In branch taken unknown branching instruction %x\n",
	     ins);
      return 0;
    }
}

/*
 * This function is sent into branch_taken but is never called
 * needed to complete compilation
 */
unsigned
getreg_val (/* */)
{
  printf("Error: getreg_val called, but has no body.\n");
  return 0;
}
