/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

#include <machine/kdbdefs.h>

#define SFLG_VALID      0x0001           /* True for any valid symbol table. */

typedef struct {
        unsigned int    bot, top;        /* low/high addresses (inclusive) */
} range_t;                               /* range covered by a symbol table */

typedef struct {
        int             nranges;         /* Number of ranges */
        range_t         range[1];        /* Actual ranges (dummy array size) */
} counted_range_t;

#define symtable_t struct symtable_st
struct symtable_st {
        symtable_t      *next_symtable;         /* Next in linked list */
        int             flags;                  /* Use SFLG bits, above */
        counted_range_t *rng;                   /* Ranges this applies to */
        int             nsyms;                  /* Number of symbols */
        struct syment   *symtable;              /* First in array of symbols */
};
#define SNRNG rng->nranges
#define SRNG  rng->range

symtable_t *curtab = NULL;

/* Address range covering whole kernel */
counted_range_t vmunix_range = {               
        1,                                /* .nranges */
        {0, 0xFFFFFFFF}};                 /* .range (all possible addresses) */

/* Staticly-allocated table header for kernel symbols */
symtable_t vmunix_symtable = {
        NULL,                                   /* .next_symtable */
        0,                                      /* .flags */
        &vmunix_range};                         /* .rng */

symtable_t *kdb_symtables = &vmunix_symtable;   /* Symbol table descriptors */

int kdb_trace = FALSE;               /* Turn on for debug messages */
int kdb_onceonly = FALSE;            /* Turn on to avoid extra symbol tables */
#define dbmsg(x) if (kdb_trace) printf x

/*
 * Find the closest symbol to val, and return
 * the difference between val and the symbol found.
 * Accept only symbols within the table's range[s].
 * Leave a pointer to the symbol found as cursym.
 */
findsym(val, type)
unsigned long val;
int type;
{
	unsigned long diff;
	struct syment *sp;
        int inrange;
        symtable_t *stab;
        range_t *rng;

        dbmsg(("findsym: val=0x%x, type=%d\n", val, type));

	cursym = NULL;
        curtab = NULL;

	diff = 0x7fffffff;
	if (type != ISYM)
		return(diff);

        for (stab=kdb_symtables; stab != NULL; stab = stab->next_symtable) {
                dbmsg(("findsym: stab=0x%x\n", stab));

                /* Watch out for transient bad tables (see addkdbsyms) */
                if (!(stab->flags & SFLG_VALID)) 
                        continue;

                /* Check address ranges for the table. We know there's always a range (see addkdbsyms) */
                inrange = FALSE;
                for (rng=stab->SRNG; rng < &stab->SRNG[stab->SNRNG]; rng++)
                        if (inrange = ((rng->bot <= val) && (val <= rng->top)) )
                                break;

                if (!inrange) continue;

                dbmsg(("findsym: in range\n"));

        	for (sp = stab->symtable; sp < &stab->symtable[stab->nsyms]; sp++) {
        		if (sp->n_zeroes && sp->n_name[0] == '.')
        			continue;
        		if (val - sp->n_value < diff && val >= sp->n_value) {
        			diff = val - sp->n_value;
                                curtab = stab;
        			cursym = sp;
        			if (diff == 0) {
                                        dbmsg(("findsym: exact match\n"));
        				return (diff);
                                }
        		}
        	}
                if (kdb_onceonly) break;
        }

        dbmsg(("findsym: diff = 0x%x\n", diff));
	return(diff);
}

/*
 * Print value v and then the string s.
 * If v is not zero, then we look for a nearby symbol
 * and print name+offset if we find a symbol for which
 * offset is small enough.
 *
 * For values which are just into kernel address space
 * that they match exactly or that they be more than maxoff
 * bytes into kernel space.
 */
psymoff(v, type, s)
long v;
int type;
char *s;
{
        long w;
        int     i;
        char    nbuf[9];
        char    *nam;

        if (v)
                w = findsym(v, type);

        if (v == 0 || w >= maxoff)
                printf(LPRMODE, v);
        else {
                if (cursym->n_zeroes == 0)
                        nam = cursym->n_nptr;
                else {
                        /* 8-char name may not have null termination, so copy it
. */
                        for (i=0; i<8 && cursym->n_name[i]!='\0'; i++)
                                nbuf[i] = cursym->n_name[i];
                        nbuf[i] = '\0';
                        nam = nbuf;
                }

                printf("%s", nam);
                if (w)
                        printf(OFFMODE, w);
        }
        printf(s);
}
