/* 
 * Copyright (c) 1990,1991,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: csw.s 1.10 94/12/14$
 *	Author: Bob Wheeler, University of Utah CSL
 */

/*
 * Context switch and cproc startup for hppa
 */

#include <mach/machine/asm.h>
#include "csw.h"

	.space	$TEXT$
	.subspa	$CODE$

/*
 * void
 * cproc_switch(cur, next, lock)
 *	int *cur;
 *	int *next;
 * 	spin_lock_t *lock;
 *
 * Take the current thread of execution and save its context on its stack.
 * Place the pointer to the context into *cur. Then take the context pointed
 * to by *next and load it. Finally clear the lock.
 */

	.export	cproc_switch,entry
	.proc
	.callinfo
cproc_switch

	.entry
	/*
	 * save all the callee save registers for the current thread. We
	 * will also save rp so we know how to get back. We will save and
	 * restore arg0 since we need to pass some state in this when we
	 * bootstrap initial threads.
	 */
	stw	r2,CSW_R2(sp)
	stw	r3,CSW_R3(sp)
	stw	r4,CSW_R4(sp)
	stw	r5,CSW_R5(sp)
	stw	r6,CSW_R6(sp)
	stw	r7,CSW_R7(sp)
	stw	r8,CSW_R8(sp)
	stw	r9,CSW_R9(sp)
	stw	r10,CSW_R10(sp)
	stw	r11,CSW_R11(sp)
	stw	r12,CSW_R12(sp)
	stw	r13,CSW_R13(sp)
	stw	r14,CSW_R14(sp)
	stw	r15,CSW_R15(sp)
	stw	r16,CSW_R16(sp)
	stw	r17,CSW_R17(sp)
	stw	r18,CSW_R18(sp)
	stw	arg0,CSW_R26(sp)

	mfsp	sr3,t1
	stw	t1,CSW_SR3(sp)
	mfctl	cr11,t1
	stw	t1,CSW_CR11(sp)

	/*
	 * Save Necessary FPU State
	 * XXX - assuming that this is a real floating point unit
	 */

	/*
	 * Real fpu so first disable any pending traps by saving fr0
	 * and then save the exception registers and floating point registers
	 */
	ldo	CSW_FR0(sp),t1
	fstds,ma fr0,8(t1)
	fstds,ma fr1,8(t1)
	fstds,ma fr2,8(t1)
	fstds,ma fr3,8(t1)
	fstds,ma fr12,8(t1)
	fstds,ma fr13,8(t1)
	fstds,ma fr14,8(t1)
	fstds	 fr15,0(t1)

	/*
	 * save the context pointer, we'll just save the bottom of the context
	 * since our stacks grow up this will be much easier.
	 */
	stw	sp,0(arg0)


	/*
	 * now load up the new context
	 */
	ldw	0(arg1),sp

	ldw	CSW_R2(sp),r2
	ldw	CSW_R3(sp),r3
	ldw	CSW_R4(sp),r4
	ldw	CSW_R5(sp),r5
	ldw	CSW_R6(sp),r6
	ldw	CSW_R7(sp),r7
	ldw	CSW_R8(sp),r8
	ldw	CSW_R9(sp),r9
	ldw	CSW_R10(sp),r10
	ldw	CSW_R11(sp),r11
	ldw	CSW_R12(sp),r12
	ldw	CSW_R13(sp),r13
	ldw	CSW_R14(sp),r14
	ldw	CSW_R15(sp),r15
	ldw	CSW_R16(sp),r16
	ldw	CSW_R17(sp),r17
	ldw	CSW_R18(sp),r18
	ldw	CSW_R26(sp),arg0

	ldw	CSW_SR3(sp),t1
	mtsp	t1,sr3
	ldw	CSW_CR11(sp),t1
	mtctl	t1,cr11

	/*
	 * Restore FPU State
	 * XXX - assuming that this is a real floating point unit
	 */

	/*
	 * Real fpu so restore status register last to avoid a potential
	 * trap.  At this interface we only need to restore the entry
	 * save floating point registers.
	 */

	ldo	CSW_FR15(sp),t1
	fldds,ma -8(t1),fr15
	fldds,ma -8(t1),fr14
	fldds,ma -8(t1),fr13
	fldds,ma -8(t1),fr12
	fldds,ma -8(t1),fr3
	fldds,ma -8(t1),fr2
	fldds,ma -8(t1),fr1
	fldds    0(t1),fr0


	/*
	 * finally clear the lock pointed to by arg2
	 */

	/*
	 * align the lock to a 16 byte boundary
	 */
	ldo	15(arg2),arg2
	depi	0,31,4,arg2

	/*
	 * clear the lock by setting it non-zero
	 */
	ldi	-1,t1

	bv	0(rp)
	stw	t1,0(arg2)
	.exit
	.procend



/*
 * cproc_start_wait(parent_context, child, stackp, lock)
 *	int *parent_context;
 *	cproc_t	child;
 *	int stackp;
 * 	spin_lock_t *lock;
 *
 *
 * This routine is called when a thread blocks using cproc_block() and there 
 * is no other thread to run. What happens is that cproc_block() goes out 
 * and gets a "waiter" thread. These threads are really just empty threads
 * without any real state to them. They are just empty shells with a stack
 * associated with them.
 *
 * We then save the context of the parent like any other context switch but
 * instead of loading a context we just keep running but we use the waiter
 * thread's cproc_t structure and stack. (So we really didn't context switch, 
 * we just kept running and now we pretend that we are this waiter thread.)
 *
 * Then we call cproc_waiting() and pass it our new cproc_t pointer. This
 * routine will then wait for a real thread that can run and it will context
 * switch from our thread (dumping our context into our throw away stack)
 * and load the context of the new thread. 
 * 
 * We will never run again. In essence we have created this little dummy 
 * thread to take us from cproc_block() to cproc_waiting(). This could probably
 * have been done a lot easier by just having the original parent thread run
 * the code in cproc_waiting() inside cproc_block().
 */

	.export	cproc_start_wait,entry
	.proc
	.callinfo
cproc_start_wait

	.entry
	/*
	 * save all the callee save registers for the current thread. We
	 * will also save rp so we know how to get back. We will save and
	 * restore arg0 since we need to pass some state in this when we
	 * bootstrap initial threads.
	 */
	stw	r2,CSW_R2(sp)
	stw	r3,CSW_R3(sp)
	stw	r4,CSW_R4(sp)
	stw	r5,CSW_R5(sp)
	stw	r6,CSW_R6(sp)
	stw	r7,CSW_R7(sp)
	stw	r8,CSW_R8(sp)
	stw	r9,CSW_R9(sp)
	stw	r10,CSW_R10(sp)
	stw	r11,CSW_R11(sp)
	stw	r12,CSW_R12(sp)
	stw	r13,CSW_R13(sp)
	stw	r14,CSW_R14(sp)
	stw	r15,CSW_R15(sp)
	stw	r16,CSW_R16(sp)
	stw	r17,CSW_R17(sp)
	stw	r18,CSW_R18(sp)
	stw	arg0,CSW_R26(sp)

	mfsp	sr3,t1
	stw	t1,CSW_SR3(sp)
	mfctl	cr11,t1
	stw	t1,CSW_CR11(sp)

	/*
	 * Save Necessary FPU State
	 * XXX - assuming that this is a real floating point unit
	 */

	/*
	 * Real fpu so first disable any pending traps by saving fr0
	 * and then save the exception registers and floating point registers
	 */
	ldo	CSW_FR0(sp),t1
	fstds,ma fr0,8(t1)
	fstds,ma fr1,8(t1)
	fstds,ma fr2,8(t1)
	fstds,ma fr3,8(t1)
	fstds,ma fr12,8(t1)
	fstds,ma fr13,8(t1)
	fstds,ma fr14,8(t1)
	fstds	 fr15,0(t1)

	/*
	 * save the context pointer, we'll just save the bottom of the context
	 * since our stacks grow up this will be much easier.
	 */
	stw	sp,0(arg0)

	/*
	 * now clear the lock that is associated with the parent thread
	 */

	ldo	15(arg3),arg3
	depi	0,31,4,arg3

	/*
	 * clear the lock by setting it non-zero
	 */
	ldi	-1,t1
	stw	t1,0(arg3)

	/*
	 * now we need to switch onto the stack pointed to by stackp. First
	 * we'll align it to a double word boundary.
	 */

	ldo	7(arg2),arg2
	depi	0,31,3,arg2

	/*
	 * next make a frame marker so that when we do get to cproc_waiting()
	 * it will have a proper frame set up.
	 */
	ldo	64(arg2),sp

	/*
	 * clear the previous stack pointer so that we know there's nothing
	 * behind this frame should we ever try and unwind this thread.
	 */
	stw	r0,FM_PSP(sp)

	/*
	 * move the cproc_t pointer for child into the first argument so that
	 * we can call cproc_waiting() with it.
	 */
	copy	arg1,arg0

	.import	cproc_waiting,code
	ldil	L%cproc_waiting,t1
	ldo	R%cproc_waiting(t1),t1

	/*
	 * don't need a return pointer in r2, we're not comming back...
	 */
	bv,n    (t1)
	.exit
	.procend


/*
 * void
 * cproc_prepare(child, child_context, stack)
 * 	int child;
 *	int *child_context;
 *	int stack;
 *
 * This routine is called to start a child process and fake up the context
 * so that it looks like it was running once. We need to fake the context
 * switch so that it "returns" from cproc_switch() to cthread_body().
 */

	.export	cproc_prepare,entry
	.proc
	.callinfo
cproc_prepare

	.entry
	/*
	 * establish a user stack for this thread. There is no guarentee
	 * that the stack is aligned correctly. We have to align it to
	 * a double word boundary.
	 *
	 * stack_base doesn't account for self-pointer, hence "+4+7"
	 */
	ldo	11(arg2),arg2
	depi	0,31,3,arg2

	/*
	 * now create a stack frame so that cthread_body thinks everything
	 * is ok.
	 *
	 * This code must mimic the stack manipulations in cproc_setup() 
	 * since the server wants all threads to start with the stack
	 * offset being the same. 
	 */
	ldo	48(arg2),arg2

	/*
	 * basically we will just be saving 0 into most registers except for
	 * rp and arg0 since we need to pass the child argument to 
	 * cthread_body().
	 */
	.import	cthread_body,code
	ldil	L%cthread_body,t1
	ldo	R%cthread_body(t1),t1

	stw	t1,CSW_R2(arg2)

	stw	r0,CSW_R3(arg2)
	stw	r0,CSW_R4(arg2)
	stw	r0,CSW_R5(arg2)
	stw	r0,CSW_R6(arg2)
	stw	r0,CSW_R7(arg2)
	stw	r0,CSW_R8(arg2)
	stw	r0,CSW_R9(arg2)
	stw	r0,CSW_R10(arg2)
	stw	r0,CSW_R11(arg2)
	stw	r0,CSW_R12(arg2)
	stw	r0,CSW_R13(arg2)
	stw	r0,CSW_R14(arg2)
	stw	r0,CSW_R15(arg2)
	stw	r0,CSW_R16(arg2)
	stw	r0,CSW_R17(arg2)
	stw	r0,CSW_R18(arg2)

	/*
	 * arg0 is already conveniently what we need it to be
	 */
	stw	arg0,CSW_R26(arg2)

	/*
	 * the new thread should not assume the contents of sr3 or cr11
	 */
	stw	r0,CSW_SR3(arg2)
	stw	r0,CSW_CR11(arg2)

	/*
	 * We don't want the new process to inherit any floating point faults
	 * from the parent. Load all the floating point status registers
	 * with 0. We should also clear all the floatint point registers
	 * so that we don't magically end up with NaN in one of them when 
	 * we context switch to the thread.
	 */
	ldo	CSW_FR0(arg2),t1
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,4(t1)
	stwm	r0,0(t1)

	/*
	 * copy the context pointer into child_context
	 */

	bv	0(rp)
	stw	arg2,0(arg1)
	.exit
	.procend

/*
 * int
 * get_dp()
 *
 * Return the contents of the dp register.
 */

	.export	get_dp,entry
	.proc
	.callinfo
get_dp
	.entry
	bv	0(rp)
	copy	dp,ret0
	.exit
	.procend
